/* Wharf.c. by Bo Yang. 
 * Modifications: Copyright 1995 by Bo Yang.
 *
 * modifications made by Frank Fejes for AfterStep 
 * Copyright 1996
 *
 * folder code Copyright 1996 by Beat Christen.
 *
 * based on GoodStuff.c by Robert Nation 
 * The GoodStuff module, and the entire GoodStuff program, and the concept for
 * interfacing that module to the Window Manager, are all original work
 * by Robert Nation
 *
 * Copyright 1993, Robert Nation. 
 * No guarantees or warantees or anything
 * are provided or implied in any way whatsoever. Use this program at your
 * own risk. Permission to use this program for any purpose is given,
 * as long as the copyright is kept intact.  */

#define TRUE 1
#define FALSE 0
#define DOUBLECLICKTIME 1

#include "../../configure.h"

#ifdef ISC
#include <sys/bsdtypes.h> /* Saul */
#endif

#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/time.h>
#if defined ___AIX || defined _AIX || defined __QNX__ || defined ___AIXV3 || defined AIXV3 || defined _SEQUENT_
#include <sys/select.h>
#endif
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include "../../afterstep/module.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>

#include "Wharf.h"
#include "../../version.h"
#define AFTER_ICONS 1
#include "../../afterstep/asbuttons.h"

char *MyName;

Display *dpy;
int x_fd,fd_width;
int ROWS = FALSE;

Window Root;
int screen;
int flags;
long d_depth;

Bool Pushed = 0;
Bool Pushable = 1;
Pixel back_pix, fore_pix, light_grey;
GC  NormalGC, HiReliefGC, HiInnerGC;
Window main_win;
int Width, Height,win_x,win_y;
unsigned int display_width, display_height;

#define MW_EVENTS   (ExposureMask | ButtonReleaseMask |\
		     ButtonPressMask)
XSizeHints mysizehints;
int num_buttons = 0;
int num_folderbuttons = MAX_BUTTONS;
int num_folders = 0;
int num_rows = 0;
int num_columns = 0;
int max_icon_width = 30,max_icon_height = 0;
int BUTTONWIDTH, BUTTONHEIGHT;
int x= -100000,y= -100000,w= -1,h= -1,gravity = NorthWestGravity;
int new_desk = 0;
int pageing_enabled = 1;
int ready = 0;

int CurrentButton = -1;
int fd[2];

struct button_info Buttons[BUTTON_ARRAY_LN];
struct folder_info Folders[FOLDER_ARRAY_LN];

char *iconPath = NULL;
char *pixmapPath = NULL;

static Atom wm_del_win;
Atom _XA_WM_PROTOCOLS;
Atom _XA_WM_NAME;

/***********************************************************************
 *
 *  Procedure:
 *	main - start of afterstep
 *
 ***********************************************************************
 */
void main(int argc, char **argv)
{
  char *display_name = NULL;
  int i,j;
  Window root;
  int x,y,border_width,button;
  int depth;
  char *temp, *s;
  char set_mask_mesg[50];
  temp = argv[0];
  
  s=strrchr(argv[0], '/');
  if (s != NULL)
    temp = s + 1;
  
  
  MyName = safemalloc(strlen(temp)+1);
  strcpy(MyName, temp);
  
  for(i=0;i<BUTTON_ARRAY_LN;i++)
    {
      Buttons[i].title = NULL;
      Buttons[i].action = NULL;
      Buttons[i].icon_file = NULL;
      Buttons[i].icon_w = 0;
      Buttons[i].icon_h = 0;
      Buttons[i].IconWin = None;
      Buttons[i].icon_maskPixmap = None;	/* pixmap for the icon mask */
      Buttons[i].iconPixmap = None;
      Buttons[i].icon_depth = 0;
      Buttons[i].up = 1;                        /* Buttons start up */
      Buttons[i].hangon = NULL;                 /* don't wait on anything yet*/
      Buttons[i].folder = -1;
    }
  signal (SIGPIPE, DeadPipe);  
  
  if((argc != 6)&&(argc != 7))
    {
      fprintf(stderr,"%s Version %s should only be executed by AfterStep!\n",
		MyName, VERSION);
      exit(1);
    }

  fd[0] = atoi(argv[1]);
  fd[1] = atoi(argv[2]);
  
  if (!(dpy = XOpenDisplay(display_name))) 
    {
      fprintf(stderr,"%s: can't open display %s", MyName,
	      XDisplayName(display_name));
      exit (1);
    }
  x_fd = XConnectionNumber(dpy);

  fd_width = GetFdWidth();

  screen= DefaultScreen(dpy);
  Root = RootWindow(dpy, screen);
  if(Root == None) 
    {
      fprintf(stderr,"%s: Screen %d is not valid ", MyName, screen);
      exit(1);
    }
  
  display_width = DisplayWidth(dpy, screen);
  display_height = DisplayHeight(dpy, screen);
  
  d_depth = DefaultDepth(dpy, screen);

  sprintf(set_mask_mesg,"SET_MASK %lu\n",
	  (unsigned long)(M_TOGGLE_PAGING| 
			  M_NEW_DESK |
			  M_END_WINDOWLIST| 
			  M_MAP|
			  M_RES_NAME|
			  M_RES_CLASS|
			  M_WINDOW_NAME));

  SendText(fd,set_mask_mesg,0);

  ParseOptions(argv[3]);
 
  if(num_buttons == 0)
    {
      fprintf(stderr,"%s: No Buttons defined. Quitting\n", MyName);
      exit(0);
    }	
  GetXPMData( BACK_BUTTON, button_xpm);
  

  for(i=0;i<num_buttons;i++)
      LoadIconFile(i);

  for(i=num_folderbuttons;i<MAX_BUTTONS;i++)
      LoadIconFile(i);

  CreateWindow();
  for(i=0;i<num_buttons;i++)
    CreateIconWindow(i, &main_win);
    
  for(i=num_folderbuttons;i<MAX_BUTTONS;i++)
    CreateIconWindow(i, Buttons[i].parent); 
  
  XGetGeometry(dpy,main_win,&root,&x,&y,
	       (unsigned int *)&Width,(unsigned int *)&Height,
 	       (unsigned int *)&border_width,(unsigned int *)&depth);

  for(i=0;i<num_rows;i++)
    for(j=0;j<num_columns; j++)
      {
	button = i*num_columns + j;
	ConfigureIconWindow(button,i,j);
      };
  
  for(i=0;i<num_folders;i++)
    for(j=0;j<Folders[i].count;j++)
      if(num_columns < num_rows)
	ConfigureIconWindow(Folders[i].firstbutton+j,0, j);
      else if ((num_columns == num_rows) && (!ROWS))
	ConfigureIconWindow(Folders[i].firstbutton+j,0, j);
      else
	ConfigureIconWindow(Folders[i].firstbutton+j,j, 0);

  XMapSubwindows(dpy,main_win);
  XMapWindow(dpy,main_win);
  for(i=0;i<num_folders;i++)
      XMapSubwindows(dpy, Folders[i].win);

  /* request a window list, since this triggers a response which
   * will tell us the current desktop and paging status, needed to
   * indent buttons correctly */
  SendText(fd,"Send_WindowList",0);

  Loop();

}
/***********************************************************************
 *
 *  Procedure:
 *	Loop - wait for data to process
 *
 ***********************************************************************/
void Loop(void)
{
  Window root;
  Window *CurrentWin;
  int x,y,w,h,xoff,yoff,border_width,CurrentRow,CurrentColumn,CurrentBase;
  long depth;
  XEvent Event;
  int NewButton,i,j,button,tw,th,i2, bl=-1;
  int LastMapped=-1;
  char *temp;
  time_t t, tl = (time_t) 0;

  while(1)
    {
      if(My_XNextEvent(dpy,&Event))
	{
	  switch(Event.type)
	    {

	    case Expose:
	      for(x=0;x<num_folders;x++)
		if(Event.xany.window == Folders[x].win )
		  {
		    RedrawWindow(&Folders[x].win,num_folderbuttons, -1, Folders[x].cols,Folders[x].rows);
		    for(y=1;y<=Folders[x].count;y++)
		    if(num_columns<num_rows)
		      RedrawUnpushedOutline(&Folders[x].win, 1, y);
		    else
		      RedrawUnpushedOutline(&Folders[x].win, y, 1);
		  }
		    if (Pushed)
                break;
	      if((Event.xexpose.count == 0)&&
		 (Event.xany.window == main_win))
		{
		  if(ready < 1)
		    ready ++;
		  RedrawWindow(&main_win,0, -1, num_rows, num_columns);
		}
	      break;
	      
	    case ButtonPress:
	      if (Event.xbutton.button != Button1)
		break;

	      CurrentWin = &Event.xbutton.window;
	      CurrentBase =0;
	      CurrentRow = (Event.xbutton.y/BUTTONHEIGHT);
	      CurrentColumn = (Event.xbutton.x/BUTTONWIDTH);
	      CurrentButton = CurrentBase + CurrentColumn + CurrentRow*num_columns;
	      
	      for(x=0;x<num_folders;x++)
		if(*CurrentWin == Folders[x].win)
		  {
		    CurrentBase = Folders[x].firstbutton;
		    CurrentButton = CurrentBase + CurrentColumn + CurrentRow;
		  }
	      i = CurrentRow+1;
	      j = CurrentColumn +1;

	      if (Buttons[CurrentButton].swallow)
		break;
	      if (mystrncasecmp(Buttons[CurrentButton].action,"Folder",6)==0)
/* kludge until Beat takes a look */
if (num_columns > 1)
		MapFolder(Buttons[CurrentButton].folder, 
			  &LastMapped,
			  Event.xbutton.x_root-Event.xbutton.x-2,
			  Event.xbutton.y_root-Event.xbutton.y,
			  CurrentRow, CurrentColumn);
else if ((num_columns == 1) && (num_rows == 1))
		MapFolder(Buttons[CurrentButton].folder, 
			  &LastMapped,
			  Event.xbutton.x_root-Event.xbutton.x,
			  Event.xbutton.y_root-Event.xbutton.y,
			  1,1);
else
		MapFolder(Buttons[CurrentButton].folder, 
			  &LastMapped,
			  Event.xbutton.x_root-Event.xbutton.x,
			  Event.xbutton.y_root-Event.xbutton.y,
			  CurrentRow, CurrentColumn);
	      if (Pushable)
		{
		  Pushed = 1;
		  RedrawPushed(CurrentWin, i, j);
		}
              break;
	      
            case ButtonRelease:
	      if ((Event.xbutton.button != Button1) ||
		  (Buttons[CurrentButton].swallow)) 
                break;


              CurrentRow = (Event.xbutton.y/BUTTONHEIGHT);
              CurrentColumn = (Event.xbutton.x/BUTTONWIDTH);

              if (Pushable) 
              {
		Pushed=0;
		RedrawUnpushed(CurrentWin, i, j);
              }
	      
	      NewButton = CurrentBase + CurrentColumn + CurrentRow*num_columns;

              if(NewButton == CurrentButton)
                {
		  t = time( 0);
		  bl = -1;
		  tl = -1;
		  if(mystrncasecmp(Buttons[CurrentButton].action,"Folder",6)!=0)
		    {
		      if (LastMapped != -1 && CurrentWin != &main_win)
			{

			  XUnmapWindow(dpy, Folders[LastMapped].win);
			  Folders[LastMapped].mapped = NOTMAPPED;
			}
		      SendInfo(fd,Buttons[CurrentButton].action,0);
		    }		  
		  if((Buttons[CurrentButton].action)&&
		     (mystrncasecmp(Buttons[CurrentButton].action,"exec",4)== 0))
		    {  
		      i=4;
		      while((Buttons[CurrentButton].action[i] != 0)&&
			    (Buttons[CurrentButton].action[i] != '"'))
			i++;
		      i2=i+1;

		      while((Buttons[CurrentButton].action[i2] != 0)&&
			    (Buttons[CurrentButton].action[i2] != '"'))
			i2++;
		      if(i2 - i >1)
			{
                          Buttons[CurrentButton].hangon = safemalloc(i2-i);
                          strncpy(Buttons[CurrentButton].hangon,
				  &Buttons[CurrentButton].action[i+1],i2-i-1);
                          Buttons[CurrentButton].hangon[i2-i-1] = 0;
                          Buttons[CurrentButton].up = 0;
                          Buttons[CurrentButton].swallow = 0;
			}
		    }
                }
              break;
	      
	      /*
		case ClientMessage:
		if ((Event.xclient.format==32) && 
		(Event.xclient.data.l[0]==wm_del_win))
		{
		DeadPipe(1);
		}
		break;
		case PropertyNotify:
		if (Pushed)
		break;
		for(i=0;i<num_rows;i++)
		for(j=0;j<num_columns; j++)
		{
		button = i*num_columns + j;
		if((Buttons[button].swallow == 3)&&
		(Event.xany.window == Buttons[button].IconWin)&&
		(Event.xproperty.atom == XA_WM_NAME))
		{
		XFetchName(dpy, Buttons[button].IconWin, &temp);
		if(strcmp(Buttons[button].title,"-")!=0)
		CopyString(&Buttons[button].title, temp);
		XFree(temp);
		XClearArea(dpy,main_win,j*BUTTONWIDTH,
		i*BUTTONHEIGHT, BUTTONWIDTH,BUTTONHEIGHT,0);
		RedrawWindow(&main_win,0, button, num_rows, num_columns);
		}
		}
		break;
		*/
	    default:
	      break;
	    }
	}
    }
  return;
}
void MapFolder(int folder, int *LastMapped, int base_x, int base_y, int row, int col)
{
  if (Folders[folder].mapped ==ISMAPPED)
    {
      XUnmapWindow(dpy, Folders[folder].win);
      Folders[folder].mapped = NOTMAPPED;
      *LastMapped = -1;
    }
  else 
    {
      int folderx, foldery;
      if (*LastMapped != -1)
	{
	  XUnmapWindow(dpy, Folders[*LastMapped].win);
	  Folders[*LastMapped].mapped = NOTMAPPED;
	  *LastMapped = -1;
	}
      Folders[folder].mapped = ISMAPPED;
      if(num_columns < num_rows)
	{
	  if((base_x % display_width) > display_width / 2 )
	    folderx = base_x+(col-Folders[folder].count)*BUTTONHEIGHT-2;
	  else
	    folderx = base_x+(col+1)*BUTTONHEIGHT+1;
	  foldery = base_y+row*BUTTONWIDTH-1;
	}
/* more kludgery */
      else if (num_columns == num_rows)
        {
/*
	  if((base_x % display_width) > display_width / 2 )
	    folderx = (col-Folders[folder].count)*BUTTONHEIGHT-2;
	  else
	    folderx = (col+1)*BUTTONHEIGHT+1;
*/
          if (ROWS) 
          {
            if ((base_y % display_height) > display_height / 2)
              foldery = base_y-(Folders[folder].count)*BUTTONHEIGHT-2; 
            else 
	      foldery = base_y+BUTTONHEIGHT+1;
            folderx = base_x-1;
          }
          else
          {
	    if((base_x % display_width) > display_width / 2 )
              folderx = base_x-(Folders[folder].count)*BUTTONWIDTH-2;
            else
              folderx = base_x+BUTTONWIDTH+1;
            foldery = base_y-1; 
          }
        }
      else
	{
	  if ((base_y % display_height) < display_height / 2)
	    foldery  =base_y+(row+1)*BUTTONWIDTH+1;
	  else
	    foldery = base_y+(row-Folders[folder].count)*BUTTONWIDTH-2;
	  folderx = base_x+col*BUTTONHEIGHT+1;
       	}
      XMoveWindow(dpy, Folders[folder].win, folderx, foldery);
      XMapWindow(dpy, Folders[folder].win);
      *LastMapped = folder;
    }
}
void RedrawUnpushed(Window *win, int i, int j)
{
		  XMoveResizeWindow(dpy, Buttons[CurrentButton].IconWin,
		  (((ICON_WIN_WIDTH-Buttons[CurrentButton].icon_w)/2))
		  +8+((j-1)*BUTTONWIDTH) ,(i-1)*BUTTONHEIGHT+
		  ((((ICON_WIN_HEIGHT-Buttons[CurrentButton].icon_h)/2))
		  +8),BUTTONWIDTH,BUTTONHEIGHT);

		RedrawWindow(win,0, CurrentButton, num_rows, num_columns);

		RedrawUnpushedOutline(win, i, j);
}

void RedrawUnpushedOutline(Window *win, int i, int j)
{
/* top */
                XDrawLine( dpy, *win, HiInnerGC, 
                j*BUTTONWIDTH-BUTTONWIDTH, i*BUTTONHEIGHT-BUTTONHEIGHT, 
                j*BUTTONWIDTH,i*BUTTONHEIGHT-BUTTONHEIGHT);

                XDrawLine( dpy, *win, HiInnerGC, j*BUTTONWIDTH-BUTTONWIDTH,
                i*BUTTONHEIGHT-BUTTONHEIGHT+1, j*BUTTONWIDTH,
                i*BUTTONHEIGHT-BUTTONHEIGHT+1); 

/* left */
                XDrawLine( dpy, *win, HiInnerGC, j*BUTTONWIDTH-BUTTONWIDTH,
                i*BUTTONHEIGHT-BUTTONHEIGHT+1, j*BUTTONWIDTH-BUTTONWIDTH,
                i*BUTTONHEIGHT-1);

                XDrawLine( dpy, *win, HiInnerGC, j*BUTTONWIDTH
                -BUTTONWIDTH+1, i*BUTTONHEIGHT-BUTTONHEIGHT+2, 
                j*BUTTONWIDTH-BUTTONWIDTH+1 ,i*BUTTONHEIGHT-1);

/* right */
                XDrawLine( dpy, *win, NormalGC, j*BUTTONWIDTH-BUTTONWIDTH
                +BUTTONWIDTH-2, i*BUTTONHEIGHT-BUTTONHEIGHT+2, j*BUTTONWIDTH
                -BUTTONWIDTH+BUTTONWIDTH-2 ,i*BUTTONHEIGHT-1);

                XDrawLine( dpy, *win, NormalGC, j*BUTTONWIDTH-BUTTONWIDTH
                +BUTTONWIDTH-1, i*BUTTONHEIGHT-BUTTONHEIGHT+1, 
                j*BUTTONWIDTH-BUTTONWIDTH+BUTTONWIDTH-1 ,i*BUTTONHEIGHT-1);

/* bottom */
                XDrawLine( dpy, *win, NormalGC, j*BUTTONWIDTH
                -BUTTONWIDTH+1, i*BUTTONHEIGHT-1, j*BUTTONWIDTH-BUTTONWIDTH
                +BUTTONWIDTH-2,i*BUTTONHEIGHT-1);

                XDrawLine( dpy, *win, NormalGC, j*BUTTONWIDTH-BUTTONWIDTH
			   +1, i*BUTTONHEIGHT-2, j*BUTTONWIDTH-BUTTONWIDTH+BUTTONWIDTH-2,
			   i*BUTTONHEIGHT-2);

}
void RedrawPushed(Window *win, int i,int j)
{
		  XMoveResizeWindow(dpy, Buttons[CurrentButton].IconWin,
				    (((ICON_WIN_WIDTH-Buttons[CurrentButton].icon_w)/2))+9+(j-1)*BUTTONWIDTH ,
				    (i-1)*BUTTONHEIGHT+((((ICON_WIN_HEIGHT-Buttons[CurrentButton].icon_h)/2))+9),BUTTONWIDTH,
				    BUTTONHEIGHT);  
		  
		  RedrawWindow(win,0, CurrentButton, num_rows, num_columns);
		  RedrawPushedOutline(win, i,j);
}

void RedrawPushedOutline(Window *win, int i, int j)
{
		  /* Top Hilite */
		  XDrawLine( dpy, *win, NormalGC, j*BUTTONWIDTH-BUTTONWIDTH,
			     i*BUTTONHEIGHT-BUTTONHEIGHT, j*BUTTONWIDTH,i*BUTTONHEIGHT
			     -BUTTONHEIGHT);                                                          
		  
		  XDrawLine( dpy, *win, NormalGC, j*BUTTONWIDTH-BUTTONWIDTH,
			     i*BUTTONHEIGHT-BUTTONHEIGHT+1, j*BUTTONWIDTH,i*BUTTONHEIGHT
			     -BUTTONHEIGHT+1);                                                          
		  /* Left Hilite */
		  
		  XDrawLine( dpy, *win, NormalGC, j*BUTTONWIDTH-BUTTONWIDTH,
			     i*BUTTONHEIGHT-BUTTONHEIGHT+1, j*BUTTONWIDTH-BUTTONWIDTH,
			     i*BUTTONHEIGHT-1);
		  
		  XDrawLine( dpy, *win, NormalGC, j*BUTTONWIDTH-BUTTONWIDTH
			     +1, i*BUTTONHEIGHT-BUTTONHEIGHT+2, j*BUTTONWIDTH-BUTTONWIDTH+1,
			     i*BUTTONHEIGHT-1);
		  
		  /* Right Hilite */
		  
		  XDrawLine( dpy, *win, HiReliefGC, j*BUTTONWIDTH
			     -BUTTONWIDTH+BUTTONWIDTH-2, i*BUTTONHEIGHT-BUTTONHEIGHT+2,
			     j*BUTTONWIDTH-BUTTONWIDTH+BUTTONWIDTH-2 ,i*BUTTONHEIGHT-1);
		  
		  XDrawLine( dpy, *win, HiReliefGC, j*BUTTONWIDTH
			     -BUTTONWIDTH+BUTTONWIDTH-1, i*BUTTONHEIGHT-BUTTONHEIGHT+1,
			     j*BUTTONWIDTH-BUTTONWIDTH+BUTTONWIDTH-1 ,i*BUTTONHEIGHT-1);
		  
		  /* Bottom Hilite */
		  
		  XDrawLine( dpy, *win, HiReliefGC, j*BUTTONWIDTH
			     -BUTTONWIDTH+1, i*BUTTONHEIGHT-1, j*BUTTONWIDTH-BUTTONWIDTH
			     +BUTTONWIDTH-2,i*BUTTONHEIGHT-1);
		  
		  XDrawLine( dpy, *win, HiReliefGC, j*BUTTONWIDTH
			     -BUTTONWIDTH+1, i*BUTTONHEIGHT-2, j*BUTTONWIDTH-BUTTONWIDTH
			     +BUTTONWIDTH-2,i*BUTTONHEIGHT-2);

}
/************************************************************************
 *
 * Draw the window 
 *
 ***********************************************************************/
void RedrawWindow(Window *win, int firstbutton, int newbutton, int num_rows, int num_columns)
{
  int i,j,w,button, b;
  XEvent dummy;
  int val1,val2;
  
  if(ready < 1)
    return;
  
  while (XCheckTypedWindowEvent (dpy, *win, Expose, &dummy));

  for(i=0;i<num_rows;i++)
    for(j=0;j<num_columns; j++)
      {
	button = firstbutton+i*num_columns + j;
	if((newbutton == -1)||(newbutton == button))
	  {
	    if((Buttons[button].swallow == 3)&&
	       (Buttons[button].IconWin != None))
              XSetWindowBorderWidth(dpy,Buttons[button].IconWin,0);
	    
	    if(Buttons[button].title != NULL)
	      {
		b = BACK_BUTTON;
                if( Buttons[b].icon_depth == d_depth)
		  {
		    if (Pushed)
		      XCopyArea( dpy, Buttons[b].iconPixmap, *win,
				 NormalGC, 2, 2, Buttons[b].icon_w-2,
				 Buttons[b].icon_h-2, j*BUTTONWIDTH+2, i*BUTTONHEIGHT+2);
		    else
		      XCopyArea( dpy, Buttons[b].iconPixmap, *win,
				 NormalGC, 0, 0, Buttons[b].icon_w,
				 Buttons[b].icon_h, j*BUTTONWIDTH, i*BUTTONHEIGHT);
		    
		  }
                else
                  XCopyPlane( dpy, Buttons[b].iconPixmap, *win, NormalGC,
			      0, 0, Buttons[b].icon_w, Buttons[b].icon_h,
			      j*BUTTONWIDTH, i*BUTTONHEIGHT, 1);
	      }
	  }
      }
}


/************************************************************************
 *
 * Sizes and creates the window 
 *
 ***********************************************************************/
void CreateWindow(void)
{
  XGCValues gcv;
  unsigned long gcm;
  int first_avail_button,i,j,k;

  wm_del_win = XInternAtom(dpy,"WM_DELETE_WINDOW",False);
  _XA_WM_PROTOCOLS = XInternAtom (dpy, "WM_PROTOCOLS", False);

  /* Allow for multi-width/height buttons */
  first_avail_button = num_buttons;
  
  if(num_buttons > MAX_BUTTONS)
    {
      fprintf(stderr,"%s: Out of Buttons!\n",MyName);
      exit(0);
    }
      
  /* size and create the window */
  if((num_rows == 0)&&(num_columns == 0))
    num_columns = 1;
  if(num_columns == 0)
    {
      num_columns = num_buttons/num_rows;
      while(num_rows * num_columns < num_buttons)
	num_columns++;
    }
  if(num_rows == 0)
    {
      num_rows = num_buttons/num_columns;
      while(num_rows * num_columns < num_buttons)
	num_rows++;
    }

  while(num_rows * num_columns < num_buttons)
    num_columns++;

  mysizehints.flags = PWinGravity| PResizeInc | PBaseSize;
  /* subtract one for the right/bottom border */
  mysizehints.width = BUTTONWIDTH*num_columns;
  mysizehints.height= BUTTONHEIGHT*num_rows;
  mysizehints.width_inc = num_columns;
  mysizehints.height_inc = num_rows;
  mysizehints.base_height = num_rows - 1;
  mysizehints.base_width = num_columns - 1;

  if(x > -100000)
    {
      if (x <= -1)
	{
          mysizehints.x = DisplayWidth(dpy,screen) + x - mysizehints.width-2;
	  gravity = NorthEastGravity;
	}
      else if ((x == 0) && (flags & 16))
        mysizehints.x = DisplayWidth(dpy,screen) - mysizehints.width-2;
      else
	mysizehints.x = x;
      if ( y<0)
	{
	  mysizehints.y = DisplayHeight(dpy,screen) + y - mysizehints.height-2;
	  gravity = SouthWestGravity;
	}
      else 
	mysizehints.y = y;

      if((x < 0) && (y < 0))
	gravity = SouthEastGravity;	
      mysizehints.flags |= USPosition;
    }

  mysizehints.win_gravity = gravity;
  if(d_depth < 2)
    {
      back_pix = GetColor("white");
      fore_pix = GetColor("black");
    }
  else
    {
      back_pix = GetColor("grey50");
      fore_pix = GetColor("grey17");
      light_grey = GetColor("grey72");
    }

  main_win = XCreateSimpleWindow(dpy,Root,mysizehints.x,mysizehints.y,
				 mysizehints.width,mysizehints.height,
				 1,fore_pix,back_pix);

  for(i=0;i<num_folders;i++)
    {
      if(num_columns <num_rows)
	{
	  Folders[i].cols = 1;
	  Folders[i].rows = Folders[i].count;
	}
      else if ((num_columns == num_rows) && (!ROWS))
      {
	  Folders[i].cols = 1;
	  Folders[i].rows = Folders[i].count;
      }
      else
	{
	  Folders[i].cols = Folders[i].count;
	  Folders[i].rows = 1;
	}
      Folders[i].win = XCreateSimpleWindow(dpy, Root, 0,0,
					   BUTTONWIDTH*Folders[i].rows,BUTTONHEIGHT*Folders[i].cols, 
					   0,fore_pix,back_pix);  
      XSetWMNormalHints(dpy,Folders[i].win,&mysizehints);
      XSelectInput(dpy, Folders[i].win, MW_EVENTS);
     }
  
  XSetWMProtocols(dpy,main_win,&wm_del_win,1);

  XSetWMNormalHints(dpy,main_win,&mysizehints);

  XSelectInput(dpy, main_win, MW_EVENTS);
  change_window_name(MyName);

  gcm = GCForeground|GCBackground;
  gcv.foreground = fore_pix;
  gcv.background = back_pix;
  NormalGC = XCreateGC(dpy, Root, gcm, &gcv);  

  gcv.foreground = back_pix;
  gcv.background = fore_pix;
  HiReliefGC = XCreateGC(dpy, Root, gcm, &gcv);

  gcv.foreground = light_grey;
  gcv.background = fore_pix;
  HiInnerGC = XCreateGC(dpy, Root, gcm, &gcv);

}


void nocolor(char *a, char *b)
{
 fprintf(stderr,"%s: can't %s %s\n", MyName, a,b);
}

/****************************************************************************
 * 
 * Loads a single color
 *
 ****************************************************************************/ 
Pixel GetColor(char *name)
{
  XColor color;
  XWindowAttributes attributes;

  XGetWindowAttributes(dpy,Root,&attributes);
  color.pixel = 0;
   if (!XParseColor (dpy, attributes.colormap, name, &color)) 
     {
       nocolor("parse",name);
     }
   else if(!XAllocColor (dpy, attributes.colormap, &color)) 
     {
       nocolor("alloc",name);
     }
  return color.pixel;
}

/************************************************************************
 *
 * Dead pipe handler
 *
 ***********************************************************************/
void DeadPipe(int nonsense)
{
  int i,j,button;

  for(i=0;i<num_rows;i++)
    for(j=0;j<num_columns; j++)
      {
	button = i*num_columns + j;
        /* delete swallowed windows, but not modules (afterstep handles those) */
	if((Buttons[button].swallow == 3)&&(Buttons[button].module == 0))
	  {
	    send_clientmessage(Buttons[button].IconWin,wm_del_win,CurrentTime);
	    XSync(dpy,0);
	  }
      }
  XSync(dpy,0);
  exit(0);
}

int TOTHEFOLDER = -1;
/*****************************************************************************
 * 
 * This routine is responsible for reading and parsing the config file
 *
 ****************************************************************************/
void ParseOptions(char *filename)
{
  FILE *fd = (FILE *)0;
  char line[256];
  char *tline,*orig_tline,*tmp;
  int Clength, len;
  
  fd = fopen(filename,"r");
  if(fd == (FILE *)0)
    {
      fprintf(stderr,"%s: can't open config file %s",MyName,filename);
      exit(1);
    }
  
  tline = fgets(line,(sizeof line)-1,fd);
  orig_tline = tline;
  Clength = strlen(MyName);
  while(tline != (char *)0)
    {
      int g_x, g_y;
      unsigned width,height;
      
      while(isspace(*tline))tline++;
      
      if((strlen(&tline[0])>1)&&
	 (mystrncasecmp(tline,CatString3("*", MyName, "Geometry"),Clength+9)==0))
	{
	  tmp = &tline[Clength+9];
	  while(((isspace(*tmp))&&(*tmp != '\n'))&&(*tmp != 0))
	    {
	      tmp++;
	    }
	  tmp[strlen(tmp)-1] = 0;
	  
	  flags = XParseGeometry(tmp,&g_x,&g_y,&width,&height);
	  if (flags & WidthValue) 
	    w = width;
	  if (flags & HeightValue) 
	    h = height;
	  if (flags & XValue) 
	    x = g_x;
	  if (flags & YValue) 
	    y = g_y;
	}
      else if((strlen(&tline[0])>1)&&
	      (mystrncasecmp(tline,CatString3("*",MyName,"Rows"),Clength+5)==0))
	{
	  len=sscanf(&tline[Clength+5],"%d",&num_rows);
	  if(len < 1)
	    num_rows = 0;
            ROWS = TRUE;
	}
      else if((strlen(&tline[0])>1)&&
	      (mystrncasecmp(tline,CatString3("*",MyName,"Columns"),Clength+8)==0))
	{
	  len=sscanf(&tline[Clength+8],"%d",&num_columns);
	  if(len < 1)
	    num_columns = 0;
            ROWS = FALSE;
	}
      else if((strlen(&tline[0])>1)&&
              (mystrncasecmp(tline,CatString3("*",MyName,"NoPush"),Clength+5)==0))
        {
	  Pushable = 0;
        }

      else if((strlen(&tline[0])>1)&&
	      (mystrncasecmp(tline,CatString3("*", MyName, ""),Clength+1)==0)&&
	      (num_buttons < MAX_BUTTONS))
	{
	  match_string(&tline[Clength+1]);
	}
      else if((strlen(&tline[0])>1)&&(mystrncasecmp(tline,"IconPath",8)==0))
	{
	  CopyString(&iconPath,&tline[8]);
	}
      else if((strlen(&tline[0])>1)&&(mystrncasecmp(tline,"PixmapPath",10)==0))
	{
	  CopyString(&pixmapPath,&tline[10]);
	}
      tline = fgets(line,(sizeof line)-1,fd);
      orig_tline = tline;
    }
  
  return;
}


/**************************************************************************
 *
 * Parses a button command line from the config file 
 *
 *************************************************************************/
void match_string(char *tline)
{
  int len,i,i2,n;
  char *ptr,*start,*end,*tmp;
  struct button_info *actual;

  /* skip spaces */
  while(isspace(*tline)&&(*tline != '\n')&&(*tline != 0))
    tline++;

  /* read next word. Its the button label. Users can specify "" 
   * NoIcon, or whatever to skip the label */
  /* read to next space */
  start = tline;
  end = tline;
  while((!isspace(*end))&&(*end!='\n')&&(*end!=0))
    end++;
  len = end - start;
  ptr = safemalloc(len+1);
  strncpy(ptr,start,len);
  ptr[len] = 0;


  if (!strncmp(ptr,"~Folder",7))
    {
      TOTHEFOLDER = -1;
      Folders[num_folders].firstbutton = num_folderbuttons;
      num_folders++;
      free(ptr);
      return;
    }

  if(TOTHEFOLDER==-1)
    {
      actual = &Buttons[num_buttons++];
      actual->parent = &main_win;
    }
  else
    {
      actual = &Buttons[--num_folderbuttons];
      actual->folder = num_folders;
      actual->parent = &Folders[num_folders].win;
    };
  
  actual->title = ptr;      

  /* read next word. Its the icon bitmap/pixmap label. Users can specify "" 
   * NoIcon, or whatever to skip the label */
  /* read to next space */
  start = end;
  /* skip spaces */
  while(isspace(*start)&&(*start != '\n')&&(*start != 0))
    start++;
  end = start;
  while((!isspace(*end))&&(*end!='\n')&&(*end!=0))
    end++;
  len = end - start;
  ptr = safemalloc(len+1);
  strncpy(ptr,start,len);
  ptr[len] = 0;
  actual->icon_file = ptr;      

  tline = end;
  /* skip spaces */
  while(isspace(*tline)&&(*tline != '\n')&&(*tline != 0))
    tline++;

  if(mystrncasecmp(tline,"swallow",7)==0 || mystrncasecmp(tline,"maxswallow",10)==0)
    {
      /* Look for swallow "identifier", in which
	 case Wharf spawns and gobbles up window */
      i=7;
      while((tline[i] != 0)&&
	    (tline[i] != '"'))
	i++;
      i2=i+1;
      while((tline[i2] != 0)&&
	    (tline[i2] != '"'))
	i2++;
      actual->maxsize =
                 mystrncasecmp(tline,"maxswallow",10) == 0 ? 1 : 0;
      if(i2 - i >1)
	{
	  actual->hangon = safemalloc(i2-i);
	  strncpy(actual->hangon,&tline[i+1],i2-i-1);
	  actual->hangon[i2-i-1] = 0;
	  actual->swallow = 1;
	}
      n = 7;
      while((isspace(tline[n]))&&(tline[n]!=0))
	n++;
      len = strlen(&tline[n]);
      tmp = tline + n + len -1;
      while(((isspace(*tmp))||(*tmp == '\n'))&&(tmp >=(tline + n)))
	{
	  tmp--;
	  len--;
	}
      ptr = safemalloc(len+6);
      if(mystrncasecmp(&tline[n],"Module",6)==0)
	{
	  ptr[0] = 0;
          actual->module = 1;
	}	  
      else
	strcpy(ptr,"Exec ");
      i2 = strlen(ptr);
      strncat(ptr,&tline[n],len);
      ptr[i2+len]=0;
      SendText(fd,ptr,0);     
      free(ptr);
      actual->action = NULL;
    }
  else
    {
      if(!TOTHEFOLDER)
	{
	  Folders[num_folders].count++;
	}

      len = strlen(tline);
      tmp = tline + len -1;
      while(((isspace(*tmp))||(*tmp == '\n'))&&(tmp >=tline))
	{
	  tmp--;
	  len--;
	}
      ptr = safemalloc(len+1);
      strncpy(ptr,tline,len);
      ptr[len]=0;
      
      if (strncmp(ptr,"Folder",6)==0)
	{
	  TOTHEFOLDER = 0;
	  Folders[num_folders].count = 0;
	  actual->folder = num_folders;
	  Folders[num_folders].mapped = NOTMAPPED;
	}
      actual->action = ptr;
    }
  return;
}

/**************************************************************************
 *  Change the window name displayed in the title bar.
 **************************************************************************/
void change_window_name(char *str)
{
  XTextProperty name;
  int i;

  if (XStringListToTextProperty(&str,1,&name) == 0) 
    {
      fprintf(stderr,"%s: cannot allocate window name",MyName);
      return;
    }
  XSetWMName(dpy,main_win,&name);
  XSetWMIconName(dpy,main_win,&name);
  for(i=0;i<num_folders;i++)
    {
      XSetWMName(dpy, Folders[i].win,&name);
      XSetWMIconName(dpy, Folders[i].win,&name);
    }
  XFree(name.value);
}



/***************************************************************************
 *
 * Waits for next X event, or for an auto-raise timeout.
 *
 ****************************************************************************/
int My_XNextEvent(Display *dpy, XEvent *event)
{
  fd_set in_fdset;
  unsigned long header[3];
  int body_length;
  int count,count2 = 0;
  static int miss_counter = 0;
  unsigned long *body;
  int total;
  char *cbody;

  if(XPending(dpy))
    {
      XNextEvent(dpy,event);
      return 1;
    }

  FD_ZERO(&in_fdset);
  FD_SET(x_fd,&in_fdset);
  FD_SET(fd[1],&in_fdset);

#ifdef __hpux
  select(fd_width,(int *)&in_fdset, 0, 0, NULL);
#else
  select(fd_width,&in_fdset, 0, 0, NULL);
#endif


  if(FD_ISSET(x_fd, &in_fdset))
    {
      if(XPending(dpy))
	{
	  XNextEvent(dpy,event);
	  miss_counter = 0;
	  return 1;
	}
      else
	miss_counter++;
      if(miss_counter > 100)
	DeadPipe(0);
    }
  
  if(FD_ISSET(fd[1], &in_fdset))
    {
      if(count = ReadASPacket(fd[1], header, &body) > 0)
	{
	  process_message(header[1],body);
	  free(body);
	}
    }
  return 0;
}

void CheckForHangon(unsigned long *body)
{
  int button,i,j;
  char *cbody;

  cbody = (char *)&body[3];
  for(i=0;i<num_rows;i++)
    for(j=0;j<num_columns; j++)
      {
	button = i*num_columns + j;      
	if(Buttons[button].hangon != NULL)
	  {
	    if(strcmp(cbody,Buttons[button].hangon)==0)
	      {
		if(Buttons[button].swallow ==  1)
		  {
		    Buttons[button].swallow = 2;
		    if(Buttons[button].IconWin != None)
		      {
			XDestroyWindow(dpy,Buttons[button].IconWin);
		      }
		    Buttons[button].IconWin = (Window)body[0];
		    free(Buttons[button].hangon);
		    Buttons[button].hangon = NULL;
		  }
		else
		  {
		    Buttons[button].up = 1;
		    free(Buttons[button].hangon);
		    Buttons[button].hangon = NULL;
		    RedrawWindow(&main_win,0, button, num_rows, num_columns);
		  }
	      }
	  }
      }
}

/**************************************************************************
 *
 * Process window list messages 
 *
 *************************************************************************/
void process_message(unsigned long type,unsigned long *body)
{
  switch(type)
    {
    case M_TOGGLE_PAGING:
      pageing_enabled = body[0];
      RedrawWindow(&main_win,0, -1, num_rows, num_columns);
      break;
    case M_NEW_DESK:
      new_desk = body[0];
      RedrawWindow(&main_win,0, -1, num_rows, num_columns);
      break;
    case M_END_WINDOWLIST:
      RedrawWindow(&main_win,0, -1, num_rows, num_columns);
    case M_MAP:
      swallow(body);
    case M_RES_NAME:
    case M_RES_CLASS:
    case M_WINDOW_NAME:
      CheckForHangon(body);
      break;
    default:
      break;
    }
}





/***************************************************************************
 *
 * ICCCM Client Messages - Section 4.2.8 of the ICCCM dictates that all
 * client messages will have the following form:
 *
 *     event type	ClientMessage
 *     message type	_XA_WM_PROTOCOLS
 *     window		tmp->w
 *     format		32
 *     data[0]		message atom
 *     data[1]		time stamp
 *
 ****************************************************************************/
void send_clientmessage (Window w, Atom a, Time timestamp)
{
  XClientMessageEvent ev;
  
  ev.type = ClientMessage;
  ev.window = w;
  ev.message_type = _XA_WM_PROTOCOLS;
  ev.format = 32;
  ev.data.l[0] = a;
  ev.data.l[1] = timestamp;
  XSendEvent (dpy, w, False, 0L, (XEvent *) &ev);
}


void swallow(unsigned long *body)
{
  char *temp;
  int button,i,j;
  long supplied;

  for(i=0;i<num_rows;i++)
    for(j=0;j<num_columns; j++)
      {
	button = i*num_columns + j; 
	if((Buttons[button].IconWin == (Window)body[0])&&
	   (Buttons[button].swallow == 2))
	  {
	    Buttons[button].swallow = 3;
	    /* "Swallow" the window! */
	    XReparentWindow(dpy,Buttons[button].IconWin, main_win,
			    j*BUTTONWIDTH+4, i*BUTTONHEIGHT+4);
	    XMapWindow(dpy,Buttons[button].IconWin);
	    XSelectInput(dpy,(Window)body[0],
			 PropertyChangeMask);
	    if (Buttons[button].maxsize) {
	      Buttons[button].icon_w = 55;
	      Buttons[button].icon_h = 57;
	    }
	    else {
	      Buttons[button].icon_w = ICON_WIN_WIDTH;
	      Buttons[button].icon_h = ICON_WIN_HEIGHT;
	    }
	    if (!XGetWMNormalHints (dpy, Buttons[button].IconWin,
				    &Buttons[button].hints, 
				    &supplied))
	      Buttons[button].hints.flags = 0;
	    
	    XResizeWindow(dpy,(Window)body[0], Buttons[button].icon_w,
			  Buttons[button].icon_h);
	    XMoveWindow(dpy,Buttons[button].IconWin,
			j*BUTTONWIDTH +
			(BUTTONWIDTH - Buttons[button].icon_w)/2,
			i*BUTTONHEIGHT + 
			(BUTTONHEIGHT - Buttons[button].icon_h)/2);
	    
	    XFetchName(dpy, Buttons[button].IconWin, &temp);
	    XClearArea(dpy, main_win,j*BUTTONWIDTH, i*BUTTONHEIGHT,
		       BUTTONWIDTH,BUTTONHEIGHT,0);
	    if(strcmp(Buttons[button].title,"-")!=0)
	      CopyString(&Buttons[button].title, temp);
	    RedrawWindow(&main_win,0, -1, num_rows, num_columns);
	    XFree(temp);
	  }
      }
}



/***********************************************************************
 *
 *  Procedure:
 *      ConstrainSize - adjust the given width and height to account for the
 *              constraints imposed by size hints
 *
 *      The general algorithm, especially the aspect ratio stuff, is
 *      borrowed from uwm's CheckConsistency routine.
 * 
 ***********************************************************************/
void ConstrainSize (XSizeHints *hints, int *widthp, int *heightp)
{
#define makemult(a,b) ((b==1) ? (a) : (((int)((a)/(b))) * (b)) )
#define _min(a,b) (((a) < (b)) ? (a) : (b))

  
  int minWidth, minHeight, maxWidth, maxHeight, xinc, yinc, delta;
  int baseWidth, baseHeight;
  int dwidth = *widthp, dheight = *heightp;

  if(hints->flags & PMinSize)
    {
      minWidth = hints->min_width;
      minHeight = hints->min_height;
      if(hints->flags & PBaseSize)
	{
	  baseWidth = hints->base_width;
	  baseHeight = hints->base_height;
	}
      else
	{
	  baseWidth = hints->min_width;
	  baseHeight = hints->min_height;
	}
    }
  else if(hints->flags & PBaseSize)
    {
      minWidth = hints->base_width;
      minHeight = hints->base_height;
      baseWidth = hints->base_width;
      baseHeight = hints->base_height;
    }
  else
    {
      minWidth = 1;
      minHeight = 1;
      baseWidth = 1;
      baseHeight = 1;
    }
  
  if(hints->flags & PMaxSize)
    {
      maxWidth = hints->max_width;
      maxHeight = hints->max_height;
    }
  else
    {
      maxWidth = 10000;
      maxHeight = 10000;
    }
  if(hints->flags & PResizeInc)
    {
      xinc = hints->width_inc;
      yinc = hints->height_inc;
    }
  else
    {
      xinc = 1;
      yinc = 1;
    }
  
  /*
   * First, clamp to min and max values
   */
  if (dwidth < minWidth) dwidth = minWidth;
  if (dheight < minHeight) dheight = minHeight;
  
  if (dwidth > maxWidth) dwidth = maxWidth;
  if (dheight > maxHeight) dheight = maxHeight;
  
  
  /*
   * Second, fit to base + N * inc
   */
  dwidth = ((dwidth - baseWidth) / xinc * xinc) + baseWidth;
  dheight = ((dheight - baseHeight) / yinc * yinc) + baseHeight;
  
  
  /*
   * Third, adjust for aspect ratio
   */
#define maxAspectX hints->max_aspect.x
#define maxAspectY hints->max_aspect.y
#define minAspectX hints->min_aspect.x
#define minAspectY hints->min_aspect.y
  /*
   * The math looks like this:
   *
   * minAspectX    dwidth     maxAspectX
   * ---------- <= ------- <= ----------
   * minAspectY    dheight    maxAspectY
   *
   * If that is multiplied out, then the width and height are
   * invalid in the following situations:
   *
   * minAspectX * dheight > minAspectY * dwidth
   * maxAspectX * dheight < maxAspectY * dwidth
   * 
   */
  
  if (hints->flags & PAspect)
    {
      if (minAspectX * dheight > minAspectY * dwidth)
	{
	  delta = makemult(minAspectX * dheight / minAspectY - dwidth,
			   xinc);
	  if (dwidth + delta <= maxWidth) 
	    dwidth += delta;
	  else
	    {
	      delta = makemult(dheight - dwidth*minAspectY/minAspectX,
			       yinc);
	      if (dheight - delta >= minHeight) dheight -= delta;
	    }
	}
      
      if (maxAspectX * dheight < maxAspectY * dwidth)
	{
	  delta = makemult(dwidth * maxAspectY / maxAspectX - dheight,
			   yinc);
	  if (dheight + delta <= maxHeight)
	    dheight += delta;
	  else
	    { 
	      delta = makemult(dwidth - maxAspectX*dheight/maxAspectY,
			       xinc);
	      if (dwidth - delta >= minWidth) dwidth -= delta;
	    }
	}
    }
  
  *widthp = dwidth;
  *heightp = dheight;
  return;
}


