/*
 * static char *rcsid_arch_c =
 *   "$Id: arch.c,v 1.2 1993/03/09 06:31:24 frankj Exp $";
 */

/*
    CrossFire, A Multiplayer game for X-windows

    Copyright (C) 1992 Frank Tore Johansen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author can be reached via e-mail to frankj@ifi.uio.no.
*/

#include "arch.h"
extern void (*draw_info_func)(object *, char *txt);

static archetype *arch_table[ARCHTABLE];
int arch_cmp=0;		/* How many strcmp's */
int arch_search=0;	/* How many searches */

void init_archetypes() { /* called from add_player() and edit() */
  if(first_archetype!=NULL) /* Only do this once */
    return;
  load_archetypes();
  empty_archetype=find_archetype("empty_archetype");
  init_blocksview();
  switch(dump_monsters) {
  case 1:
    print_monsters();
    exit(0);
  case 2:
    dump_abilities();
    exit(0);
  }
}

void arch_info(object *op) {
  sprintf(errmsg,"%d searches and %d strcmp()'s",arch_search,arch_cmp);
  (*draw_info_func)(op,errmsg);
}

void clear_archetable() {
  memset(arch_table,0,ARCHTABLE*sizeof(archetype *)); /* NULL = 0, I hope */
}

void init_archetable() {
  archetype *at;
  LOG(llevDebug,"Setting up archetable...");
  for(at=first_archetype;at!=NULL;at=(at->more==NULL)?at->next:at->more)
    add_arch(at);
  if(debug)
  LOG(llevDebug,"done\n");
}

void dump_arch(archetype *at) {
  int i;
  dump_object(&at->clone);
  LOG(llevDebug,"Arch: %s [%d] = %s\n",at->name,at->animations,errmsg);
  for(i=0;i<at->animations;i++)
    LOG(llevDebug,"Anim %d=%d\n",i,at->faces[i]);
}

void dump_all_archetypes() {
  archetype *at;
  for(at=first_archetype;at!=NULL;at=(at->more==NULL)?at->next:at->more)
    dump_arch(at);
}

archetype *get_archetype_struct() {
  archetype *new;

  new=(archetype *)CALLOC(1,sizeof(archetype));
  if(new==NULL)
    fatal(OUT_OF_MEMORY);
  new->faces=NULL;
  new->next=NULL;
  new->name=NULL;
  new->randomitems=NULL;
  new->clone.other_arch=NULL;
  new->clone.name=NULL;
  new->clone.race=NULL;
  new->clone.slaying=NULL;
  new->clone.msg=NULL;
  UNSET_FREED((&new->clone)); /* This shouldn't matter, since copy_object() */
  SET_REMOVED((&new->clone)); /* doesn't copy these flags... */
  new->head=NULL;
  new->more=NULL;
  new->animations=0;
  return new;
}

void first_arch_pass(FILE *fp) {
  object *op;
  archetype *at,*prev=NULL,*last_more=NULL;
  int i;

  op=get_object();
  op->arch=first_archetype=at=get_archetype_struct();
  while((i=load_object(fp,op))) {
    copy_object(op,&at->clone);
    at->clone.speed_left= -0.1;
    switch(i) {
    case 1: /* A new archetype, just link it with the previous */
      if(last_more!=NULL)
        last_more->next=at;
      if(prev!=NULL)
        prev->next=at;
      prev=last_more=at;
      break;
    case 2:
      LOG(llevDebug,"Error: Archetype with inventory is illegal!\n");
      exit(-1);
    case 3: /* Another part of the previous archetype, link it correctly */
      at->head=prev;
      if(last_more!=NULL)
        last_more->more=at;
      last_more=at;
      break;
    }
    at=get_archetype_struct();
    clear_object(op);
    op->arch=at;
  }
  free_object(op);
}

void second_arch_pass(FILE *fp) {
  char buf[MAX_BUF],*variable=buf,*argument,*cp;
  archetype *at=NULL,*other;

  while(fgets(buf,MAX_BUF,fp)!=NULL) {
    if(*buf=='#')
      continue;
    if((argument=strchr(buf,' '))!=NULL) {
      *argument='\0',argument++;
      if((cp=strchr(argument,'\n'))!=NULL)
        *cp='\0';
    }
    if(!strcmp("Object",variable)) {
      if((at=find_archetype(argument))==NULL)
        LOG(llevError,"Warning: failed to find arch %s\n",argument);
    } else if(!strcmp("other_arch",variable)) {
      if(at!=NULL&&at->clone.other_arch==NULL) {
        if((other=find_archetype(argument))==NULL)
          LOG(llevError,"Warning: failed to find other_arch %s\n",argument);
        else if(at!=NULL)
          at->clone.other_arch=other;
      }
    } else if(!strcmp("randomitems",variable)) {
      if(at!=NULL) {
        treasurelist *tl=find_treasurelist(argument);
        if(tl==NULL)
          LOG(llevError,"Failed to link treasure to arch: %s\n",variable);
        else
          at->randomitems=tl->items;
      }
    }
  }
}

#if 0
void check_generators() {
  archetype *at;
  for(at=first_archetype;at!=NULL;at=at->next)
    if(IS_GENERATOR((&at->clone))&&at->clone.other_arch==NULL)
      LOG(llevError,"Warning: %s is generator but lacks other_arch.\n",
              at->name);
}
#endif

void load_archetypes() {
  FILE *fp;
  char filename[MAX_BUF];

  sprintf(filename,"%s/%s",LibDir,ArcheTypes);
  if((fp=open_and_uncompress(filename,0))==NULL) {
    LOG(llevError,"Can't open archetype file.\n");
    return;
  }
  clear_archetable();
  LOG(llevDebug,"arch-pass 1...");
  first_arch_pass(fp);
  LOG(llevDebug,"done\n");
  init_archetable();
  warn_archetypes=1;
  rewind(fp);
  LOG(llevDebug,"loading treasure...");
  load_treasures();
  LOG(llevDebug,"done\narch-pass 2...");
  second_arch_pass(fp);
  LOG(llevDebug,"done\n");
#if 0
  check_generators();
#endif
  close_and_delete(fp);
}

object *arch_to_object(archetype *at) {
  object *op;
  if(at==NULL) {
    if(warn_archetypes)
      LOG(llevError,"Couldn't find archetype.\n");
    return NULL;
  }
  op=get_object();
  copy_object(&at->clone,op);
  op->arch=at;
  return op;
}

object *get_archetype(char *name) {
  archetype *at;
  char *tmp=find_string(name);

  if(tmp!=NULL)
    for(at=first_archetype;at!=NULL;at=(at->more==NULL)?at->next:at->more)
      if(tmp==at->name)
        return arch_to_object(at);
  return NULL;
}

char *get_archename(int t) {
  archetype *at;

  for(at=first_archetype;at!=NULL;at=(at->more==NULL)?at->next:at->more)
    if(at->clone.type==t)
      return at->clone.name;
  return NULL;
}

unsigned long
hasharch(char *str) {
    unsigned long hash = 0;
    int i = 0, rot = 0;
    char *p;

    for (p = str; i < MAXSTRING && *p; p++, i++) {
        hash ^= (unsigned long) *p << rot;
        rot += 2;
        if (rot >= (sizeof(long) - sizeof(char)) * 8)
            rot = 0;
    }
    return (hash % ARCHTABLE);
}

archetype *find_archetype(char *name) {
  archetype *at;
  unsigned long index=hasharch(name);
  arch_search++;
  for(;;) {
    at = arch_table[index];
    if (at==NULL) {
      if(warn_archetypes)
        LOG(llevError,"Couldn't find archetype %s\n",name);
      return NULL;
    }
    arch_cmp++;
    if (!strcmp(at->name,name))
      return at;
    if(++index>ARCHTABLE)
      index=0;
  }
}

void add_arch(archetype *at) {
  int index=hasharch(at->name),org_index=index;
  for(;;) {
    if(arch_table[index]==NULL) {
      arch_table[index]=at;
      return;
    }
    if(++index==ARCHTABLE)
      index=0;
    if(index==org_index)
      fatal(ARCHTABLE_TOO_SMALL);
  }
}

#if 0
archetype *find_archetype(char *name) {
  archetype *at;
  char *tmp=find_string(name);
  int i;

  for(i=0;i<ARCH_CACHE_SIZE;i++)
    if(arch_cache[i].name==tmp)
      return arch_cache[i].arch;
  if(tmp!=NULL)
    for(at=first_archetype;at!=NULL;at=(at->more==NULL)?at->next:at->more)
      if(tmp==at->name) {
        arch_cache[last_arch_cache].name=at->name;
        arch_cache[last_arch_cache].arch=at;
        if(++last_arch_cache==ARCH_CACHE_SIZE)
          last_arch_cache=0;
        return at;
      }
  return NULL;
}
#endif

archetype *face_to_archetype(Fontindex face) {
  archetype *at;

  for(at=first_archetype;at!=NULL;at=(at->more==NULL)?at->next:at->more)
    if(at->clone.face==face)
      return at;
  return NULL;
}

archetype *type_to_archetype(int type) {
  archetype *at;

  for(at=first_archetype;at!=NULL;at=(at->more==NULL)?at->next:at->more)
    if(at->clone.type==type)
      return at;
  return NULL;
}

object *clone_arch_from_face(Fontindex face) {
  archetype *at;
  object *op=get_object();

  if((at=face_to_archetype(face))==NULL) {
    LOG(llevError,"Can't clone archetype with face %d\n",face);
    free_object(op);
    return NULL;
  }
  copy_object(&at->clone,op);
  return op;
}

object *clone_arch(int type) {
  archetype *at;
  object *op=get_object();

  if((at=type_to_archetype(type))==NULL) {
    LOG(llevError,"Can't clone archetype %d\n",type);
    free_object(op);
    return NULL;
  }
  copy_object(&at->clone,op);
  return op;
}
