/*
 * static char *rcsid_living_c =
 *   "$Id: living.c,v 1.1.1.1 1993/03/07 08:31:18 frankj Exp $";
 */

/*
    CrossFire, A Multiplayer game for X-windows

    Copyright (C) 1992 Frank Tore Johansen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author can be reached via e-mail to frankj@ifi.uio.no.
*/

#include <global.h>
#include <living.h>

extern void (*draw_stats_func)(object *);
extern void (*draw_info_func)(object *, char *);

void
set_attr_value(living *stats,int attr,signed char value) {
  switch(attr) {
  case STR:
    stats->Str=value;
    break;
  case DEX:
    stats->Dex=value;
    break;
  case CON:
    stats->Con=value;
    break;
  case WIS:
    stats->Wis=value;
    break;
  case CHA:
    stats->Cha=value;
    break;
  case INT:
    stats->Int=value;
    break;
  }
}

void
change_attr_value(living *stats,int attr,signed char value) {
  switch(attr) {
  case STR:
    stats->Str+=value;
    break;
  case DEX:
    stats->Dex+=value;
    break;
  case CON:
    stats->Con+=value;
    break;
  case WIS:
    stats->Wis+=value;
    break;
  case CHA:
    stats->Cha+=value;
    break;
  case INT:
    stats->Int+=value;
    break;
  }
}

signed char
get_attr_value(living *stats,int attr) {
  switch(attr) {
  case STR:
    return(stats->Str);
  case DEX:
    return(stats->Dex);
  case CON:
    return(stats->Con);
  case WIS:
    return(stats->Wis);
  case CHA:
    return(stats->Cha);
  case INT:
    return(stats->Int);
  }
  return 0;
}

void check_stat_bounds(object *op) {
  int i,v;
  for(i=0;i<6;i++)
    if((v=get_attr_value(&(op->stats),i))>25)
      set_attr_value(&(op->stats),i,25);
    else if(v<1)
      set_attr_value(&(op->stats),i,1);
}

#define ORIG_S(xyz,abc)	(op->contr->orig_stats.abc)

int change_abil(object *op, object *tmp) {
  int flag=IS_APPLIED(tmp)?1:-1,i,j;
  if(op->type==PLAYER) {
    if(tmp->type==POTION) {
      for(j=0;j<6;j++)
        if((i=get_attr_value(&(op->contr->orig_stats),j)
           +flag*get_attr_value(&(tmp->stats),j))
           >(20+tmp->stats.sp)+classbonus[op->contr->face][j] || i<1)
          return 0;
      for(j=0;j<6;j++)
        change_attr_value(&(op->contr->orig_stats),j,
                          flag*get_attr_value(&(tmp->stats),j));
      tmp->stats.sp=0;/* Fix it up for super potions */
    }
    for(j=0;j<6;j++)
      change_attr_value(&(op->stats),j,flag*get_attr_value(&(tmp->stats),j));
    check_stat_bounds(op);
  }
  if(flag==1)
    op->immune|=tmp->immune,
    op->protected|=tmp->protected,
    op->vulnerable|=tmp->vulnerable,
    op->attacktype|=tmp->attacktype;
  else
    op->immune&=~tmp->immune,
    op->protected&=~tmp->protected,
    op->vulnerable&=~tmp->vulnerable,
    op->attacktype&=~tmp->attacktype;
  if(tmp->attacktype & AT_CONFUSION) {
    if(flag>0)
      (*draw_info_func)(op,"Your hands begin to glow red.");
    else
      (*draw_info_func)(op,"Your hands stop glowing red.");
  }
  if(LIFESAVE(tmp)) {
    if(flag>0) {
      (*draw_info_func)(op,"You feel very protected.");
      SET_LIFESAVE(op);
    } else {
      (*draw_info_func)(op,"You don't feel protected anymore.");
      UNSET_LIFESAVE(op);
    }
  }
  if(REFL_MISSILE(tmp)) {
    if(flag>0) {
      (*draw_info_func)(op,"A magic force shimmers around you.");
      SET_REFL_MISSILE(op);
    } else {
      (*draw_info_func)(op,"The magic force fades away.");
      UNSET_REFL_MISSILE(op);
    }
  }
  if(REFL_SPELL(tmp)) {
    if(flag>0) {
      (*draw_info_func)(op,"You feel more safe now, somehow.");
      SET_REFL_SPELL(op);
    } else {
      (*draw_info_func)(op,"Suddenly you feel less safe, somehow.");
      UNSET_REFL_SPELL(op);
    }
  }
  if(IS_FLYING(tmp)) {
    if(flag>0) {
      if(IS_WIZ(op))
        (*draw_info_func)(op,"You float a little higher in the air.");
      else {
        (*draw_info_func)(op,"You start to float in the air!.");
        SET_FLY(op);
        if(op->speed>1)
          op->speed=1;
      }
    } else {
      if(IS_WIZ(op))
        (*draw_info_func)(op,"You float a little lower in the air.");
      else {
        (*draw_info_func)(op,"You float down to the ground.");
        UNSET_FLY(op);
        if(op->speed<=1)
          fix_player(op);
      }
    }
  }
  if(STEALTH(tmp)) {
    if(flag>0) {
      (*draw_info_func)(op,"You walk more quietly.");
      SET_STEALTH(op);
    } else {
      (*draw_info_func)(op,"You walk more noisily.");
      UNSET_STEALTH(op);
    }
  }
  if(XRAYS(tmp)) {
    if(flag>0) {
      if(IS_WIZ(op))
        (*draw_info_func)(op,"Your vision becomes a little clearer.");
      else {
        (*draw_info_func)(op,"Everything becomes transparent.");
        SET_XRAYS(op);
        if(op->type==PLAYER)
          op->contr->do_los=1;
      }
    } else {
      if(IS_WIZ(op))
        (*draw_info_func)(op,"Your vision becomes a bit out of focus.");
      else {
        (*draw_info_func)(op,"Everything suddenly looks very solid.");
        UNSET_XRAYS(op);
        if(op->type==PLAYER)
          op->contr->do_los=1;
      }
    }
  }
  if(tmp->stats.luck)
    if(flag>0) {
      (*draw_info_func)(op,"You feel more lucky.");
      op->stats.luck+=tmp->stats.luck;
    } else {
      (*draw_info_func)(op,"You feel less lucky.");
      op->stats.luck-=tmp->stats.luck;
    }
  if(tmp->stats.hp && op->type==PLAYER) {
    op->contr->gen_hp+=tmp->stats.hp*flag;
    if(flag*tmp->stats.hp>0)
      (*draw_info_func)(op,"You feel much more healthy!");
    else
      (*draw_info_func)(op,"You feel much less healthy!");
  }
  if(tmp->stats.sp && op->type==PLAYER) {
    op->contr->gen_sp+=tmp->stats.sp*flag;
    if(flag*tmp->stats.sp>0)
      (*draw_info_func)(op,"You feel one with the powers of magic!");
    else
      (*draw_info_func)(op,"You suddenly feel very mundane.");
  }
  if(tmp->stats.food && op->type==PLAYER) {
    op->contr->digestion+=tmp->stats.food*flag;
    if(tmp->stats.food*flag>0)
      (*draw_info_func)(op,"You feel your digestion slowing down.");
    else
      (*draw_info_func)(op,"You feel your digestion speeding up.");
  }
  if(tmp->immune&AT_PHYSICAL)
    if(flag>0)
      (*draw_info_func)(op,"You feel much less solid.");
    else
      (*draw_info_func)(op,"You suddenly feel very solid.");
  else if(tmp->protected&AT_PHYSICAL)
    if(flag>0)
      (*draw_info_func)(op,"You feel less solid.");
    else
      (*draw_info_func)(op,"You feel more solid.");
  if(tmp->immune&AT_MAGIC)
    if(flag>0)
      (*draw_info_func)(op,"You feel immune to magic.");
    else
      (*draw_info_func)(op,"You feel less immune to magic.");
  else if(tmp->protected&AT_MAGIC)
    if(flag>0)
      (*draw_info_func)(op,"You feel more resistant to magic.");
    else
      (*draw_info_func)(op,"You feel less resistant to magic.");
  if(tmp->immune&AT_FIRE)
    if(flag>0)
      (*draw_info_func)(op,"You feel immune to fire.");
    else
      (*draw_info_func)(op,"You feel less immune to fire.");
  else if(tmp->protected&AT_FIRE)
    if(flag>0)
      (*draw_info_func)(op,"You feel more resistant to fire.");
    else
      (*draw_info_func)(op,"You feel less resistant to fire.");
  if(tmp->vulnerable&AT_FIRE)
    if(flag>0)
      (*draw_info_func)(op,"You feel exposed to fire.");
    else
      (*draw_info_func)(op,"You feel less exposed to fire.");
  if(tmp->immune&AT_COLD)
    if(flag>0)
      (*draw_info_func)(op,"You feel immune to cold.");
    else
      (*draw_info_func)(op,"You feel less immune to cold.");
  else if(tmp->protected&AT_COLD)
    if(flag>0)
      (*draw_info_func)(op,"You feel more resistant to cold.");
    else
      (*draw_info_func)(op,"You feel less resistant to cold.");
  if(tmp->vulnerable&AT_COLD)
    if(flag>0)
      (*draw_info_func)(op,"You feel exposed to cold.");
    else
      (*draw_info_func)(op,"You feel less exposed to cold.");
  if(tmp->immune&AT_ELECTRICITY)
    if(flag>0)
      (*draw_info_func)(op,"You feel immune to electricity.");
    else
      (*draw_info_func)(op,"You feel less immune to electricity.");
  else if(tmp->protected&AT_ELECTRICITY)
    if(flag>0)
      (*draw_info_func)(op,"You feel more resistant to electricity.");
    else
      (*draw_info_func)(op,"You feel less resistant to electricity.");
  if(tmp->immune&AT_DRAIN)
    if(flag>0)
      (*draw_info_func)(op,"You feel very full of life.");
    else
      (*draw_info_func)(op,"You shiver, everything seems so bleak.");
  else if(tmp->protected&AT_DRAIN)
    if(flag>0)
      (*draw_info_func)(op,"You feel more resistant to draining.");
    else
      (*draw_info_func)(op,"You feel less resistant to draining.");
  if(tmp->immune&AT_POISON)
    if(flag>0)
      (*draw_info_func)(op,"You feel extremely healthy.");
    else
      (*draw_info_func)(op,"You feel extremely less healthy!");
  else if(tmp->protected&AT_POISON)
    if(flag>0)
      (*draw_info_func)(op,"You feel more resistant to poison.");
    else
      (*draw_info_func)(op,"You feel less resistant to poison.");
  if(tmp->immune&AT_SLOW)
    if(flag>0)
      (*draw_info_func)(op,"You feel in sync with time.");
    else
      (*draw_info_func)(op,"You feel out of sync with time.");
  else if(tmp->protected&AT_SLOW)
    if(flag>0)
      (*draw_info_func)(op,"You feel more in sync with time.");
    else
      (*draw_info_func)(op,"You feel less in sync with time.");
  if(tmp->immune&AT_PARALYZE)
    if(flag>0)
      (*draw_info_func)(op,"You feel very unrestrained.");
    else
      (*draw_info_func)(op,"You feel more restrained.");
  else if(tmp->protected&AT_PARALYZE)
    if(flag>0)
      (*draw_info_func)(op,"You feel more resistant to paralyzation.");
    else
      (*draw_info_func)(op,"You feel less resistant to paralyzation.");
  if(tmp->stats.Int)
    if(tmp->stats.Int*flag>0)
      (*draw_info_func)(op,"You feel smarter.");
    else
      (*draw_info_func)(op,"You feel dumb!");
  if(tmp->stats.Dex)
    if(tmp->stats.Dex*flag>0)
      (*draw_info_func)(op,"You feel more agile.");
    else
      (*draw_info_func)(op,"You feel clumsy!");
  if(tmp->stats.Str)
    if(tmp->stats.Str*flag>0)
      (*draw_info_func)(op,"You feel stronger.");
    else
      (*draw_info_func)(op,"You feel weaker!");
  if(tmp->stats.Wis)
    if(tmp->stats.Wis*flag>0)
      (*draw_info_func)(op,"You feel wiser.");
    else
      (*draw_info_func)(op,"You feel stupid!");
  if(tmp->stats.Cha)
    if(tmp->stats.Cha*flag>0)
      (*draw_info_func)(op,"You seem to look better.");
    else
      (*draw_info_func)(op,"You look ugly!");
  if(tmp->stats.Con)
    if(tmp->stats.Con*flag>0)
      (*draw_info_func)(op,"You feel healthy.");
    else
      (*draw_info_func)(op,"You feel less healthy!");
  return 1;
}

#ifdef NASTY_HACKS
void remove_statbonus(object *op) {
  int i;
  char *p;
  for(i=0,p= (char *) &op->stats;i<6;i++,p++)
    *p-=classbonus[op->contr->face][i];
}

void add_statbonus(object *op) {
  int i;
  char *p;
  for(i=0,p= (char *) &op->stats;i<6;i++,p++)
    *p+=classbonus[op->contr->face][i];
}
#else
void remove_statbonus(object *op) {
  int i,j;
  for(j=0;j<6;j++) {
    i= -classbonus[op->contr->face][j];
    change_attr_value(&(op->stats),j,i);
    change_attr_value(&(op->contr->orig_stats),j,i);
  }
}

void add_statbonus(object *op) {
  int i,j;
  for(j=0;j<6;j++) {
    i= classbonus[op->contr->face][j];
    change_attr_value(&(op->stats),j,i);
    change_attr_value(&(op->contr->orig_stats),j,i);
  }
}

#endif

void fix_player(object *op) {
  int i,j;
  float f,max=9,added_speed=0,bonus_speed=0;
  float M,W,s,D,K,S,M2;
  int weapon_weight=0,weapon_speed=0;
  int best_ac=0;
  object *tmp;
  if(op->type==PLAYER)
    for(i=0;i<6;i++)
      set_attr_value(&(op->stats),i,get_attr_value(&(op->contr->orig_stats),i));
  if(op->slaying!=NULL) {
    free_string(op->slaying);
    op->slaying=NULL;
  }
  if(!IS_WIZ(op)) {
    UNSET_FLY(op);
    UNSET_XRAYS(op);
  }
  if(op->arch==NULL) {
    UNSET_REFL_SPELL(op);
    UNSET_REFL_MISSILE(op);
    UNSET_LIFESAVE(op);
    UNSET_STEALTH(op);
    op->protected=0;
    op->immune=0;
    op->vulnerable=0;
    op->armour=0;
    op->stats.ac=10;
    op->stats.wc=21;
    op->stats.dam=1;
    op->stats.luck=0;
  } else {
    op->protected=op->arch->clone.protected;
    op->vulnerable=op->arch->clone.vulnerable;
    op->immune=op->arch->clone.immune;
    op->armour=op->arch->clone.armour;
    op->stats.ac=op->arch->clone.stats.ac;
    op->stats.wc=op->arch->clone.stats.wc;
    op->stats.dam=op->arch->clone.stats.dam;
    op->stats.luck=op->arch->clone.stats.luck;
    op->speed = op->arch->clone.speed;
  }
  for(tmp=op->inv;tmp!=NULL;tmp=tmp->below)
    if(IS_APPLIED(tmp)) {
      if(op->type==PLAYER)
        for(i=0;i<6;i++)
          change_attr_value(&(op->stats),i,get_attr_value(&(tmp->stats),i));
      op->protected|=tmp->protected;
      op->vulnerable|=tmp->vulnerable;
      op->immune|=tmp->immune;
      op->stats.luck+=tmp->stats.luck;
      if(LIFESAVE(tmp))
        SET_LIFESAVE(op);
      if(REFL_SPELL(tmp))
        SET_REFL_SPELL(op);
      if(REFL_MISSILE(tmp))
        SET_REFL_MISSILE(op);
      if(STEALTH(tmp))
        SET_STEALTH(op);
      if(IS_FLYING(tmp)) {
        SET_FLY(op);
        if(!IS_WIZ(op))
          max=1;
      }
      if(XRAYS(tmp))
        SET_XRAYS(op);
      if(tmp->stats.exp) {
        if(tmp->stats.exp > 0) {
          added_speed+=tmp->stats.exp-tmp->stats.exp/3 - 1;
          bonus_speed+=1+tmp->stats.exp/3;
        } else
          added_speed+=(float)tmp->stats.exp;
      }
      switch(tmp->type) {
      case RING:
      case AMULET:
      case GIRDLE:
      case HELMET:
      case SHIELD:
      case BOOTS:
      case GLOVES:
        if(tmp->armour)
          op->armour+=((100-op->armour)*tmp->armour)/100;
        if(tmp->stats.wc)
          op->stats.wc-=(tmp->stats.wc+tmp->magic);
        if(tmp->stats.dam)
          op->stats.dam+=(tmp->stats.dam+tmp->magic);
        if(tmp->stats.ac)
          op->stats.ac-=(tmp->stats.ac+tmp->magic);
        break;
      case WEAPON:
        op->stats.wc-=(tmp->stats.wc+tmp->magic);
        if(tmp->stats.ac&&tmp->stats.ac+tmp->magic>0)
          op->stats.ac-=tmp->stats.ac+tmp->magic;
        if(tmp->armour)
          op->armour+=((100-op->armour)*tmp->armour)/100;
        op->stats.dam+=(tmp->stats.dam+tmp->magic);
        weapon_weight=tmp->weight;
        weapon_speed=((int)WEAPON_SPEED(tmp)*2-tmp->magic)/2;
        if(weapon_speed<0) weapon_speed=0;
        if(tmp->slaying!=NULL)
          add_refcount(op->slaying = tmp->slaying);
        break;
      case ARMOUR: /* Only the best of these three are used: */
      case BRACERS:
      case FORCE:
        if(tmp->armour)
          op->armour+=((100-op->armour)*tmp->armour)/100;
        if(tmp->stats.ac) {
          if(best_ac<tmp->stats.ac+tmp->magic) {
            op->stats.ac+=best_ac; /* Remove last bonus */
            best_ac=tmp->stats.ac+tmp->magic;
          }
          else /* To nullify the below effect */
            op->stats.ac+=tmp->stats.ac+tmp->magic;
        }
        if(tmp->stats.ac) op->stats.ac-=(tmp->stats.ac+tmp->magic);
        if(ARMOUR_SPEED(tmp)&&ARMOUR_SPEED(tmp)/10.0<max)
          max=ARMOUR_SPEED(tmp)/10.0;
        break;
      }
    }
  if(op->type==PLAYER) {
    check_stat_bounds(op);
    for(i=1,op->stats.maxhp=0;i<=op->level&&i<=10;i++) {
      j=op->contr->levhp[i]+con_bonus[op->stats.Con]/2;
      if(i%2&&con_bonus[op->stats.Con]%2)
        j++;
      op->stats.maxhp+=j>1?j:1;
    }
    for(i=11;i<=op->level;i++)
      op->stats.maxhp+=2;
    if(op->stats.hp>op->stats.maxhp)
      op->stats.hp=op->stats.maxhp;
    for(i=1,op->stats.maxsp=0;i<=op->level&&i<=10;i++) {
      j=op->contr->levsp[i]+int_bonus[op->stats.Int]/2;
      if(i%2&&int_bonus[op->stats.Int]%2)
        j++;
      op->stats.maxsp+=j>1?j:1;
    }
    for(i=11;i<=op->level;i++)
      op->stats.maxsp+=2;
    if(op->stats.sp>op->stats.maxsp)
      op->stats.sp=op->stats.maxsp;
    if(op->contr->braced)
      op->stats.ac+=2;
    else
      op->stats.ac-=dex_bonus[op->stats.Dex];
    op->stats.wc-=(op->level+thaco_bonus[op->stats.Str]);
    if(op->contr->braced)
      op->stats.wc+=4;
    op->stats.dam+=dam_bonus[op->stats.Str];
    if(op->stats.dam<1)
      op->stats.dam=1;
    op->speed=1.0+speed_bonus[op->stats.Dex];
  }
  if(added_speed>=0)
    op->speed+=added_speed/10.0;
  else /* Something wrong here...: */
    op->speed /= (float)(1-added_speed);
  if(op->speed>max)
    op->speed=max;

  if(op->type == PLAYER) {
    f=(op->carrying/1000)-max_carry[op->stats.Str];
    if(f>0) op->speed=op->speed/(1+f/max_carry[op->stats.Str]);
  }

  op->speed+=bonus_speed/10.0; /* Not affected by limits */

  if(op->type == PLAYER) {
/* (This formula was made by vidarl@ifi.uio.no) */
    M=(max_carry[op->stats.Str]-121)/121.0;
    M2=max_carry[op->stats.Str]/100.0;
    W=weapon_weight/20000.0;
    s=2-weapon_speed/10.0;
    D=(op->stats.Dex-14)/14.0;
    K=1 + M/3.0 - W/(3*M2) + op->speed/5.0 + D/2.0;
    K*=(4+op->level)/(float)(6+op->level)*1.2;
    if(K<=0) K=0.01;
      S=op->speed/(K*s);
    op->contr->weapon_sp=S;
    if(op->stats.hp!=-10000)
      (*draw_stats_func)(op);
  }
  /* I want to limit the power of small monsters with big weapons: */
  if(op->arch!=NULL&&op->stats.dam>op->arch->clone.stats.dam*3)
    op->stats.dam=op->arch->clone.stats.dam*3;
}

#ifdef NASTY_HACKS
int allowed_class(object *op) {
  int i;
  char *p;
  for(i=0,p= (char *) &op->stats;i<6;i++,p++)
    if(*p<0) return 0;
  return 1;
}
#else
int allowed_class(object *op) {
  return op->stats.Dex>0&&op->stats.Str>0&&op->stats.Con>0&&
         op->stats.Int>0&&op->stats.Wis>0&&op->stats.Cha>0;
}
#endif

int level_exp(int level) {
  static int bleep=250000;
  if(level<=10) return levels[level];
  return levels[10]+bleep*(level-10);
}

void add_exp(object *op,int exp) {
  char buf[MAX_BUF];
  if(op->type==PLAYER&&op->contr->braced)
    exp=exp/5;
  op->stats.exp+=exp;
  if(op->stats.exp<0) op->stats.exp=0;
  if(op->stats.exp>99999999)
    op->stats.exp=99999999;
  if(op->type==PLAYER) {
    if(op->level<100&&op->stats.exp>=level_exp(op->level+1)) {
      op->level++;
      if(op->level<11)
        op->contr->levhp[op->level]=(int)RANDOM()%4+(int)RANDOM()%4+3,
        op->contr->levsp[op->level]=(int)RANDOM()%3+(int)RANDOM()%3+2;
      fix_player(op);
      if(op->level>1) {
        sprintf(buf,"You are now level %d.",op->level);
        (*draw_info_func)(op,buf);
      }
      add_exp(op,0); /* To increase more levels */
    } else if(op->level>1&&op->stats.exp<level_exp(op->level)) {
      op->level--;
      fix_player(op);
      sprintf(buf,"You are now level %d.",op->level);
      (*draw_info_func)(op,buf);
      add_exp(op,0); /* To decrease more levels */
    }
  }
}

char *describe_item(object *op) {
  char buf[MAX_BUF];
  errmsg[0]='\0';
  if(op->type!=PLAYER && !IS_MONSTER(op)) {
    if(op->stats.Str) {
      sprintf(buf,"(Str%+d)",op->stats.Str);
      strcat(errmsg,buf);
    }
    if(op->stats.Int) {
      sprintf(buf,"(Int%+d)",op->stats.Int);
      strcat(errmsg,buf);
    }
    if(op->stats.Dex) {
      sprintf(buf,"(Dex%+d)",op->stats.Dex);
      strcat(errmsg,buf);
    }
    if(op->stats.Con) {
      sprintf(buf,"(Con%+d)",op->stats.Con);
      strcat(errmsg,buf);
    }
    if(op->stats.Wis) {
      sprintf(buf,"(Wis%+d)",op->stats.Wis);
      strcat(errmsg,buf);
    }
    if(op->stats.Cha) {
      sprintf(buf,"(Cha%+d)",op->stats.Cha);
      strcat(errmsg,buf);
    }
    if(op->stats.wc) {
      sprintf(buf,"(wc%+d)",op->stats.wc);
      strcat(errmsg,buf);
    }
    if(op->stats.dam) {
      sprintf(buf,"(dam%+d)",op->stats.dam);
      strcat(errmsg,buf);
    }
    if(op->stats.ac) {
      sprintf(buf,"(ac%+d)",op->stats.ac);
      strcat(errmsg,buf);
    }
    if(op->stats.food) {
      if(op->stats.food>0)
        sprintf(buf,"(sustenance%+d)",op->stats.food);
      else
        sprintf(buf,"(hunger%+d)",op->stats.food);
      strcat(errmsg,buf);
    }
    if(op->stats.sp&&op->type!=BOW&&op->type!=WEAPON) {
      sprintf(buf,"(magic%+d)",op->stats.sp);
      strcat(errmsg,buf);
    }
    if(op->stats.hp) {
      sprintf(buf,"(regeneration%+d)",op->stats.hp);
      strcat(errmsg,buf);
    }
    if(op->stats.exp) {
      sprintf(buf,"(speed %+d)",op->stats.exp);
      strcat(errmsg,buf);
    }
  } else if(op->type == PLAYER) {
    if(op->contr->digestion) {
      if(op->contr->digestion>0)
        sprintf(buf,"(sustenance%+d)",op->contr->digestion);
      else
        sprintf(buf,"(hunger%+d)",op->contr->digestion);
      strcat(errmsg,buf);
    }
    if(op->contr->gen_sp) {
      sprintf(buf,"(magic%+d)",op->contr->gen_sp);
      strcat(errmsg,buf);
    }
    if(op->contr->gen_hp) {
      sprintf(buf,"(regeneration%+d)",op->contr->gen_hp);
      strcat(errmsg,buf);
    }
  }
  if(op->armour) {
    sprintf(buf,"(armour%+d)",op->armour);
    strcat(errmsg,buf);
  }
  if(op->stats.luck) {
    sprintf(buf,"(luck%+d)",op->stats.luck);
    strcat(errmsg,buf);
  }
  if(op->vulnerable&AT_PHYSICAL)
    strcat(errmsg,"(vulnerable physical)");
  if(op->protected&AT_PHYSICAL)
    strcat(errmsg,"(prot physical)");
  if(op->immune&AT_PHYSICAL)
    strcat(errmsg,"(immune physical)");

  if(op->vulnerable&AT_MAGIC)
    strcat(errmsg,"(vulnerable magic)");
  if(op->protected&AT_MAGIC)
    strcat(errmsg,"(prot magic)");
  if(op->immune&AT_MAGIC)
    strcat(errmsg,"(immune magic)");

  if(op->vulnerable&AT_ELECTRICITY)
    strcat(errmsg,"(vulnerable electricity)");
  if(op->protected&AT_ELECTRICITY)
    strcat(errmsg,"(prot electricity)");
  if(op->immune&AT_ELECTRICITY)
    strcat(errmsg,"(immune electricity)");

  if(op->vulnerable&AT_PARALYZE)
    strcat(errmsg,"(vulnerable paralyze)");
  if(op->protected&AT_PARALYZE)
    strcat(errmsg,"(protected paralyze)");
  if(op->immune&AT_PARALYZE)
    strcat(errmsg,"(free action)");

  if(op->vulnerable&AT_DRAIN)
    strcat(errmsg,"(vulnerable drain)");
  if(op->protected&AT_DRAIN)
    strcat(errmsg,"(prot drain)");
  if(op->immune&AT_DRAIN)
    strcat(errmsg,"(immune drain)");

  if(op->vulnerable&AT_SLOW)
    strcat(errmsg,"(vulnerable slow)");
  if(op->protected&AT_SLOW)
    strcat(errmsg,"(prot slow)");
  if(op->immune&AT_SLOW)
    strcat(errmsg,"(immune slow)");

  if(op->vulnerable&AT_COLD)
    strcat(errmsg,"(vulnerable cold)");
  if(op->protected&AT_COLD)
    strcat(errmsg,"(prot cold)");
  if(op->immune&AT_COLD)
    strcat(errmsg,"(immune cold)");

  if(op->vulnerable&AT_FIRE)
    strcat(errmsg,"(vulnerable fire)");
  if(op->protected&AT_FIRE)
    strcat(errmsg,"(prot fire)");
  if(op->immune&AT_FIRE)
    strcat(errmsg,"(immune fire)");

  if(op->vulnerable&AT_POISON)
    strcat(errmsg,"(vulnerable poison)");
  if(op->protected&AT_POISON)
    strcat(errmsg,"(prot poison)");
  if(op->immune&AT_POISON)
    strcat(errmsg,"(immune poison)");

  if(op->vulnerable&AT_CONFUSION)
    strcat(errmsg,"(vulnerable confusion)");
  if(op->protected&AT_CONFUSION)
    strcat(errmsg,"(prot confusion)");
  if(op->immune&AT_CONFUSION)
    strcat(errmsg,"(immune confusion)");

  if(op->vulnerable&AT_ACID)
    strcat(errmsg,"(vulnerable acid)");
  if(op->protected&AT_ACID)
    strcat(errmsg,"(prot acid)");
  if(op->immune&AT_ACID)
    strcat(errmsg,"(immune acid)");

  if(op->vulnerable&AT_FEAR)
    strcat(errmsg,"(vulnerable fear)");
  if(op->protected&AT_FEAR)
    strcat(errmsg,"(prot fear)");
  if(op->immune&AT_FEAR)
    strcat(errmsg,"(immune fear)");

  if(LIFESAVE(op))
    strcat(errmsg,"(lifesaving)");
  if(REFL_SPELL(op))
    strcat(errmsg,"(reflect spells)");
  if(REFL_MISSILE(op))
    strcat(errmsg,"(reflect missiles)");
  if(IS_FLYING(op))
    strcat(errmsg,"(levitate)");
  if(STEALTH(op))
    strcat(errmsg,"(stealth)");
  if(XRAYS(op))
    strcat(errmsg,"(xray-vision)");
  if(errmsg[0]=='\0'&&op->type==RING)
    strcat(errmsg,"(adorment +1)");
  if(op->slaying!=NULL) {
    strcat(errmsg,"(slay ");
    strcat(errmsg,op->slaying);
    strcat(errmsg,")");
  }
  if(IS_MONSTER(op)) {
    if(UNDEAD(op))
      strcat(errmsg,"(undead)");
    if(CAN_PASS_THRU(op))
      strcat(errmsg,"(pass through doors)");
    if(SEE_INVISIBLE(op))
      strcat(errmsg,"(see invisible)");
    if(CAN_USE_WEAPON(op))
      strcat(errmsg,"(wield weapon)");
    if(CAN_USE_BOW(op))
      strcat(errmsg,"(archer)");
    if(CAN_USE_ARMOUR(op))
      strcat(errmsg,"(wear armour)");
    if(CAN_USE_RING(op))
      strcat(errmsg,"(wear ring)");
    if(CAN_USE_SCROLL(op))
      strcat(errmsg,"(read scroll)");
    if(CAN_USE_WAND(op))
      strcat(errmsg,"(fire wand)");
    if(CAN_CAST_SPELL(op))
      strcat(errmsg,"(spellcaster)");
    if(IS_FRIENDLY(op))
      strcat(errmsg,"(friendly)");
    if(UNAGGRESSIVE(op))
      strcat(errmsg,"(unaggressive)");
    if(HITBACK(op))
      strcat(errmsg,"(hitback)");
    if(op->arch->randomitems != NULL) {
      treasure *t;
      int first = 1;
      for(t=op->arch->randomitems; t != NULL; t=t->next)
        if(t->item->clone.type == ABILITY) {
          if(first) {
            first = 0;
            strcat(errmsg,"(Spell abilities:)");
          }
          strcat(errmsg,"(");
          strcat(errmsg,t->item->clone.name);
          strcat(errmsg,")");
        }
    }
    if(op->attacktype && op->attacktype != 1) {
      int i;
      strcat(errmsg,"(Attacks:)");
      for(i=0;i<15;i++)
        if(op->attacktype & (2<<(i-1))) {
          strcat(errmsg,"(");
          strcat(errmsg,attacks[i]);
          strcat(errmsg,")");
        }
    }
  }
  return errmsg;
}
