/*
 * static char *rcsid_map_c =
 *   "$Id: map.c,v 1.2 1993/03/12 21:52:02 frankj Exp $";
 */

/*
    CrossFire, A Multiplayer game for X-windows

    Copyright (C) 1992 Frank Tore Johansen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author can be reached via e-mail to frankj@ifi.uio.no.
*/

#include <graphics.h>
#include <global.h>

#ifdef NeXT
extern char *tempnam(char *dir, char *pfx);
#endif

extern void (*process_active_maps_func)();
extern void (*update_buttons_func)(mapstruct *);
extern void (*draw_info_func)(object *, char *txt);
extern void (*fix_auto_apply_func) (mapstruct *);
extern int nrofallocobjects,nroffreeobjects;

int has_been_loaded(int i) {
  mapstruct *map;
  for(map = first_map; map != NULL; map = map->next)
    if(map->level == i)
      return 1;
  return 0;
}

void dump_map(mapstruct *m) {
  LOG(llevError,"Map %s status: %d.\n",m->name,m->in_memory);
  LOG(llevError,"Level: %d\n",m->level);
  LOG(llevError,"Size: %dx%d Start: %d,%d\n",
          m->mapx,m->mapy,m->startx,m->starty);
  LOG(llevError,"Message: %d lines, %d bytes.\n",m->mess_lines,m->mess_size);
  if(m->message!=NULL)
    LOG(llevError,"%s",m->message);
  if(m->tmpname!=NULL)
    LOG(llevError,"Tmpname: %s\n",m->tmpname);
}

void dump_all_maps() {
  mapstruct *m;
  for(m=first_map;m!=NULL;m=m->next) {
    LOG(llevError,"Map %s status %d.\n",m->name,m->in_memory);
    LOG(llevError,"Tmpname: %s\n",m->tmpname);
  }
}

int wall(mapstruct *m, int x,int y) {
  object *op;
  if(out_of_map(m,x,y)||get_omap(m,x,y)!=G_BLANK)
    return 1;
  for(op=get_map_ob(m,x,y);op!=NULL;op=op->above)
    if(NO_PASS(op))
      return 1;
  return 0;
}

int blocks_view(mapstruct *m, int x, int y) {
#if 0 /* I have made something smarter and faster 8) */
  object *op;
  if(out_of_map(m,x,y)||get_omap(m,x,y)!=G_BLANK)
    return 1;
  for(op=get_map_ob(m,x,y);op!=NULL;op=op->above)
    if(BLOCKSVIEW(op))
      return 1;
  return 0;
#else
  if(out_of_map(m,x,y))
    return 1;
  return blocksview[get_map(m,x,y)];
#endif
}

int blocks_magic(mapstruct *m, int x, int y) {
  object *op;
  if(out_of_map(m,x,y))
    return 1;
  for(op=get_map_ob(m,x,y);op!=NULL;op=op->above)
    if(NO_MAGIC(op))
      return 1;
  return 0;
}

int blocked(mapstruct *m, int x, int y) {
  object *tmp;
  if(out_of_map(m,x,y)||get_omap(m,x,y)!=G_BLANK)
    return 1;
  for(tmp=get_map_ob(m,x,y);tmp!=NULL;tmp=tmp->above)
    if(IS_ALIVE(tmp)||NO_PASS(tmp))
      return 1;
  return 0;
}

int blocked_link(object *ob, int x, int y) {
  object *tmp;
  if(out_of_map(ob->map,x,y) || get_omap(ob->map,x,y) != G_BLANK)
    return 1;
  if(ob->head != NULL)
    ob=ob->head;
  for(tmp = get_map_ob(ob->map,x,y); tmp!= NULL; tmp = tmp->above)
    if (NO_PASS(tmp) || (IS_ALIVE(tmp) && tmp->head != ob && tmp != ob))
      return 1;
  return 0;
}

/* Eneq(@csd.uu.se): This is a new version of blocked, this one handles objects
   that can be passed through by monsters with the CAN_PASS_THRU defined. */

int blocked_two(object *op, int x,int y) {
  object *tmp;
  if(out_of_map(op->map,x,y)||get_omap(op->map,x,y)!=G_BLANK)
    return 1;
  tmp=get_map_ob(op->map,x,y);
  for(tmp=get_map_ob(op->map,x,y);tmp!=NULL;tmp=tmp->above)
    if((IS_ALIVE(tmp)&&tmp->type!=DOOR)||(NO_PASS(tmp)&&!PASS_THRU(tmp))||
       (NO_PASS(tmp)&&PASS_THRU(tmp)&&!CAN_PASS_THRU(op)))
      return 1;
  return 0;
}

int arch_blocked(archetype *at,mapstruct *m,int x,int y) {
  archetype *tmp;
  if(at==NULL)
    return blocked(m,x,y);
  for(tmp=at;tmp!=NULL;tmp=tmp->more)
    if(blocked(m,x+tmp->clone.x,y+tmp->clone.y))
      return 1;
  return 0;
}

int arch_out_of_map(archetype *at,mapstruct *m,int x,int y) {
  archetype *tmp;
  if(at==NULL)
    return out_of_map(m,x,y);
  for(tmp=at;tmp!=NULL;tmp=tmp->more)
    if(out_of_map(m,x+tmp->clone.x,y+tmp->clone.y))
      return 1;
  return 0;
}

#ifndef SPEED_GAME

int out_of_map(mapstruct *m, int x, int y) {
  if(m==NULL) {
    LOG(llevError,"Out of map called with NULL\n");
    return 1;
  }
  return x<0||x>=m->mapx||y<0||y>=m->mapy;
}

#endif

void refresh_map(mapstruct *m) {
  int x,y;
  object *tmp,*tmp2,*active=NULL;
  Fontindex f;

  if(m==NULL)
    return;
  for(x=0;x<m->mapx;x++)
    for(y=0;y<m->mapy;y++) {
      f=get_omap(m,x,y);
/* Eneq(@csd.uu.se): Hunting down inappropriate objects in the map. The game
   sometime hangs and tries to remove_removed objects etc. */
      for(active=tmp=get_map_ob(m,x,y);tmp!=NULL;tmp=tmp2)
      { /* how can we get objects with pointer<1024, mta */
        if (tmp < (object *) 1024) /* map is uninitialized? */
          break;
        tmp2=tmp->above;
        active=tmp;
        if (IS_REMOVED(tmp)) {
          fprintf (stderr, "Crossfire: Found removed object in map.\n");
          active=(tmp->above==NULL?tmp->below:tmp->above);
          if (tmp->below==NULL&&tmp2==NULL)
            m->map_ob[x+m->mapx*y]=NULL;
          else if (tmp->below==NULL&&tmp2!=NULL) {
            tmp2->below=NULL;
            m->map_ob[x+m->mapx*y]=tmp2;
          } else if (tmp->below!=NULL&&tmp2==NULL)
            tmp->below->above=NULL;
          else if (tmp->below!=NULL&&tmp2!=NULL) {
            tmp->below->above=tmp2;
            tmp2->below=tmp->below;
          }
          if (!IS_FREED(tmp))
            free_object(tmp);
        }
        if (IS_FREED(tmp)&&!IS_REMOVED(tmp)) {
          fprintf (stderr, "Crossfire: Found freed object in map.\n");
          remove_ob(tmp);
        }
      }

      if(active != NULL)
        update_object(active);
      else
        set_map(m,x,y,get_omap(m,x,y));
    }
}

/*
 * open_and_uncompress() first searches for the original filename.
 * if it exist, then it opens it and returns the file-pointer.
 * if not, it does two things depending on the flag.  If the flag
 * is set, it tries to create the original file by uncompressing a .Z file.
 * If the flag is not set, it creates a temporarily file and uncompresses
 * the .Z file into that file.  The temporary file will be deleted in
 * close_and_delete().
 */

FILE *open_and_uncompress(char *name,int flag) {
  FILE *fp;
  char buf[MAX_BUF],buf2[MAX_BUF];
  if((fp=fopen(name,"r"))!=NULL)
    return fp;
  strcpy(buf,name);
  strcat(buf,".Z");
  if((fp=fopen(buf,"r"))==NULL)
    return NULL;
  fclose(fp);
  if((fp=fopen(UNCOMPRESS,"r"))==NULL) {
    LOG(llevError,"Error: Can't find uncompress(%s).\n",UNCOMPRESS);
    return NULL;
  }
  fclose(fp);
  if(editor) /* Hack that always uncompress the original file if in editor */
    flag=1;
  if(flag)
    sprintf(buf2,"%s %s",UNCOMPRESS,buf);
  else
    sprintf(buf2,"%s < %s > %s/uncomp.%d",UNCOMPRESS,buf,TMPDIR,getpid());
  system(buf2);
  if(flag) {
    chmod(name,0664);
    return fopen(name,"r");
  }
  delete_last_file=1;
  sprintf(buf,"%s/uncomp.%d",TMPDIR,getpid());
  return fopen(buf,"r");
}

void close_and_delete(FILE *fp) {
  char buf[MAX_BUF];
  fclose(fp);
  if(!delete_last_file)
    return;
  delete_last_file=0;
  sprintf(buf,"%s/uncomp.%d",TMPDIR,getpid());
  unlink(buf);
}

void save_omap(mapstruct *m,int flag) {
  FILE *fp;
  char filename[MAX_BUF];
  int i,j;

  if(!strcmp("Unnamed",m->name)) {
    LOG(llevError,"Tried to save unnamed map.\n");
    return;
  }
  if(flag) {
    if(m->path) {
      sprintf(filename,"%s/%s/%s.om",LibDir,MapDir,m->path);
      make_path_to_file(filename);
    } else
      sprintf(filename,"%s/%s/%s.om",LibDir,MapDir,m->name);
  } else {
    if(m->tmpname==NULL)
      m->tmpname=tempnam(TMPDIR,NULL);
    strcpy(filename,m->tmpname);
    strcat(filename,".om");
  }
  LOG(llevDebug,"Saving map %s (%d)\n",m->name,m->level);
  if((fp=fopen(filename,"w"))==NULL) {
    perror("Can't save omap file");
    return;
  }
  fprintf(fp,"mapx %d\n",m->mapx);
  fprintf(fp,"mapy %d\n",m->mapy);
  fprintf(fp,"startx %d\n",m->startx);
  fprintf(fp,"starty %d\n",m->starty);
  fprintf(fp,"mess_size %d\n",m->mess_size);
  fprintf(fp,"mess_lines %d\n",m->mess_lines);
  if(m->mess_lines)
    fprintf(fp,"%s",m->message);
  for(i=0;i<m->mapx;i++)
    for(j=0;j<m->mapy;j++)
      fprintf(fp,"%d\n",get_omap(m,i,j));
  fclose(fp);
  if(flag&&m->read_only)
    chmod(filename,0644);
  else
    chmod(filename,0664);
}

#if defined (MACH) || defined (NeXT)
#ifndef S_IWGRP
#define S_IWGRP 0000020
#endif
#ifndef S_IWUSR
#define S_IWUSR 0000002
#endif
#endif
void check_read_only(mapstruct *m) {
  char buf[MAX_BUF];
  struct stat statbuf;

  if(m->path!=NULL)
    sprintf(buf,"%s/%s/%s.om",LibDir,MapDir,m->path);
  else
    sprintf(buf,"%s/%s/%s.om",LibDir,MapDir,m->name);
  if(stat(buf,&statbuf)) {
    strcat(buf,".Z");
    if(stat(buf,&statbuf)) {
      LOG(llevError,"Can't open %s\n",buf);
      perror("open");
      return;
    }
  }
  m->read_only=0;
#if 0
  m->read_only=(!(statbuf.st_mode&S_IWGRP&&getegid()==statbuf.st_gid)&&
                !(statbuf.st_mode&S_IWUSR&&geteuid()==statbuf.st_uid));
#endif
}
    
void load_omap(mapstruct *m,int flag) { /* flag 1=original, 0=temporary */
  FILE *fp;
  char filename[MAX_BUF];
  int i,j,c;
  char buf[MAX_BUF];
  char bigbuf[VERY_BIG_BUF];

  if(flag)
    check_read_only(m);
  if(editor&&m->read_only) {
    LOG(llevError,"Fatal error: Tried to read closed map.\n");
    exit(0);
  }

  bigbuf[0]='\0';
  if(flag) {
    if(m->path!=NULL)
      sprintf(filename,"%s/%s/%s.om",LibDir,MapDir,m->path);
    else
      sprintf(filename,"%s/%s/%s.om",LibDir,MapDir,m->name);
  } else {
    if(m->tmpname!=NULL) {
      strcpy(filename,m->tmpname);
      strcat(filename,".om");
    } else
      sprintf(filename,"%s/%s.om.%d",TMPDIR,m->name,getpid());
  }
  if((fp=open_and_uncompress(filename,0))==NULL) {
    LOG(llevError,"Can't open %s\n",filename);
    perror("Can't read omap file");
    return;
  }
  fscanf(fp,"mapx %d\n",&m->mapx);
  fscanf(fp,"mapy %d\n",&m->mapy);
  fscanf(fp,"startx %d\n",&m->startx);
  fscanf(fp,"starty %d\n",&m->starty);
  fscanf(fp,"mess_size %d\n",&m->mess_size);
  fscanf(fp,"mess_lines %d\n",&m->mess_lines);
  allocate_map(m);
  clear_map(m);
  if(m->mess_lines) {
    for(i=0;i<m->mess_lines;i++) {
      fgets(buf,MAX_BUF,fp);
      strcat(bigbuf,buf);
    }
    m->message=strdup_local(bigbuf);
  } else
    m->message=NULL;
  for(i=0;i<m->mapx;i++)
    for(j=0;j<m->mapy;j++) {
      if(!*fgets(buf,MAX_BUF,fp)) {
        LOG(llevError,"Error in omap file: Too early EOF.\n");
        close_and_delete(fp);
        return;
      }
      c=atol(buf);
      set_map(m,i,j,(Fontindex) c);
      set_omap(m,i,j,(Fontindex) c);
    }
  close_and_delete(fp);
}

#if defined(MACH) || defined(vax) || defined(ibm032) || defined(NeXT)
#ifndef S_ISDIR
#define S_ISDIR(x) (((x) & S_IFMT) == S_IFDIR)
#endif
#endif
void make_path_to_file (char *filename)
{
    char buf[MAX_BUF], *cp = buf;
    struct stat statbuf;

    if (!filename || !*filename)
	return;
    strcpy (buf, filename);
    while ((cp = strchr (cp + 1, (int) '/'))) {
	*cp = '\0';
	fprintf (stderr, "Checking %s...", buf);
	if (stat (buf, &statbuf) || !S_ISDIR (statbuf.st_mode)) {
	    fprintf (stderr, "Was not dir...");
	    if (mkdir (buf, 0777)) {
		perror ("Couldn't make path to file");
		return;
	    }
	    fprintf (stderr, "Made dir.\n");
	} else
	    fprintf (stderr, "Was dir\n");
	*cp = '/';
    }
}
#if 0
void make_path_to_file(char *filename) {
  char buf[MAX_BUF],*cp;
  struct stat statbuf;
  for(cp=filename;*cp!='\0';cp++)
    if(*cp=='/') {
      strncpy(buf,filename,(int)(cp-filename+1));
      buf[cp-filename+2]='\0';
      LOG(llevDebug,"Checking %s...",buf);
      if(stat(buf,&statbuf)||!S_ISDIR(statbuf.st_mode)) {
        LOG(llevDebug,"Was not dir...");
        if(mkdir(buf,0660)) {
          perror("Couldn't make path to file");
          return;
        }
        LOG(llevDebug,"Made dir.\n");
      } else LOG(llevDebug,"Was dir\n");
    }
}
#endif

void save_map(mapstruct *m, int flag) { /* flag: 0=tmp map, 1=original map */
  FILE *fp;
  char filename[MAX_BUF];
  int i,j;
  object *m_ob,*i_ob, *m2_ob;;

  if(m->level==(-1)||!strcmp("Unnamed",m->name)) {
    LOG(llevError,"Trying to save unnamed map.\n");
    return;
  }
  if(flag) {
    if(m->path) {
      sprintf(filename,"%s/%s/%s.oo",LibDir,MapDir,m->path);
      make_path_to_file(filename);
    } else
      sprintf(filename,"%s/%s/%s.oo",LibDir,MapDir,m->name);
  } else {
    if(m->tmpname==NULL)
      m->tmpname=tempnam(TMPDIR,NULL);
    strcpy(filename,m->tmpname);
    strcat(filename,".oo");
  }
  if((fp=fopen(filename,"w"))==NULL) {
    LOG(llevError,"Can't open %s\n",filename);
    perror("Can't save temporary file");
    return;
  }

/* Eneq(@csd.uu.se): Refresh the map before saving it. Hopefully this will mean
   that we won't have any removed/freed objects in the map. */
  refresh_map(m);

  for(i=0;i<m->mapx;i++)
    for(j=0;j<m->mapy;j++)
      for(m_ob=get_map_ob(m,i,j);m_ob!=NULL;m_ob=m2_ob) {
        m2_ob=m_ob->above;
        if(m_ob->type==PLAYER||m_ob->head!=NULL)
          continue;
        if(m_ob->owner!=NULL)
          continue;
        if (m_ob->more==NULL)
          save_object(fp,m_ob,flag ? 3 : 1);/* don't check and remove */
        else {
          save_object(fp,m_ob,3); /* don't check and don't remove */
          for(i_ob=m_ob->more;i_ob!=NULL;i_ob=i_ob->more) {
            fprintf(fp,"More\n");
            save_object(fp,i_ob,3); /* don't check and don't remove */
          }
	  if (!flag) {
	      remove_ob (m_ob);
	      free_object (m_ob);
	  }
        }
#if 0 /* Why is inventory and links removed? */
        for(i_ob=m_ob->inv;i_ob!=NULL;i_ob=i_ob->below) {
          fprintf(fp,"Inventory\n");
          save_object(fp,i_ob);
        }
        for(i_ob=m_ob->more;i_ob!=NULL;i_ob=i_ob->more) {
          fprintf(fp,"More\n");
          save_object(fp,i_ob);
        }
#endif
      }
  fclose(fp);
  if(flag&&m->read_only)
    chmod(filename,0644);
  else
    chmod(filename,0664);
}

void check_background(mapstruct *m,int x,int y) {
  if(get_omap(m,x,y)!=G_BLANK) {
    LOG(llevError,"Fixed invisible wall.\n");
    set_omap(m,x,y,G_BLANK);
  }
}

void restore_map(mapstruct *m, char *filename,int flag) {
  FILE *fp;
  object *op,*prev=NULL,*last_more=NULL, *tmp, *otmp;
  int i, j = 0;

  if((fp = open_and_uncompress(filename,0))==NULL) {
    perror("Can't load object file");
    LOG(llevError,"Filename: %s\n",filename);
    return;
  }
  op=get_object();
  while((i=load_object(fp,op))) {
    switch(i) {
    case 1:
      insert_ob_in_map(op,m);
      for (tmp=op->inv, op->inv=NULL;tmp!=NULL;tmp=otmp) {
        if (tmp->nrof)
            op->carrying-=tmp->weight*tmp->nrof;
        else
            op->carrying-=tmp->weight+tmp->carrying;
        SET_NO_FIX_PLAYER(tmp);
        otmp=tmp->below;
        insert_ob_in_ob(tmp,op);
        UNSET_NO_FIX_PLAYER(tmp);
      }
      check_background(m,op->x,op->y);
      prev=op,last_more=op;
      break;
    case 2:
      insert_ob_in_ob(op,prev);
      break;
    case 3:
      insert_ob_in_map(op,m);
      check_background(m,op->x,op->y);
      op->head=prev,last_more->more=op,last_more=op;
      break;
    }
    if(!editor&&m->first_visit&&op->arch!=NULL&&op->arch->randomitems!=NULL)
      create_treasure(op->arch->randomitems,op,GT_INVENTORY,m->difficulty);
    op=get_object();
    if(j++>CHECK_ACTIVE_MAPS) {
      j=0;
      (*process_active_maps_func)(); /* We don't want them to wait any longer */
    }
  }
  free_object(op);
  close_and_delete(fp);
  (*update_buttons_func)(m); /* In case other objects press some buttons down */
}

void clear_map(mapstruct *m) {
  Fontindex *aptr,*bptr,*endptr;

  aptr=m->map;
  bptr=m->omap;
  endptr=m->map+m->mapx*m->mapy;
  for(;aptr<endptr;++aptr,++bptr) {
    *aptr = G_BLANK;
    *bptr = G_BLANK;
  }
#if 0
  memset(m->map,G_BLANK,sizeof(Fontindex)*m->mapx*m->mapy);
  memset(m->omap,G_BLANK,sizeof(Fontindex)*m->mapx*m->mapy);
#endif
  memset(m->map_ob, 0,sizeof(object *)*m->mapx*m->mapy);
}

void copy_map(mapstruct *m1, mapstruct *m2) { /* This routine only copies pointers, */
  int x,y;                                    /* *NOT* objects! */
  m2->level=m1->level;
  strncpy(m2->name,m1->name,BIG_NAME);
  if(m1->message==NULL)
    m2->message=NULL;
  else
    m2->message=strdup_local(m1->message);
  if(m1->path)
    add_refcount(m2->path = m1->path);
  else
    m2->path=NULL;
  m2->mess_size=m1->mess_size,m2->mess_lines=m1->mess_lines;
  for(x=0;x<m1->mapx&&x<m2->mapx;x++)
    for(y=0;y<m1->mapy&&y<m2->mapy;y++) {
      set_omap(m2,x,y,get_omap(m1,x,y));
      set_map(m2,x,y,get_omap(m1,x,y));
    }
}

void clean_object(object *op) {
  object *tmp;
  while((tmp=op->inv)!=NULL) {
    clean_object(tmp);

/* Eneq(@csd.uu.se): We'll have to move up the inventory or else op->inv
   always will point to the same object. */
    
    if (tmp->below!=NULL)
      tmp->below->above=NULL;
    op->inv=tmp->below;
    tmp->below=NULL;

    remove_ob(tmp);
    free_object(tmp);
  }
}

void free_all_objects(mapstruct *m) {
  int i,j;
  object *op;

  for(i=0;i<m->mapx;i++)
    for(j=0;j<m->mapy;j++) {
      object *previous_obj=NULL;
        while((op=get_map_ob(m,i,j))!=NULL) {
          if (op==previous_obj)
            break;
          previous_obj=op;
          if(op->head!=NULL)
            op=op->head;
          clean_object(op);
          remove_ob(op);
          free_object(op);
        }
      }
}

/* move_all_objects(): Only used from the editor(s?) */

void move_all_objects(mapstruct *m1, mapstruct *m2) {
  int i,j;
  object *op;

  for(i=0;i<m1->mapx&&i<m2->mapx;i++)
    for(j=0;j<m1->mapy&&j<m2->mapy;j++) {
      while((op=get_map_ob(m1,i,j))!=NULL&&op->head==NULL) {
        remove_ob(op);
        op->x=i,op->y=j; /* Not really needed */
        insert_ob_in_map(op,m2);
      }
    }
  free_all_objects(m1);
}

void free_map(mapstruct *m,int flag) {
  if(!m->in_memory) {
    LOG(llevError,"Trying to free freed map.\n");
    return;
  }
  if(flag)
    free_all_objects(m);
  m->in_memory=MAP_SWAPPED;
  CFREE(m->map);
  CFREE(m->omap);
  CFREE(m->map_ob);
  if(m->path!=NULL) {
    free_string(m->path);
    m->path=NULL;
  }
  if (m->message) free(m->message);
  m->message=NULL;
}

void load_temporary_objects(mapstruct *m) {
  char filename[MAX_BUF];

  if(m->path!=NULL)
    sprintf(filename,"%s/xf.%d/%s.oo",TMPDIR,getpid(),m->path);
  else {
    if(m->tmpname!=NULL) {
      strcpy(filename,m->tmpname);
      strcat(filename,".oo");
    } else
      sprintf(filename,"%s/%s.oo.%d",TMPDIR,m->name,getpid());
  }
  restore_map(m,filename,0);
}

void load_original_objects(mapstruct *m, int flag) {
  char filename[MAX_BUF];

  if(m->path!=NULL)
    sprintf(filename,"%s/%s/%s.oo",LibDir,MapDir,m->path);
  else
    sprintf(filename,"%s/%s/%s.oo",LibDir,MapDir,m->name);
  m->difficulty=5; /* For create_object() in monsters */
  restore_map(m,filename,flag);
  m->difficulty=calculate_difficulty(m);
  if(!flag)
    (*fix_auto_apply_func)(m); /* Chests which open as default */
}

void draw_map_message(object *op) {
  if(op->map->message!=NULL)
    (*draw_info_func)(op,op->map->message);
}

mapstruct *get_empty_map(int sizex, int sizey) {
  mapstruct *tmp;
  mapstruct *m=(mapstruct *) CALLOC(1,sizeof(mapstruct));

  if(m==NULL)
    fatal(OUT_OF_MEMORY);
  for(tmp=first_map;tmp!=NULL&&tmp->next!=NULL;tmp=tmp->next);
  if(tmp==NULL)
    first_map=m;
  else
    tmp->next=m;
  m->tmpname=NULL;
  m->mapx=sizex,m->mapy=sizey,m->next=NULL,m->mess_lines=0,
  m->level= -1,m->in_memory = MAP_IN_MEMORY;
  m->players=0,m->startx=2,m->starty=2,
  m->read_only=0;
  m->mess_size=0,m->message=NULL;
  m->pending = (objectlink *) NULL;
  m->path=NULL;
  m->omap=(Fontindex *) CALLOC(sizex*sizey,sizeof(Fontindex));
  m->map=(Fontindex *) CALLOC(sizex*sizey,sizeof(Fontindex));
  m->map_ob=(object **) CALLOC(sizex*sizey,sizeof(object *));
  if(m->omap==NULL||m->map==NULL||m->map_ob==NULL)
    fatal(OUT_OF_MEMORY);
  clear_map(m);
  return m;
}

void allocate_map(mapstruct *m) {
  if(m->in_memory != MAP_SWAPPED )
    return;
  m->in_memory = MAP_IN_MEMORY;
  m->omap=(Fontindex *) CALLOC(m->mapx*m->mapy,sizeof(Fontindex));
  m->map=(Fontindex *) CALLOC(m->mapx*m->mapy,sizeof(Fontindex));
  m->map_ob=(object **) CALLOC(m->mapx*m->mapy,sizeof(object *));
  if(m->omap==NULL||m->map==NULL||m->map_ob==NULL)
    fatal(OUT_OF_MEMORY);
}

mapstruct *ready_map(int level, int flag) {
  mapstruct *tmp;
  int first_visit=0;

/* Have we been at this level before? */
  for(tmp=first_map;tmp!=NULL;tmp=tmp->next)
    if(tmp->level==level)
      break;
  if(tmp==NULL) {
    tmp=find_and_load(level); /* No, load the mapstruct (size) */
    first_visit=1;
  }
  if(tmp==NULL)
    return NULL;
  if(tmp->in_memory == MAP_LOADING)
    return tmp;	/* I assume the caller will check this again (and block) */
  if(tmp->in_memory == MAP_IN_MEMORY)
    return tmp;
  if(first_visit || (tmp->reset_time != -1 && seconds() > tmp->reset_time)) {
    if(!first_visit)
      clean_tmp_map(tmp);	/* Doesn't make much difference */
    load_omap(tmp,1);                 /* Reads the .om file */
    tmp->in_memory=MAP_LOADING;
    load_original_objects(tmp,flag);  /* Reads the .oo file */
    tmp->in_memory=MAP_IN_MEMORY;
    tmp->first_visit=0;
  } else {
    if(tmp->message!=NULL)
      LOG(llevError,"Last message hasn't been freed: %s\n",
              tmp->message);
    load_omap(tmp,0);              /* Read the background */
    tmp->in_memory=MAP_LOADING;
    load_temporary_objects(tmp);
    tmp->in_memory=MAP_IN_MEMORY;
    clean_tmp_map(tmp);		/* It's going to be saved anew anyway */
    tmp->tmpname = NULL;
  }
  return tmp;
}

mapstruct *ready_map_name(char *name, int flag) {
  mapstruct *tmp;
  int first_visit=0;

/* Have we been at this level before? */
  for(tmp=first_map;tmp!=NULL;tmp=tmp->next)
    if(tmp->path==name)
      break;
  if(tmp==NULL) {
    tmp=get_linked_map(name); /* No, load the mapstruct (size) */
    LOG(llevDebug,"Got linked map.\n");
    first_visit=1;
  }
  if(tmp==NULL||(tmp->in_memory!=MAP_SWAPPED))
    return tmp;
  if(first_visit) {
    LOG(llevDebug,"First visit.\n");
    load_omap(tmp,1);                 /* Reads the .om file */
    load_original_objects(tmp,flag);  /* Reads the .oo file */
  } else {
    if(tmp->message!=NULL)
      LOG(llevError,"Last message hasn't been freed: %s\n",
              tmp->message);
    load_omap(tmp,0);                 /* Read the background */
    load_temporary_objects(tmp);
  }
  return tmp;
}

#ifndef SPEED_GAME

void set_map(mapstruct *m, int x, int y, Fontindex c) {
  if(m==NULL) {
    LOG(llevError,"Error: set_map called with NULL\n");
    return;
  }
  if(out_of_map(m,x,y)) {
    LOG(llevError,"Error: set_map out of boundary: %d,%d of %d,%d\n",
            x,y,m->mapx,m->mapy);
    return;
  }
  if(!m->in_memory!=MAP_IN_MEMORY) {
    LOG(llevError,"Error: set_map called when mas was not in memory.\n");
    return;
  }
  m->map[x+m->mapx*y]=c;
}

Fontindex get_map(mapstruct *m, int x, int y) {
  if(m==NULL) {
    LOG(llevError,"Error: get_map called with NULL\n");
    return G_BLANK;
  }
  if(out_of_map(m,x,y)) {
    LOG(llevError,"Error: get_map out of boundary: %d,%d of %d,%d\n",
            x,y,m->mapx,m->mapy);
    return G_BLANK;
  }
  if(!m->in_memory!=MAP_IN_MEMORY) {
    LOG(llevError,"Error: get_map called when mas was not in memory.\n");
    return G_BLANK;
  }
  return m->map[x+m->mapx*y];
}

void set_omap(mapstruct *m, int x, int y, Fontindex c) {
  if(m==NULL) {
    LOG(llevError,"Error: set_omap called with NULL\n");
    return;
  }
  if(out_of_map(m,x,y)) {
    LOG(llevError,"Error: set_omap out of boundary: %d,%d of %d,%d\n",
            x,y,m->mapx,m->mapy);
    return;
  }
  if(!m->in_memory!=MAP_IN_MEMORY) {
    LOG(llevError,"Error: set_omap called when mas was not in memory.\n");
    return;
  }
  m->omap[x+m->mapx*y]=c;
}

Fontindex get_omap(mapstruct *m, int x, int y) {
  if(m==NULL) {
    LOG(llevError,"Error: get_omap called with NULL\n");
    return G_BLANK;
  }
  if(out_of_map(m,x,y)) {
    LOG(llevError,"Error: get_omap out of boundary: %d,%d of %d,%d\n",
            x,y,m->mapx,m->mapy);
    return G_BLANK;
  }
  if(!m->in_memory!=MAP_IN_MEMORY) {
    LOG(llevError,"Error: get_omap called when mas was not in memory.\n");
    return G_BLANK;
  }
  return m->omap[x+m->mapx*y];
}

void set_map_ob(mapstruct *m, int x, int y, object *tmp) {
  if(m==NULL) {
    LOG(llevError,"Error: set_map_ob called with NULL\n");
    return;
  }
  if(out_of_map(m,x,y)) {
    LOG(llevError,"Error: set_map_ob out of boundary: %d,%d of %d,%d\n",
            x,y,m->mapx,m->mapy);
    return;
  }
  if(!m->in_memory!=MAP_IN_MEMORY) {
    dump_object(tmp);
    LOG(llevError,"Error: set_map_ob called when mas was not in memory.\n%s\n",
        errmsg);
    return;
  }
  m->map_ob[x+m->mapx*y]=tmp;
}

object *get_map_ob(mapstruct *m, int x, int y) {
  if(m==NULL) {
    LOG(llevError,"Error: get_map_ob called with NULL\n");
    return NULL;
  }
  if(out_of_map(m,x,y)) {
    LOG(llevError,"Error: get_map_ob out of boundary: %d,%d of %d,%d\n",
            x,y,m->mapx,m->mapy);
    return NULL;
  }
  if(!m->in_memory!=MAP_IN_MEMORY) {
    LOG(llevError,"Error: get_map_ob called when mas was not in memory.\n");
    return NULL;
  }
  return m->map_ob[x+m->mapx*y];
}

#endif

void no_maps_file(char *filename) {
  LOG(llevError,"Can't open the %s file.\n",filename);
  LOG(llevError,"If you have no maps, you must either make them yourself\n");
  LOG(llevError,"with the mapeditor, or fetch them from ftp.ifi.uio.no\n");
  LOG(llevError,"in the /pub/crossfire directory.  Be sure not to fetch\n");
  LOG(llevError,"maps belonging to a later version than this version.\n");
  exit(-1);
}

int find_first_level() {
  FILE *fp;
  char filename[MAX_BUF],buf[MAX_BUF];
  int start= -1;

  sprintf(filename,"%s/%s",LibDir,MAPINDEX);
  if((fp=open_and_uncompress(filename,0))==NULL)
    no_maps_file(filename);
  while(fgets(buf,MAX_BUF,fp)!=NULL)
    if(sscanf(buf,"start %d",&start))
      break;
  close_and_delete(fp);
  if(start == -1) {
    LOG(llevError,"Warning: Your maps file lacks a first-map declaration.\n");
    start=1;
  }
  return start;
}

char *find_map_name(int level) {
  FILE *fp;
  char filename[MAX_BUF];
  char buf[MAX_BUF];
  int i;
  static char mapname[BIG_NAME];

  sprintf(filename,"%s/%s",LibDir,MAPINDEX);
  if((fp=open_and_uncompress(filename,0))==NULL)
    no_maps_file(filename);
  while(fgets(buf,MAX_BUF,fp)!=NULL) {
    if(*buf=='#')
      continue;
    if(sscanf(buf,"%d %s",&i,mapname)!=2) 
      continue;
    if(i==level) {
      close_and_delete(fp);
      return mapname;
    }
  }
  close_and_delete(fp);
  LOG(llevDebug,"Couldn't find maplevel: %d\n",level);
  return NULL;
}

void update_map_name(int level, char *name) {
  FILE *fp;
  char filename[MAX_BUF];
  char **names;
  int levels[MAXMAPS];
  char buf[MAX_BUF];
  int i,nrofmaps=0,flag=0;

  if(!strcmp("Unnamed",name)) {
    LOG(llevError,"Tried to save unnamed map.\n");
    return;
  }
  sprintf(filename,"%s/%s",LibDir,MAPINDEX);
  if((fp=open_and_uncompress(filename,1))==NULL) {
    perror("Can't open maps file");
    return;
  }
  names = (char **) malloc(sizeof(char *) * MAXMAPS);
  for(i=0;i<MAXMAPS;i++)
    names[i] = (char *) malloc(sizeof(char) * BIG_NAME);
  while(fgets(buf,MAX_BUF,fp)!=NULL) {
    if(sscanf(buf,"%d %s",&levels[nrofmaps],names[nrofmaps])!=2)
      continue;
    if(level==levels[nrofmaps]) {
      if(!strncmp(names[nrofmaps],name,BIG_NAME)) {
        LOG(llevDebug,"No name need changing.\n");
        close_and_delete(fp);
        goto leave;
      }
      strncpy(names[nrofmaps],name,BIG_NAME);
      flag=1;
    }
    nrofmaps++;
  }
  close_and_delete(fp);
  if(!flag) {
    if((fp=fopen(filename,"a"))==NULL) {
      perror("Can't append to maps file");
      goto leave;
    }
    fprintf(fp,"%d %s\n",level,name);
    close_and_delete(fp);
    goto leave;
  }
  if((fp=fopen(filename,"w"))==NULL) {
    perror("Can't write to maps file");
    goto leave;
  }
  fprintf(fp,"start %d\n",first_map_level);
  for(i=0;i<nrofmaps;i++)
    fprintf(fp,"%d %s\n",levels[i],names[i]);
  close_and_delete(fp);

  leave:
    for(i=0;i<MAXMAPS;i++)
      free(names[i]);
    free(names);
}

void set_map_reset_time(mapstruct *map) {
  if(MAP_RESET == 0)
    map->reset_time = (-1); /* Will never be reset */
  else
    map->reset_time = seconds() + MAP_RESET;
}

/*
 * find_and_load(int level) allocates space for the new map-structure, links it
 * into the list of maps and reads the .om file to allocate space to the
 * message string and fills it with the map-message.
 */

mapstruct *find_and_load(int level) {
  char *name=find_map_name(level);
  mapstruct *map=(mapstruct *) CALLOC(1,sizeof(mapstruct));
  mapstruct *mp;

  if(map==NULL)
    fatal(OUT_OF_MEMORY);
  if(name==NULL) {
    CFREE(map);
    return NULL;
  }
  for(mp=first_map;mp!=NULL&&mp->next!=NULL;mp=mp->next);
  if(mp==NULL)
    first_map=map;
  else
    mp->next=map;
  strncpy(map->name,name,BIG_NAME);
  map->tmpname=NULL;
  map->level=level,map->mapx=0,map->mapy=0,map->startx=2,map->starty=2;
  map->path=NULL;
  map->next=NULL;
  map->in_memory=MAP_SWAPPED; /* Not loaded yet */
  map->first_visit=1;
  map->timeout=0;
  map->players=0;
  map->need_refresh=0;
  map->pending = (objectlink *) NULL;
  set_map_reset_time(map);
  return map;
}

mapstruct *get_linked_map(char *path) {
  mapstruct *map=(mapstruct *) CALLOC(1,sizeof(mapstruct));
  mapstruct *mp;

  if(map==NULL)
    fatal(OUT_OF_MEMORY);
  for(mp=first_map;mp!=NULL&&mp->next!=NULL;mp=mp->next);
  if(mp==NULL)
    first_map=map;
  else
    mp->next=map;
  map->name[0]='\0';
  map->level=0,map->mapx=0,map->mapy=0,map->startx=2,map->starty=2;
  map->path=path;
  map->next=NULL;
  map->in_memory=MAP_SWAPPED;
  return map;
}

/* This routine is supposed to find out which level the players should have
 * before visiting this map.  It is used to calculate which bonuses to put
 * on magic items.
 */

int calculate_difficulty(mapstruct *m) {
  object *op;
  archetype *at;
  int x,y;
  int diff=0;
  int total_exp=0,exp_pr_sq;
  int i;

  for(x=0;x<m->mapx;x++)
    for(y=0;y<m->mapy;y++)
      for(op=get_map_ob(m,x,y);op!=NULL;op=op->above) {
        if(IS_MONSTER(op))
          total_exp+=op->stats.exp;
        if(IS_GENERATOR(op)) {
          total_exp+=op->stats.exp;
          at=type_to_archetype(GENERATE_TYPE(op));
          if(at!=NULL)
            total_exp+=at->clone.stats.exp*8;
        }
      }
#ifdef NEWCALC
  (int)exp_pr_sq=((double)1000*total_exp)/(m->mapx*m->mapy+1);
  for(i=20;i>0;i--)
    if(exp_pr_sq>level_exp(i)) {
      diff=i;
      break;
    }
#else
  exp_pr_sq=((double)1000*total_exp)/(m->mapx*m->mapy+1);
  diff=20;
  for(i=1;i<20;i++)
    if(exp_pr_sq<=level_exp(i)) {
      diff=i;
      break;
    }
#endif
  return diff;
}

void clean_tmp_map(mapstruct *m) {
  char buf[MAX_BUF];
  if(m->tmpname == NULL)
    return;
  sprintf(buf,"%s.oo",m->tmpname);
  (void) unlink(buf);
  sprintf(buf,"%s.om",m->tmpname);
  (void) unlink(buf);
}
