/*
 * static char *rcsid_xutil_c =
 *   "$Id: xutil.c,v 1.2 1993/03/13 03:16:49 frankj Exp $";
 */

/*
    CrossFire, A Multiplayer game for X-windows

    Copyright (C) 1992 Frank Tore Johansen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author can be reached via e-mail to frankj@ifi.uio.no.
*/

#include <global.h>
#include <color.h>
#include <graphics.h>

extern void (*process_active_maps_func)();

XChar2b fontindex_to_XChar2b(Fontindex s)
{
  XChar2b c;

  c.byte1 = s/256;
  c.byte2 = s%256;

  return c;
}

/*
 * read_bmaps(): When bitmaps are used instead of fonts, this function
 * does the actual reading of all the bitmap-files, and sets up the
 * player->bitmaps array.  It assumes the bitmap directory and the
 * bmaps file are found in the LIBDIR directory.
 */

void read_bmaps(player *pl) {
  char buf[MAX_BUF], **filenames, *name, *buf2;
  FILE *fp;
  int value, nroffiles=0, *values, i, count = 0, nrofbmaps = 0;

  sprintf(buf,"%s/bmaps",LIBDIR);
  if((fp=fopen(buf,"r"))==NULL) {
    perror("Can't open bmaps file");
    exit(-1);
  }
  LOG(llevDebug,"Reading bmaps...");

  /* First count how many bitmaps we have, so we can allocate correctly */
  while(fgets(buf,MAX_BUF,fp)!=NULL)
    if(buf[0] != '#')
      nrofbmaps++;
  rewind(fp);

  filenames = (char **) malloc(sizeof(char *) * nrofbmaps);
  values = (int *) malloc(sizeof(int *) * nrofbmaps);
  nrofpixmaps = 0;

  while(fgets(buf,MAX_BUF,fp)!=NULL) {
    char *cp = buf;
    if (*cp == '#')
      continue;
    if((name = strchr(cp,'\n'))!=NULL)
      *name = '\0';
    name = strchr(cp,'\t');
    if(name == NULL) {
      LOG(llevDebug,"Warning, no whitespace: %s\n",cp);
      continue;
    }
    *name++='\0';
    if (*cp == '\\')
      value = atol(++cp);
    else
      value = (long) *cp;
    filenames[nroffiles] = strdup_local(name);
    values[nroffiles++] = value;
    if(value > nrofpixmaps)
      nrofpixmaps = value;
  }
  fclose(fp);
  nrofpixmaps++;
  LOG(llevDebug,"done (got %d/%d)\n",nrofpixmaps,nroffiles);
  if(ximages != NULL)
    return;
  pl->pixmaps = (Pixmap **) malloc(sizeof(Pixmap *) * nrofpixmaps);
  for (i=0; i<nrofpixmaps; i++)
    pl->pixmaps[i] = NULL;
  strcpy(buf,LIBDIR);
  strcat(buf,"/bitmaps/");
  buf2=buf + strlen(buf);
  LOG(llevDebug,"Building ximages...");
  for (i=0; i<nroffiles; i++) {
    unsigned int width,height;
    int hotx,hoty;	/* Dummies */
    int fault=1;
    if(pl->pixmaps[values[i]] != NULL) {
      LOG(llevError,"Warning: two entries in bmaps: %s\n",filenames[i]);
      continue;
    }
    strcpy(buf2,filenames[i]);
    pl->pixmaps[values[i]] = (Pixmap *) malloc(sizeof(Pixmap));

    switch(XReadBitmapFile(pl->gdisp,pl->win_game,
                           (_Xconst char *) buf,&width,&height,
                           pl->pixmaps[values[i]],
                           &hotx, &hoty)) {
    case BitmapOpenFailed:
      LOG(llevError,"Failed to open bitmap file %s\n",buf);
      break;
    case BitmapFileInvalid:
      LOG(llevError,"Invalid bitmap file %s\n",buf);
      break;
    case BitmapSuccess:
      fault = 0;
      break;
    default:
      LOG(llevError,"Unspecified error file %s\n",buf);
      break;
    }
    if(fault) {
      free(pl->pixmaps[values[i]]);
      pl->pixmaps[values[i]] = NULL;
      continue;
    }
    if(width != 24 || height != 24) {
      LOG(llevError,"Warning, dimmension %dx%d in %s\n",
              width,height,filenames[i]);
      free(pl->pixmaps[values[i]]);
      pl->pixmaps[values[i]] = NULL;
      continue;
    }
    if(++count > CHECK_ACTIVE_MAPS/10) {
      count = 0;
      (*process_active_maps_func)();
    }
    if(debug)
      XCopyPlane(pl->gdisp,*pl->pixmaps[values[i]],pl->win_game,pl->gc_game,
                 0,0,24,24,RANDOM()%pl->gamehint.width,
                 RANDOM()%pl->gamehint.height,1);
  }
  free(filenames);
  free(values);
  if(debug) {
    XClearWindow(pl->gdisp,pl->win_game);
    LOG(llevDebug,"done\n");
  }
  /*
   * Now fill out the unused holes with pointers to a blank pixmap
   * to avoid crashes in case trying to draw a nonexistant pixmap.
   */
  for (i = 0; i < nrofpixmaps; i++)
    if (pl->pixmaps[i] == (Pixmap *) NULL)
      pl->pixmaps[i] = pl->pixmaps[G_BLANK];
}

void set_font_path(Display *dpy, char *path) {
  char **currentList = NULL; int ncurrent = 0;
  char **directoryList = NULL; int ndirs = 0;

  currentList = XGetFontPath (dpy, &ncurrent);
  if(currentList==NULL) {
    LOG(llevError,"Unable to get old font path.\n");
    return;
  }
  {
    register char *cp = path;
    ndirs=1;
    while((cp=strchr(cp, ','))!=NULL)
      ndirs++,cp++;
    directoryList=(char **) malloc(ndirs*sizeof(char *));
    if(!directoryList) {
      LOG(llevError,"Unable to allocate memory for font path directory.\n");
      return;
    }
  }
  {
    int i=0;
    char *cp = path;
    directoryList[i++]=cp;
    while((cp=strchr(cp, ','))!=NULL)
      directoryList[i++]=cp+1,
      *cp++='\0';
    if(i!=ndirs) {
      LOG(llevError,"Internal error, only parsed %d of %d dirs.\n",i,ndirs);
      return;
    }
  }
  {
    int nnew=ndirs+ncurrent;
    char **newList = (char **) malloc (nnew * sizeof(char *));

    if(!newList) {
      LOG(llevError,"Couldn't get memory for new fontpath.\n");
      return;
    }
#if defined(SYSV) || defined(SVR4)
    memcpy((char *)directoryList,(char *)newList,
           (unsigned) (ndirs*sizeof (char *)));
    memcpy((char *) currentList, (char *) (newList + ndirs),
           (unsigned) (ncurrent*sizeof (char *)));
#else
    bcopy((char *)directoryList,(char *)newList,
          (unsigned) (ndirs*sizeof (char *)));
    bcopy((char *) currentList, (char *) (newList + ndirs),
          (unsigned) (ncurrent*sizeof (char *)));
#endif
    XSetFontPath(dpy,newList, nnew);
    free((char *)newList);
  }
  if (directoryList)
    free((char *) directoryList);
  if (currentList)
    XFreeFontPath (currentList);
}

int check_font_path(Display *dpy) {
  int count;
  char **list;

  list = XListFonts(dpy, font_graphic, 1, &count);
  LOG(llevDebug, "Matching fonts to %s: %d (%s)\n",
      font_graphic,count,count?*list:"");
  XFreeFontNames(list);
  return count;
}

int fixfontpath(player *pl) {

  if (check_font_path(pl->gdisp))
    return 0;

  if(fix_fontpath) {
      LOG(llevError,"Trying to fix fontpath for display %s.\n",pl->name);
      fflush(stderr);
      set_font_path(pl->gdisp,FontDir);
      if(check_font_path(pl->gdisp))
        return 0;
  }
  LOG(llevError,"Failed, switching to pixmaps (this might take a while).\n");
  pl->use_pixmaps = 1;
  sprintf(errmsg,"Display %s doesn't have the right fontpath.",pl->name);
  return 1;
}

/***************************************************************************
Any problems with colors e-mail  tvangod@icst.csuchico.edu.
Tyler Van Gorder, June 1, 1992
***************************************************************************/

void allocate_colors(player *p, Window w) {

  static int i, tried = 0;
  Status status;

  if(no_color) {
    p->iscolor=0;
    return;
  }
  p->iscolor = 1;
  p->depth = 0;
  p->vis = DefaultVisual(p->gdisp,p->gscreen);
  if ( p->vis->class == PseudoColor ){
    p->colormap = DefaultColormap(p->gdisp,p->gscreen);
    p->depth = DefaultDepth(p->gdisp,p->gscreen);
  }
  else{
    LOG(llevError,"Switching to black and white.\n");
    p->iscolor = 0;
  }
try_private:
  if (p->depth > 3 && p->iscolor) {
    unsigned long pixels[13];
    for (i=0; i<13; i++){
      status = XLookupColor(p->gdisp,p->colormap, colorname[i],&p->exactcolor,
                            &p->discolor[i]);
      if (!status){
        LOG(llevError,"Can't find colour.\n");
        LOG(llevError,"Switching to black and white.\n");
        p->iscolor = 0;
        break;
      }
      status = XAllocColor(p->gdisp,p->colormap,&p->discolor[i]);
      if (!status) {
        if (!tried) {
          LOG(llevError, "Not enough colours. Trying a private colourmap.\n");
          XFreeColors(p->gdisp, p->colormap, pixels, i-1, 0);
          p->colormap = XCreateColormap(p->gdisp, w, p->vis, AllocNone);
          XSetWindowColormap(p->gdisp, w, p->colormap);
          tried = 1;
          goto try_private;
        } else {
          LOG(llevError, "Failed. Switching to black and white.\n");
          p->iscolor = 0;
          break;
        }
      }
      pixels[i] = p->discolor[i].pixel;
    }
  }
  else {
     LOG(llevError,"You have a black and white terminal.\n");
     p->iscolor = 0;
  }
}
