/*
 * static char *rcsid_hiscore_c =
 *   "$Id: hiscore.c,v 1.2 1993/03/12 21:53:06 frankj Exp $";
 */

/*
    CrossFire, A Multiplayer game for X-windows

    Copyright (C) 1992 Frank Tore Johansen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author can be reached via e-mail to frankj@ifi.uio.no.
*/

#include <graphics.h>
#include <hiscore.h>
#include <sproto.h>

char *spool(char *bp,char *error) {
  if((bp=strchr(bp,':'))==NULL) {
    LOG(llevError,"Error after %s.\n",error);
    return NULL;
  }
  if(*bp=='\0') {
    LOG(llevError,"End of line.\n");
    return NULL;
  }
  if(*(++bp)=='\0') {
    LOG(llevError,"End of line++.\n");
    return NULL;
  }
  return bp;
}

void copy_score(score *sc1,score *sc2) {
  strncpy(sc2->name,sc1->name,BIG_NAME);
  strncpy(sc2->killer,sc1->killer,BIG_NAME);
  sc2->exp=sc1->exp,sc2->maplevel=sc1->maplevel,
  sc2->maxhp=sc1->maxhp,sc2->maxsp=sc1->maxsp;
}

char *put_score(score *sc) {
  static char buf[MAX_BUF];
  sprintf(buf,"%s:%d:%s:%d:%d:%d",sc->name,sc->exp,sc->killer,sc->maplevel,
          sc->maxhp,sc->maxsp);
  return buf;
}

score *get_score(char *bp) {
  static score sc;
  char *cp;

  if((cp=strchr(bp,'\n'))!=NULL)
    *cp='\0';
  strncpy(sc.name,bp,BIG_NAME);
  if((cp=strchr(sc.name,':'))!=NULL)
    *cp='\0';
  if((bp=spool(bp,"name"))==NULL) return NULL;
  sscanf(bp,"%d",&sc.exp);
  if((bp=spool(bp,"score"))==NULL) return NULL;
  strncpy(sc.killer,bp,BIG_NAME);
  if((cp=strchr(sc.killer,':'))!=NULL)
    *cp='\0';
  else
    LOG(llevError,"Couldn't find : in killer.\n");
  if((bp=spool(bp,"killer"))==NULL) return NULL;
  sscanf(bp,"%d",&sc.maplevel);
  if((bp=spool(bp,"maplevel"))==NULL) return NULL;
  sscanf(bp,"%d",&sc.maxhp);
  if((bp=spool(bp,"maxhp"))==NULL) return NULL;
  sscanf(bp,"%d",&sc.maxsp);
  return &sc;
}

score *add_score(score *new_score) {
  FILE *fp;
  static score old_score;
  score *tmp_score,pscore[HIGHSCORE_LENGTH];
  char buf[MAX_BUF], filename[MAX_BUF], *bp;
  int nrofscores=0,flag=0,i;
 
  new_score->position=HIGHSCORE_LENGTH+1;
  old_score.position= -1;
  sprintf(filename,"%s/%s",LibDir,HIGHSCORE);
  if((fp=open_and_uncompress(filename,1))!=NULL) {
    while(fgets(buf,MAX_BUF,fp)!=NULL&&nrofscores<HIGHSCORE_LENGTH) {
      if((tmp_score=get_score(buf))==NULL) break;
      if(!flag&&new_score->exp>=tmp_score->exp) {
        copy_score(new_score,&pscore[nrofscores]);
        new_score->position=nrofscores;
        flag=1;
        if(++nrofscores>=HIGHSCORE_LENGTH)
          break;
      }
      if(!strcmp(new_score->name,tmp_score->name)) { /* Another entry */
        copy_score(tmp_score,&old_score);
        old_score.position=nrofscores;
        if(flag)
          continue;
      }
      copy_score(tmp_score,&pscore[nrofscores++]);
    }
    close_and_delete(fp);
  }
  if(old_score.position!=-1&&old_score.exp>=new_score->exp)
    return &old_score; /* Did not beat old score */
  if(!flag&&nrofscores<HIGHSCORE_LENGTH)
    copy_score(new_score,&pscore[nrofscores++]);
  if((fp=fopen(filename,"w"))==NULL) {
    perror("Couldn't write to highscore-list.\n");
    return NULL;
  }
  for(i=0;i<nrofscores;i++) {
    bp=put_score(&pscore[i]);
    fprintf(fp,"%s\n",bp);
  }
  fclose(fp);
  if(flag) {
/* Eneq(@csd.uu.se): Patch to fix error in adding a new score to the
   hiscore-list */
    if(old_score.position==-1)
      return new_score;
    return &old_score;
  }
  new_score->position= -1;
  if(old_score.position!=-1)
    return &old_score;
  if(nrofscores) {
    copy_score(&pscore[nrofscores-1],&old_score);
    return &old_score;
  }
  LOG(llevError,"Highscore error.\n");
  return NULL;
}

void check_score(object *op) {
  char buf[MAX_BUF];
  score new_score;
  score *old_score;
  if(op->stats.exp==0)
    return;
  if(!op->contr->name_changed) {
    if(op->stats.exp>0) {
      draw_info(op,"As you haven't changed your name, you won't");
      draw_info(op,"get into the high-score list.");
    }
    return;
  }
  if(WAS_WIZ(op)) {
    draw_info(op,"Since you have been in wizard mode,");
    draw_info(op,"you can't enter the high-score list.");
    return;
  }
  sprintf(buf,"%s the %s",op->name,op->contr->title);
  strncpy(new_score.name,buf,BIG_NAME);
  strncpy(new_score.killer,op->contr->killer,BIG_NAME);
  if(new_score.killer[0]=='\0')
    strcpy(new_score.killer,"a dungeon collaps");
  new_score.exp=op->stats.exp;
  if(op->map == NULL)
    new_score.maplevel = 0;
  else
    new_score.maplevel=op->map->level,
  new_score.maxhp=(int) op->stats.maxhp, new_score.maxsp=(int) op->stats.maxsp;
  if((old_score=add_score(&new_score))==NULL) {
    draw_info(op,"Error in the highscore list.");
    return;
  }
  if(new_score.position == -1) {
    new_score.position = HIGHSCORE_LENGTH+1; /* Not strictly correct... */
    if(!strcmp(old_score->name,new_score.name))
      draw_info(op,"You didn't beat your last highscore:");
    else
      draw_info(op,"You didn't enter the highscore list:");
    draw_one_high_score(old_score);
    draw_info(op,errmsg);
    draw_one_high_score(&new_score);
    draw_info(op,errmsg);
    return;
  }
  if(old_score->exp>=new_score.exp)
    draw_info(op,"You didn't beat your last score:");
  else
    draw_info(op,"You beat your last score:");
  draw_one_high_score(old_score);
  draw_info(op,errmsg);
  draw_one_high_score(&new_score);
  draw_info(op,errmsg);
}

void draw_one_high_score(score *sc) {
  if(!strncmp(sc->killer,"quit",MAX_NAME))
    sprintf(errmsg,"%3d %8d %s quit the game on map %d [%d][%d].",
            sc->position,sc->exp,sc->name,sc->maplevel,sc->maxhp,sc->maxsp);
  else if(!strncmp(sc->killer,"left",MAX_NAME))
    sprintf(errmsg,"%3d %8d %s left the game on map %d [%d][%d].",
            sc->position,sc->exp,sc->name,sc->maplevel,sc->maxhp,sc->maxsp);
  else
    sprintf(errmsg,"%3d %8d %s was killed by %s on map %d [%d][%d].",
            sc->position,sc->exp,sc->name,sc->killer,sc->maplevel,
            sc->maxhp,sc->maxsp);
}


void display_high_score(object *op,int max) {
  FILE *fp;
  char buf[MAX_BUF],buf2[MAX_BUF], filename[MAX_BUF], *bp;
  int i=0,j=0,maxchar=(op==NULL?80:INFOCHARS);
  score *sc;

  sprintf(filename,"%s/highscore",LibDir);
  if((fp=open_and_uncompress(filename,0))==NULL) {
    perror("Couldn't open highscore file");
    if(op!=NULL)
      draw_info(op,"There is no highscore file.");
    return;
  }
  sprintf(buf,"Nr  Score    Who [max hitpoints][max spellpoints]");
  if(op==NULL && active_socket == (sockets *) NULL)
    printf("%s\n",buf);
  else {
    if(op != NULL)
      clear_win_info(op);
    draw_info(op,buf);
  }
  while(fgets(buf,MAX_BUF,fp)!=NULL) {
    if(j>=HIGHSCORE_LENGTH||i>=(max-1))
      break;
    if((sc=get_score(buf))==NULL)
      break;
    i++,j++;
    sc->position=j;
    draw_one_high_score(sc);
    strncpy(buf,errmsg,MAX_BUF);
    bp=buf+maxchar;
    if(strlen(buf)>maxchar)
      while(*bp!=' '&&bp>buf)
        bp--;
    strncpy(buf2,buf,bp-buf);
    buf2[bp-buf]='\0';
    if (op == NULL && active_socket == (sockets *) NULL)
      printf("%s\n",buf2);
    else
      draw_info(op,buf2);
    if(strlen(buf)>maxchar) {
      i++;
      strcpy(buf2,"            ");
      strcat(buf2,bp);
      if(op == NULL && active_socket == (sockets *) NULL)
        printf("%s\n",buf2);
      else
        draw_info(op,buf2);
    }
  }
  close_and_delete(fp);
}
