/*
 * static char *rcsid_shop_c =
 *   "$Id: shop.c,v 1.1.1.1 1993/03/07 08:31:09 frankj Exp $";
 */

/*
    CrossFire, A Multiplayer game for X-windows

    Copyright (C) 1992 Frank Tore Johansen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author can be reached via e-mail to frankj@ifi.uio.no.

*/

#include <global.h>
#include <sproto.h>

extern spell spells[];
extern int cha_bonus[26];

int query_cost(object *tmp, object *who, int flag) {
  int i=tmp->value;
  if(tmp->magic&&(tmp->arch==NULL||!tmp->arch->clone.magic))
    if(tmp->magic>0)
      i*=(3*tmp->magic*tmp->magic*tmp->magic); /* Don't multiply artifacts. */
    else
      i/=(1-tmp->magic);
  if(flag==F_BUY)
    i*=6; /* For charisma 25, the price is now x1.5 */
  if(tmp->type==WAND)
    i=(i*tmp->stats.food)/spells[tmp->stats.sp].charges;
  if(who!=NULL)
    if(flag==F_BUY)
      i=(i*(100-cha_bonus[who->stats.Cha]))/100;
    else {
      if(i>10000)
        i=8000+isqrt(i)*20;
      i=(i*100)/(100-cha_bonus[who->stats.Cha]);
    }
  if(tmp->nrof)
    i*=tmp->nrof;
  return i;
}

char *query_cost_string(object *tmp,object *who,int flag) {
  static char buf[MAX_BUF];
  int i=query_cost(tmp,who,flag),gold=i/10,silver=i%10;

  if(!gold&&!silver)
    strcpy(buf,"nothing");
  else if(gold&&!silver)
    sprintf(buf,"%d gold coins",gold);
  else if(gold&&silver)
    sprintf(buf,"%d gold coins and %d silver coins",gold,silver);
  else
    sprintf(buf,"%d silver coins",silver);
  return buf;
}

int query_money(object *op) {
  object *platina_ob=present_in_ob(PLATINACOIN,op);
  object *gold_ob=present_in_ob(GOLDCOIN,op);
  object *silver_ob=present_in_ob(SILVERCOIN,op);
  return (platina_ob==NULL?0:platina_ob->nrof*50)+
         (gold_ob==NULL?0:gold_ob->nrof*10)+
         (silver_ob==NULL?0:silver_ob->nrof);
}

int pay_for_item(object *op,object *pl) {
  int i=query_cost(op,pl,F_BUY),paid=0;
  object *platina_ob=present_in_ob(PLATINACOIN,pl);
  object *gold_ob=present_in_ob(GOLDCOIN,pl);
  object *silver_ob=present_in_ob(SILVERCOIN,pl);
  object *tmp;

  if(!i||i>query_money(pl))
    return 0;
 
  if(silver_ob!=NULL) {
    if(silver_ob->nrof>=i) {
      tmp=get_split_ob(silver_ob,i);
      if(tmp!=NULL)
        free_object(tmp);
      return 1;
    } else if (silver_ob->nrof >= i%10) {
      paid= i%10 + ((silver_ob->nrof - i%10)/10)*10;
      tmp=get_split_ob(silver_ob, paid);
      if(tmp!=NULL)
        free_object(tmp);
    }
  }
  tmp=get_split_ob(gold_ob,(i-paid+9)/10);
  if(tmp!=NULL) {
    paid+=tmp->nrof*10;
    free_object(tmp);
  }
  if(paid>i) {
    tmp=clone_arch(SILVERCOIN);
    tmp->nrof=paid-i;
    insert_ob_in_ob(tmp,pl);
    paid=i;
  }
  if(paid<i) {
    tmp=get_split_ob(platina_ob,(i-paid+49)/50);
    if(tmp!=NULL) {
      paid+=tmp->nrof*50;
      free_object(tmp);
    }
    if(paid>i) {
      if((paid-i)%10) {
        tmp=clone_arch(SILVERCOIN);
        tmp->nrof=(paid-i)%10;
        paid+=tmp->nrof;
        insert_ob_in_ob(tmp,pl);
      }
      if((paid-i)/10) {
        tmp=clone_arch(GOLDCOIN);
        tmp->nrof=(paid-i)/10;
        paid+=tmp->nrof*10;
        insert_ob_in_ob(tmp,pl);
      }
    }
  }
  if(WAS_WIZ(op))
    SET_WAS_WIZ(tmp);
  return 1;
}

/* Eneq(@csd.uu.se): Better get_payment, descends containers looking for
   unpaid items. get_payment is now used as a link. To make it simple
   we need the player-object here. */

int get_payment2 (object *pl, object *op) {
    char buf[MAX_BUF],buf2[MAX_BUF];
    int ret=1;

    if (op!=NULL&&op->inv)
        ret = get_payment2(pl, op->inv);

    if (!ret) 
        return 0;

    if (op!=NULL&&op->below) 
        ret = get_payment2 (pl, op->below);

    if (!ret) 
        return 0;
   
    if(op!=NULL&&IS_UNPAID(op)) {
        strncpy(buf,query_cost_string(op,pl,F_BUY),MAX_BUF);
        if(!pay_for_item(op,pl)) {
#if 0
            sprintf(buf2,"You can't afford %s.",query_name(op));
#else
            int i=query_cost(op,pl,F_BUY) - query_money(pl);
            UNSET_UNPAID(op);
            sprintf(buf2,"You lack %d.%1d gold to by %s.",
                    i/10,i%10,query_name(op));
            SET_UNPAID(op);
#endif
            draw_info(pl,buf2);
            D_UNLOCK(pl);
            draw_all_inventory(pl);
            return 0;
        } else {
            UNSET_UNPAID(op);
            sprintf(buf2,"You paid %s for %s.",buf,query_name(op));
            merge_ob(op,NULL);
            draw_info(pl,buf2);
        }
    }
    return 1;
}

int get_payment(object *pl) {
  int ret;

  D_LOCK(pl);

  ret = get_payment2 (pl, pl->inv);

  D_UNLOCK(pl);
  draw_all_inventory(pl);
  return ret;
}

void sell_item(object *op, object *pl) {
  int i=query_cost(op,pl,F_SELL),silver=i%10,gold=i/10;
  char buf[MAX_BUF];
  object *tmp, *otmp2;

  if(pl==NULL||pl->type!=PLAYER) {
    LOG(llevDebug,"Object other than player tried to sell something.\n");
    return;
  }
  if(!i) {
    sprintf(buf,"We're not interested in %s.",query_name(op));
    draw_info(pl,buf);
    return;
  }

  otmp2=pl->container;
  pl->container=NULL;

  if(silver) {
    tmp=clone_arch(SILVERCOIN);
    tmp->nrof=silver;
    insert_ob_in_ob(tmp,pl);
  }
  if(gold) {
    tmp=clone_arch(GOLDCOIN);
    tmp->nrof=gold;
    insert_ob_in_ob(tmp,pl);
  }

  pl->container=otmp2;

  sprintf(buf,"You receive %s for %s.",query_cost_string(op,pl,1),
          query_name(op));
  SET_UNPAID(op);
  draw_info(pl,buf);
}
