/*
 * static char *rcsid_xio_c =
 *   "$Id: xio.c,v 1.3 1993/03/13 03:21:09 frankj Exp $";
 */

/*
    CrossFire, A Multiplayer game for X-windows

    Copyright (C) 1992 Frank Tore Johansen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author can be reached via e-mail to frankj@ifi.uio.no.
*/

#ifdef ds
#include <stddef.h>
#endif
#include <graphics.h>
#include <bitmaps.h>
#include <global.h>
#include <sproto.h>
#include <xio.h>
#include <buttons.h>

extern object *objects;
extern object *free_objects;
extern object objarray[STARTMAX];
extern Fontindex pface[NROFPFACES][4];
extern char colorname[13][16];
extern char coloralias[13][16];
extern int itemcolor[][2];

int IOerrors(Display *d) {
  player *pl;
  char buf[MAX_BUF];

  LOG(llevError,"Fatal error on display.\n");
/* Mol(mol@meryl.csd.uu.se 9211211 Temporary patch. The game locks in an 
   infinite save-loop when using the original emergency_save(2)
 */
  emergency_save(0);

  if(editor)
    exit(-1);
  for(pl=first_player;pl!=NULL;pl=pl->next)
    if(pl->gdisp==d)
      break;
  if(pl==NULL)
    return 0;
  LOG(llevError,"Player %s lost the display.\n",pl->name);
  (void) sprintf(buf,"%s lost connection.",pl->name);
  remove_lock(pl);
  if(pl->ob->map!=NULL)
    pl->ob->map->players--;
  if(pl->ob->state==0||pl->ob->state==3||pl->ob->state==5)
    remove_ob(pl->ob);
#if 0
  free_object(pl->ob);
#endif
  free_player(pl);
  if(first_player==NULL)
    exit(0);
  info_all(buf,5);
  info_flush();
  longjmp(jump_addr,1);
  return 0;
}

int Xerrors(Display *d,XErrorEvent *err) {
  char buf[MAX_BUF];
#if 0 /* This isn't fatal */
  emergency_save(0); /* Just in case */
#endif
  XGetErrorText(d,err->error_code,buf,MAX_BUF);
  LOG(llevError,"Error code %s\n",buf);
  return 0;
}

void setuperrors() {
  XSetIOErrorHandler(IOerrors);
  XSetErrorHandler(Xerrors);
}

int get_root_display(player *p, char *display) {
#ifdef HAVE_SAVE_UID
  uid_t olduid;
#endif
  char *cp;
#if 0
  XSetWindowAttributes attr;
#endif

#ifdef HAVE_SAVE_UID
  olduid = geteuid();		/* I like to run it suid :) */
  setuid(getuid());		/* Have to this to open display, */
				/* if anyone is using xauth like me. */
#endif

  p->gdisp=XOpenDisplay(display);

  if (!p->gdisp) {
    sprintf(errmsg, "Can't open display %s.", display);
    return 1;
  }

  if ((cp=XGetDefault(p->gdisp,name,"splitwindow")) != NULL)
    if (!strcmp("on",cp) || !strcmp("yes",cp))
      p->split_window = 1;
    else if (!strcmp("off",cp) || !strcmp("no",cp))
      p->split_window = 0;
  if ((cp=XGetDefault(p->gdisp,"crossfire","pixmap")) != NULL)
    if (!strcmp("on",cp) || !strcmp("yes",cp))
      p->use_pixmaps = 1;
    else if (!strcmp("off",cp) || !strcmp("no",cp))
      p->use_pixmaps = 0;


#ifdef HAVE_SAVE_UID
  setuid(olduid);		/* Back to owner of suid crossfire */
#else
  setuid(geteuid());		/* We want both ruid == euid */
#endif

/*
  XAutoRepeatOff(p->gdisp);
*/
  p->gscreen=DefaultScreen(p->gdisp);
  p->gbackground=WhitePixel(p->gdisp,p->gscreen);
  p->gforeground=BlackPixel(p->gdisp,p->gscreen);
  p->roothint.x=0,p->roothint.y=0;
  p->roothint.width=582+6+INFOCHARS*FONTWIDTH;
  p->roothint.height=482;
  p->roothint.max_width=p->roothint.min_width=p->roothint.width;
  p->roothint.max_height=p->roothint.min_height=p->roothint.height;
  p->roothint.flags=PPosition | PSize;
  if(!p->split_window) {
#if 0
    p->win_root=XCreateWindow(p->gdisp,DefaultRootWindow(p->gdisp),
	p->roothint.x,p->roothint.y,p->roothint.width,p->roothint.height,2,
	CopyFromParent,CopyFromParent,CopyFromParent,0,&attr);
#else
    p->win_root=XCreateSimpleWindow(p->gdisp,DefaultRootWindow(p->gdisp),
        p->roothint.x,p->roothint.y,p->roothint.width,p->roothint.height,2,
        p->gbackground,p->gforeground);
#endif
    allocate_colors(p, p->win_root);
    if (p->iscolor){
       p->gforeground=p->discolor[0].pixel;
       p->gbackground=p->discolor[8].pixel;
    }
    p->pixmap=XCreateBitmapFromData(p->gdisp,p->win_root,
          (_Xconst char *) crossfire_bits,
	  (unsigned int) crossfire_width, (unsigned int)crossfire_height);
    XSetStandardProperties(p->gdisp,p->win_root,name,name,
                           p->pixmap,gargv,gargc,&(p->roothint));
    p->gc_root=XCreateGC(p->gdisp,p->win_root,0,0);
    XSetForeground(p->gdisp,p->gc_root,p->gforeground);
    XSetBackground(p->gdisp,p->gc_root,p->gbackground);
  }
#ifdef BEFORE_PIXMAPS
  if(fixfontpath(p))
    return 1;
#else
  if(!p->use_pixmaps)
    (void) fixfontpath(p);
#endif
  if(!p->split_window) {
    if(!p->use_pixmaps) {
      p->font=XLoadFont(p->gdisp,font_graphic);
      XSetFont(p->gdisp,p->gc_root,p->font);
    }
    XSelectInput(p->gdisp,p->win_root,KeyPressMask|
                 KeyReleaseMask|ExposureMask|StructureNotifyMask);
    XMapRaised(p->gdisp,p->win_root);
    XNextEvent(p->gdisp,&p->gevent);
  }
  setuperrors();
  return 0;
}

int get_game_display(player *p,char *display) {
  char *cp;
   
  p->gamehint.x=305,p->gamehint.y=104;
  p->gamehint.width=269,p->gamehint.height=269;
  p->gamehint.max_width=p->gamehint.min_width=p->gamehint.width;
  p->gamehint.max_height=p->gamehint.min_height=p->gamehint.height;
  p->gamehint.flags=PPosition | PSize;
  p->win_game=XCreateSimpleWindow(p->gdisp,
      p->split_window?DefaultRootWindow(p->gdisp):p->win_root,
      p->gamehint.x,p->gamehint.y,p->gamehint.width,p->gamehint.height,2,
      p->gbackground,p->gforeground);
  p->pixmap=XCreateBitmapFromData(p->gdisp,p->win_game,
          (_Xconst char *) crossfire_bits,
	  (unsigned int) crossfire_width, (unsigned int)crossfire_height);
  if (p->split_window) {
    allocate_colors(p, p->win_game);
    if (p->iscolor){
       p->gforeground=p->discolor[0].pixel;
       p->gbackground=p->discolor[8].pixel;
    }
  } else {
    XSetWindowColormap(p->gdisp, p->win_game, p->colormap);
  }
  XSetStandardProperties(p->gdisp,p->win_game,"Crossfire",name,
                         p->pixmap,gargv,gargc, &(p->gamehint));

  /* Eneq(@csd.uu.se): Has to load the chrfont from the resources here. */

  if((cp=XGetDefault(p->gdisp,name,"chrfont"))!=NULL||chrfont!=NULL) {
      XFontStruct *tmp;

      if (chrfont!=NULL)
        cp=chrfont;

      if ((tmp=XLoadQueryFont(p->gdisp, cp)) != NULL) {
        strcpy (p->font_str, cp);
        XFreeFont(p->gdisp, tmp);
      } else
        LOG(llevError, "Crossfire: Couldn't load font %s.\n", cp);
  }

  p->gc_game=XCreateGC(p->gdisp,p->win_game,0,0);
  if (p->iscolor){
     XSetForeground(p->gdisp,p->gc_game,p->discolor[0].pixel);
     XSetBackground(p->gdisp,p->gc_game,p->discolor[12].pixel);
  }
  else{
     XSetForeground(p->gdisp,p->gc_game,p->gforeground);
     XSetBackground(p->gdisp,p->gc_game,p->gbackground);
  }
  if(!p->use_pixmaps) {
    p->game_font=p->font=XLoadFont(p->gdisp,font_graphic);
    XSetFont(p->gdisp,p->gc_game,p->font);
  }
  XSelectInput(p->gdisp,p->win_game,
               ButtonPressMask|KeyPressMask|KeyReleaseMask|ExposureMask);
  XMapRaised(p->gdisp,p->win_game);
  return 0;
}

int get_info_display(player *p,char *name) {
  p->infohint.x=579,p->infohint.y=0;
  p->infohint.width=6+INFOCHARS*FONTWIDTH,p->infohint.height=11+p->infolines*13;
  p->infohint.min_width=100;
  p->infohint.min_height=30;
  p->infohint.flags=PPosition | PSize;
  p->win_info=XCreateSimpleWindow(p->gdisp,
      p->split_window?DefaultRootWindow(p->gdisp):p->win_root,
      p->infohint.x,p->infohint.y,p->infohint.width,p->infohint.height,2,
      p->gforeground,p->gbackground);
  XSetWindowColormap(p->gdisp, p->win_info, p->colormap);
  p->pixmap=XCreateBitmapFromData(p->gdisp,p->win_info,
          (_Xconst char *) crossfire_bits,
	  (unsigned int) crossfire_width, (unsigned int)crossfire_height);
  XSetStandardProperties(p->gdisp,p->win_info,"Crossfire - text","Crosstext",
			 p->pixmap,gargv,gargc,&(p->infohint));
  p->gc_info=XCreateGC(p->gdisp,p->win_info,0,0);
  XSetForeground(p->gdisp,p->gc_info,p->gforeground);
  XSetBackground(p->gdisp,p->gc_info,p->gbackground);
  p->font=XLoadFont(p->gdisp,font_text_info);
  XSetFont(p->gdisp,p->gc_info,p->font);
  XSelectInput(p->gdisp,p->win_info,
               ButtonPressMask|KeyPressMask|KeyReleaseMask|ExposureMask|
               StructureNotifyMask);
  XMapRaised(p->gdisp,p->win_info);
  return 0;
}

void resize_win_info(player *p,XEvent *event) {
  int chars=(event->xconfigure.width/FONTWIDTH)-1;
  int lines=(event->xconfigure.height/FONTHEIGHT);
  int i;
  char **info;
  if(chars==p->infochars&&lines==p->infolines)
    return;
  if(chars<3||lines<3)
    return;
  info=(char **) malloc(sizeof(char *) * lines);
  for(i=0;i<lines;i++) {
    info[i]= (char *) malloc(sizeof(char) * (chars+1));
    if(i<p->infolines) {
      strncpy(info[i],p->info[i],chars);
      info[i][chars]='\0';
    } else
      info[i][0]='\0';
  }
  for(i=0;i<p->infolines;i++)
    free(p->info[i]);
  free(p->info);
  if(p->infoline>=lines)
    p->infoline=lines-1;
  if(p->writing>=chars)
    p->writing=chars-1;
  p->info=info;
  p->infochars=chars;
  p->infolines=lines;
  refresh_win_info(p->ob);
}

void resize_win_inv(player *p,XEvent *event) {
  int x=event->xconfigure.width;
  int y=event->xconfigure.height;
  int lines=(y-FONTHEIGHT-8)/24;
  int chars=(x-60)/FONTWIDTH;
  if(x==p->invhint.x&&y==p->invhint.y)
    return;
  if(lines>MAX_INV_SIZE)
    lines=MAX_INV_SIZE;
  p->inv_size=lines;
  p->inv_chars=chars;
  p->invhint.width=x;
  p->invhint.height=y;
  p->barlength_inv=lines*24;
  sprintf(p->format_inv,"%%-%d.%ds%%-6s",chars-6,chars-6);
  draw_all_inventory(p->ob);
}

void resize_win_look(player *p,XEvent *event) {
  int x=event->xconfigure.width;
  int y=event->xconfigure.height;
  int lines=(y-FONTHEIGHT-8)/24;
  int chars=(x-60)/FONTWIDTH;
  if(x==p->lookhint.x&&y==p->lookhint.y)
    return;
  if(lines>MAX_LOOK_SIZE)
    lines=MAX_LOOK_SIZE;
  p->look_size=lines;
  p->look_chars=chars;
  p->lookhint.width=x;
  p->lookhint.height=y;
  p->barlength_look=lines*24;
  sprintf(p->format_look,"%%-%d.%ds%%-6s",chars-6,chars-6);
  draw_all_look(p->ob);
}

int get_message_display(player *p,char *name) { /* Misc-window */
  p->messagehint.x=303,p->messagehint.y=378;
  p->messagehint.width=273,p->messagehint.height=101;
  p->messagehint.max_width=p->messagehint.min_width=p->messagehint.width;
  p->messagehint.max_height=p->messagehint.min_height=p->messagehint.height;
  p->messagehint.flags=PPosition | PSize;
  p->win_message=XCreateSimpleWindow(p->gdisp,
      p->split_window?DefaultRootWindow(p->gdisp):p->win_root,
      p->messagehint.x,p->messagehint.y,p->messagehint.width,
      p->messagehint.height,2,p->gforeground,p->gbackground);
  XSetWindowColormap(p->gdisp, p->win_message, p->colormap);
  p->pixmap=XCreateBitmapFromData(p->gdisp,p->win_message,
          (_Xconst char *) crossfire_bits,
	  (unsigned int) crossfire_width, (unsigned int)crossfire_height);
  XSetStandardProperties(p->gdisp,p->win_message,"Crossfire - vitals",
                         "crossvitals",p->pixmap,
                         gargv,gargc,&(p->messagehint));
  p->gc_message=XCreateGC(p->gdisp,p->win_message,0,0);
  XSetForeground(p->gdisp,p->gc_message,p->gforeground);
  XSetBackground(p->gdisp,p->gc_message,p->gbackground);
  p->font=XLoadFont(p->gdisp,font_text_info);
  XSetFont(p->gdisp,p->gc_message,p->font);
  XSelectInput(p->gdisp,p->win_message,
               ButtonPressMask|KeyPressMask|KeyReleaseMask|ExposureMask);
  XMapRaised(p->gdisp,p->win_message);
  return 0;
}

int get_inv_display(player *p,char *name) {
#ifdef OBWIN
  p->inv=create_obwin(p,0,0,300,310,p->split_window,"inventory");
#else
  p->invhint.x=0,p->invhint.y=0;
  p->invhint.width=300,p->invhint.height=310;
  p->invhint.min_width=60+10*FONTWIDTH;
  p->invhint.min_height=FONTHEIGHT+8+24*2;
  p->invhint.flags=PPosition | PSize;
  p->win_inv=XCreateSimpleWindow(p->gdisp,
      p->split_window?DefaultRootWindow(p->gdisp):p->win_root,
      p->invhint.x,p->invhint.y,p->invhint.width,p->invhint.height,2,
      p->gforeground,p->gbackground);
  XSetWindowColormap(p->gdisp, p->win_inv, p->colormap);
  p->pixmap=XCreateBitmapFromData(p->gdisp,p->win_inv,
          (_Xconst char *) crossfire_bits,
	  (unsigned int) crossfire_width, (unsigned int)crossfire_height);
  XSetStandardProperties(p->gdisp,p->win_inv,"Crossfire - inventory",
                         "crossinventory",p->pixmap,gargv,gargc,
                         &(p->invhint));
  XSelectInput(p->gdisp,p->win_inv,
               ButtonPressMask|KeyPressMask|KeyReleaseMask|ExposureMask|
               StructureNotifyMask);
  XMapRaised(p->gdisp,p->win_inv);
  p->gc_inv_text=XCreateGC(p->gdisp,p->win_inv,0,0);
  p->gc_inv_icon=XCreateGC(p->gdisp,p->win_inv,0,0);
  XSetForeground(p->gdisp,p->gc_inv_text,p->gforeground);
  XSetBackground(p->gdisp,p->gc_inv_text,p->gbackground);
  XSetForeground(p->gdisp,p->gc_inv_icon,p->gforeground);
  XSetBackground(p->gdisp,p->gc_inv_icon,p->gbackground);
  p->font=XLoadFont(p->gdisp,font_inv_text);
  XSetFont(p->gdisp,p->gc_inv_text,p->font);
  if(!p->use_pixmaps) {
    p->font=XLoadFont(p->gdisp,font_graphic);
    XSetFont(p->gdisp,p->gc_inv_icon,p->font);
  }
#endif
  return 0;
}

int get_look_display(player *p,char *name) {
  p->lookhint.x=0,p->lookhint.y=313;
  p->lookhint.width=300,p->lookhint.height=166;
  p->lookhint.min_width=60+10*FONTWIDTH;
  p->lookhint.min_height=FONTHEIGHT+8+24*2;
  p->lookhint.flags=PPosition | PSize;
  p->win_look=XCreateSimpleWindow(p->gdisp,
      p->split_window?DefaultRootWindow(p->gdisp):p->win_root,
      p->lookhint.x,p->lookhint.y,p->lookhint.width,p->lookhint.height,2,
      p->gforeground,p->gbackground);
  XSetWindowColormap(p->gdisp, p->win_look, p->colormap);
  p->pixmap=XCreateBitmapFromData(p->gdisp,p->win_look,
          (_Xconst char *) crossfire_bits,
	  (unsigned int) crossfire_width, (unsigned int)crossfire_height);
  XSetStandardProperties(p->gdisp,p->win_look,"Crossfire - look",
                         "crosslook",p->pixmap,gargv,gargc,
                         &(p->lookhint));
  p->gc_look_text=XCreateGC(p->gdisp,p->win_look,0,0);
  p->gc_look_icon=XCreateGC(p->gdisp,p->win_look,0,0);
  XSetForeground(p->gdisp,p->gc_look_text,p->gforeground);
  XSetBackground(p->gdisp,p->gc_look_text,p->gbackground);
  XSetForeground(p->gdisp,p->gc_look_icon,p->gforeground);
  XSetBackground(p->gdisp,p->gc_look_icon,p->gbackground);
  p->font=XLoadFont(p->gdisp,font_inv_text);
  XSetFont(p->gdisp,p->gc_look_text,p->font);
  if(!p->use_pixmaps) {
    p->font=XLoadFont(p->gdisp,font_graphic);
    XSetFont(p->gdisp,p->gc_look_icon,p->font);
  }
  XSelectInput(p->gdisp,p->win_look,
               ButtonPressMask|KeyPressMask|KeyReleaseMask|ExposureMask|
               StructureNotifyMask);
  XMapRaised(p->gdisp,p->win_look);
  return 0;
}

int get_stats_display(player *p,char *name) {
  p->stathint.x=303,p->stathint.y=0;
  p->stathint.width=273,p->stathint.height=100;
  p->stathint.min_width=p->stathint.max_width=p->stathint.width;
  p->stathint.min_height=p->stathint.max_height=p->stathint.height;
  p->stathint.flags=PPosition | PSize;
  p->win_stats=XCreateSimpleWindow(p->gdisp,
      p->split_window?DefaultRootWindow(p->gdisp):p->win_root,
      p->stathint.x,p->stathint.y,p->stathint.width,p->stathint.height,2,
      p->gforeground,p->gbackground);
  XSetWindowColormap(p->gdisp, p->win_stats, p->colormap);
  p->pixmap=XCreateBitmapFromData(p->gdisp,p->win_stats,
          (_Xconst char *) crossfire_bits,
	  (unsigned int) crossfire_width, (unsigned int)crossfire_height);
  XSetStandardProperties(p->gdisp,p->win_stats,"Crossfire - status",
                         "crosstatus",p->pixmap,gargv,gargc,
                         &(p->stathint));
  p->gc_stats=XCreateGC(p->gdisp,p->win_stats,0,0);
  XSetForeground(p->gdisp,p->gc_stats,p->gforeground);
  XSetBackground(p->gdisp,p->gc_stats,p->gbackground);
  p->font=XLoadFont(p->gdisp,font_text_stats);
  XSetFont(p->gdisp,p->gc_stats,p->font);
  XSelectInput(p->gdisp,p->win_stats,
               KeyPressMask|KeyReleaseMask|ExposureMask);
  XMapRaised(p->gdisp,p->win_stats);
  return 0;
}

#ifdef SET_TITLE
void redraw_title(object *pl)
{
  char buf[MAX_BUF];
  if (pl->contr->own_title[0]=='\0')
    sprintf(buf,"Player: %s the %s",pl->name,pl->contr->title);
  else
    sprintf(buf,"Player: %s the %s",pl->name,pl->contr->own_title);
  strcat(buf,"                    ");
  XDrawImageString(pl->contr->gdisp,pl->contr->win_stats,
                  pl->contr->gc_stats,10,10,buf,strlen(buf));
}
#endif /* SET_TITLE */

void draw_stats(object *pl) {
  char buff[7][MAX_BUF];
  int flag[7];
  int i;
  if(pl->type!=PLAYER) {
    LOG(llevError,"Warning: draw_stats() in non-player: %s (%d)\n",
            pl->name,pl->count);
    return;
  }
  for(i=0;i<7;i++)
    flag[i]=0;
  if(pl->contr->last_value==-1) {
/* Eneq(@csd.uu.se): Let players define their own titles. */

      if (pl->contr->own_title[0]=='\0')
        sprintf(buff[0],"Player: %s the %s",pl->name,pl->contr->title);
      else
        sprintf(buff[0],"Player: %s the %s",pl->name,pl->contr->own_title);
    flag[0]=1;
  }
  if(pl->contr->last_value==-1||pl->stats.exp!=pl->contr->last_stats.exp||
     pl->contr->last_level!=pl->level) {
    sprintf(buff[1],"Score: %5d  Level: %d",pl->stats.exp,pl->level);
    flag[1]=1;
    pl->contr->last_stats.exp=pl->stats.exp;
    pl->contr->last_level=pl->level;
  }
  if(pl->contr->last_value==-1||
     pl->stats.hp!=pl->contr->last_stats.hp||
     pl->stats.maxhp!=pl->contr->last_stats.maxhp||
     pl->stats.sp!=pl->contr->last_stats.sp||
     pl->stats.maxsp!=pl->contr->last_stats.maxsp) {
    sprintf(buff[2],"Hp: %d/%d  Sp: %d/%d",
            pl->stats.hp,pl->stats.maxhp,pl->stats.sp,pl->stats.maxsp);
    flag[2]=1;
    pl->contr->last_stats.hp=pl->stats.hp;
    pl->contr->last_stats.maxhp=pl->stats.maxhp;
    pl->contr->last_stats.sp=pl->stats.sp;
    pl->contr->last_stats.maxsp=pl->stats.maxsp;
  }
  if(pl->contr->last_value==-1||
    pl->stats.Dex!=pl->contr->last_stats.Dex ||
    pl->stats.Con!=pl->contr->last_stats.Con ||
    pl->stats.Str!=pl->contr->last_stats.Str ||
    pl->stats.Int!=pl->contr->last_stats.Int ||
    pl->stats.Wis!=pl->contr->last_stats.Wis ||
    pl->stats.Cha!=pl->contr->last_stats.Cha) {
    sprintf(buff[3],"St%2d Dx%2d Co%2d In%2d Wi%2d Ch%2d",
            pl->stats.Str,pl->stats.Dex,pl->stats.Con,
            pl->stats.Int,pl->stats.Wis,pl->stats.Cha);
    flag[3]=1;
    pl->contr->last_stats.Str=pl->stats.Str;
    pl->contr->last_stats.Con=pl->stats.Con;
    pl->contr->last_stats.Dex=pl->stats.Dex;
    pl->contr->last_stats.Int=pl->stats.Int;
    pl->contr->last_stats.Wis=pl->stats.Wis;
    pl->contr->last_stats.Cha=pl->stats.Cha;
  }
  if(pl->contr->last_value==-1||
     pl->stats.wc!=pl->contr->last_stats.wc||
     pl->stats.ac!=pl->contr->last_stats.ac||
     pl->armour!=pl->contr->last_armour||
     pl->stats.dam!=pl->contr->last_stats.dam) {
     sprintf(buff[4],"Wc:%3d Dam:%3d Ac:%3d Arm:%3d",
             pl->stats.wc,pl->stats.dam,pl->stats.ac,pl->armour);
     flag[4]=1;
     pl->contr->last_stats.wc=pl->stats.wc;
     pl->contr->last_stats.ac=pl->stats.ac;
     pl->contr->last_stats.dam=pl->stats.dam;
     pl->contr->last_armour=pl->armour;
  }
  if(pl->contr->last_value==-1||pl->contr->last_speed!=pl->speed||
     pl->stats.food!=pl->contr->last_stats.food||
     pl->contr->last_weapon_sp!=pl->contr->weapon_sp) {
     flag[5]=1;
     if(pl->stats.food<100&&(pl->stats.food&4))
       sprintf(buff[5],"Speed: %3.2f (%1.2f) Food: *%d* HUNGRY!",
               pl->speed,pl->speed/pl->contr->weapon_sp,pl->stats.food);
     else
       sprintf(buff[5],"Speed: %3.2f (%1.2f)  Food: %3d",
               pl->speed,pl->speed/pl->contr->weapon_sp,pl->stats.food);
     pl->contr->last_stats.food=pl->stats.food;
     pl->contr->last_speed=pl->speed;
     pl->contr->last_weapon_sp=pl->contr->weapon_sp;
  }
  if(pl->contr->last_value==-1||pl->contr->last_shoot!=pl->contr->shoottype||
     (pl->contr->shoottype==2&&
      pl->contr->last_spell!=pl->contr->chosen_spell)) {
    flag[6]=1;
    switch(pl->contr->shoottype) {
    case 0:
      strcpy(buff[6],"Range: nothing");
      break;
    case 1:
      strcpy(buff[6],"Range: bow");
      break;
    case 2:
      sprintf(buff[6],"Range: spell (%s)",
              spells[pl->contr->chosen_spell].name);
      break;
    case 3:
      sprintf(buff[6],"Range: wand (%s)",
              spells[pl->contr->chosen_spell].name);
      break;
    case 4:
      sprintf(buff[6],"Range: scroll(%s)",
              spells[pl->contr->chosen_spell].name);
      break;
    default:
      strcpy(buff[6],"Range: illegal");
    }
    pl->contr->last_shoot=pl->contr->shoottype;
    pl->contr->last_spell=pl->contr->chosen_spell;
  }
  for(i=0;i<7;i++)
    if(flag[i]) {
      strcat(buff[i],"                     "); /* In case it became shorter */
      XDrawImageString(pl->contr->gdisp,pl->contr->win_stats,
        pl->contr->gc_stats,10,i*14+10, buff[i],strlen(buff[i]));
    }
  if(pl->contr->last_value!= -1)
    draw_message_window(pl); /* Update bars */
  pl->contr->last_value=pl->value;
}

void draw_all_info(object *pl) {
  int i;
  XClearWindow(pl->contr->gdisp,pl->contr->win_info);
  for(i=0;i<pl->contr->infolines;i++)
    XDrawImageString(pl->contr->gdisp,pl->contr->win_info,
     pl->contr->gc_info,FONTWIDTH,(i+1)*FONTHEIGHT,
     pl->contr->info[i],strlen(pl->contr->info[i]));
}

void draw_inventory(object *pl) {
  signed short items;
  int i,j;
  object *tmp;
  char buf[MAX_BUF];
#ifdef BEFORE_PIXMAP
  XChar xbuf;
#endif
  player *p=pl->contr;

  if(pl->type!=PLAYER)
    return;
  if(p->freeze_inv)
    return;
  for(tmp=pl->inv,items=0;tmp!=NULL;tmp=tmp->below)
    if(!tmp->invisible)
      items++;
  if(p->scroll_inv>(items - p->inv_size))
    p->scroll_inv=items-p->inv_size;
  if(p->scroll_inv<0)
    p->scroll_inv=0;
  if(p->last_weight!=pl->weight+pl->carrying) {
    p->last_weight=pl->weight+pl->carrying;
    sprintf(buf,"Inventory: (%s)",query_weight(pl));
    XDrawImageString(p->gdisp,p->win_inv,
                     p->gc_inv_text,8,13,buf,strlen(buf));
  }
  for(tmp=pl->inv,i=0;tmp!=NULL;tmp=tmp->below,i++) {
    if(tmp->invisible) {
      i--;
      continue;
    }
    if(i-p->scroll_inv>=p->inv_size||i<p->scroll_inv)
      continue;
    if(p->inv_face[i-p->scroll_inv]==tmp->face)
      continue;
    p->inv_face[i-p->scroll_inv]=tmp->face;
#ifdef BEFORE_PIXMAP
    xbuf=FontindexToXChar(tmp->face);
    XDRAWIMAGESTRING(p->gdisp,p->win_inv,
                     p->gc_inv_icon,4,40+24*(i-p->scroll_inv)
                     ,&xbuf,1);
#else
    draw_face(p,p->win_inv,p->gc_inv_icon,4,40+24*(i-p->scroll_inv),tmp->face);
#endif
  }
  for(tmp=pl->inv,i=0;tmp!=NULL;tmp=tmp->below,i++) {
    if(tmp->invisible) {
      i--;
      continue;
    }
    if(i-p->scroll_inv>=p->inv_size||i<p->scroll_inv)
      continue;
    sprintf(buf,p->format_inv,query_name(tmp),query_weight(tmp));
    if(!strncmp(buf,p->inv_name[i-p->scroll_inv],BIG_NAME))
      continue;
    strncpy(p->inv_name[i-p->scroll_inv],buf,BIG_NAME);
    XDrawImageString(p->gdisp,p->win_inv,
                     p->gc_inv_text,32,34+24*(i-p->scroll_inv)
                     ,buf,strlen(buf));
  }
  if((i-p->scroll_inv)<=p->nrofdrawn_inv) {
    int start_height=24+24*(i-p->scroll_look);
    int height;
    for(j=i;(j-p->scroll_inv)<=p->nrofdrawn_inv&&
        (j-p->scroll_inv)<p->inv_size;j++)
      p->inv_name[j-p->scroll_inv][0]='\0';
    if(p->inv_size<p->nrofdrawn_inv)
      height=p->inv_size*24;
    else
      height=p->nrofdrawn_inv*24;
    XClearArea(p->gdisp,p->win_inv,0,start_height,
               p->invhint.width-23,height*24,0);
#if 0 /* Lots of XDrawImageString probably isn't the best way to clear an area*/
      XDrawImageString(p->gdisp,p->win_inv,
                       p->gc_inv_text,32,34+24*(j-p->scroll_inv)
                       ,"                             ",30);
    }
#endif
#ifdef BEFORE_PIXMAP
    xbuf=FontindexToXChar(G_BLANK);
#endif
    for(j=i;(j-p->scroll_inv)<=p->nrofdrawn_inv&&
        (j-p->scroll_inv)<p->inv_size;j++) {
      p->inv_face[j-p->scroll_inv]=G_BLANK;
#ifdef BEFORE_PIXMAP
      XDRAWIMAGESTRING(p->gdisp,p->win_inv,
                       p->gc_inv_icon,4,40+24*(j-p->scroll_inv)
                       ,&xbuf,1);
#else
      draw_face(p,p->win_inv,p->gc_inv_icon,4,40+24*(j-p->scroll_inv),G_BLANK);
#endif
    }
  }
  if(i<p->inv_size)
    i=p->inv_size;
  p->nrofdrawn_inv=i;
  j=(p->inv_size*p->barlength_inv/i);
  if(p->scrollsize_inv!=j||
     p->last_scroll_inv!=p->scroll_inv)
  { /* Change the scrollbar if different */
    int offset=p->scroll_inv*p->barlength_inv/i;
    XClearArea(p->gdisp,p->win_inv,p->invhint.width-20,17+p->scrollstart_inv,
               16,p->scrollsize_inv,0);
    p->scrollsize_inv=j;
    p->scrollstart_inv=offset;
    p->last_scroll_inv=p->scroll_inv;
    XFillRectangle(p->gdisp,p->win_inv,p->gc_inv_text,
                   p->invhint.width-20,17+offset,16,p->scrollsize_inv);
  }
}

void draw_inventory_faces(object *pl) {
  int i,j,items;
  object *tmp, *inv;
#ifdef BEFORE_PIXMAP
  XChar xbuf;
#endif
  player *p=pl->contr;

  inv=pl->inv;

  if(p->freeze_inv)
    return;
  for(tmp=inv,items=0;tmp!=NULL;tmp=tmp->below)
    if(!tmp->invisible)
      items++;
  if(p->scroll_inv>items-p->inv_size)
    p->scroll_inv=items-p->inv_size;
  if(p->scroll_inv<0)
    p->scroll_inv=0;
  for(tmp=inv,i=0;tmp!=NULL;tmp=tmp->below,i++) {
    if(tmp->invisible) {
      i--;
      continue;
    }
    if(i-p->scroll_inv>=p->inv_size||i<p->scroll_inv)
      continue;
    if(p->inv_face[i-p->scroll_inv]==tmp->face)
      continue;
    p->inv_face[i-p->scroll_inv]=tmp->face;
#ifdef BEFORE_PIXMAP
    xbuf=FontindexToXChar(tmp->face);
    XDRAWIMAGESTRING(p->gdisp,p->win_inv,
                     p->gc_inv_icon,4,40+24*(i-p->scroll_inv)
                     ,&xbuf,1);
#else
    draw_face(p,p->win_inv,p->gc_inv_icon,4,40+24*(i-p->scroll_inv),tmp->face);
#endif
  }
  if((i-p->scroll_inv)<=p->nrofdrawn_inv) {
#ifdef BEFORE_PIXMAP
    xbuf=FontindexToXChar(G_BLANK);
#endif
    for(j=i;(j-p->scroll_inv)<=p->nrofdrawn_inv&&
        (j-p->scroll_inv)<p->inv_size;j++) {
      p->inv_face[j-p->scroll_inv]=G_BLANK;
#ifdef BEFORE_PIXMAP
      XDRAWIMAGESTRING(p->gdisp,p->win_inv,
                       p->gc_inv_icon,4,40+24*(j-p->scroll_inv)
                       ,&xbuf,1);
#else
      draw_face(p,p->win_inv,p->gc_inv_icon,4,40+24*(j-p->scroll_inv),G_BLANK);
#endif
    }
  }
}

void draw_all_inventory(object *pl) {
  int i;
  player *p=pl->contr;
  for(i=0;i<p->inv_size;i++)
    p->inv_face[i]=G_BLANK,p->inv_name[i][0]='\0';
  XClearWindow(p->gdisp,p->win_inv);
  XDrawRectangle(p->gdisp,p->win_inv,
                 p->gc_inv_text,p->invhint.width-22,15,20,p->barlength_inv+4);
  p->scrollsize_inv=1,p->scrollstart_inv=1,
  p->nrofdrawn_inv=0;
  p->last_weight= -1;
  p->freeze_inv=0;
  draw_inventory(pl);
}

void draw_look(object *pl) {
  signed short i,j,items;
  object *tmp,*top;
  char buf[MAX_BUF];
#ifdef BEFORE_PIXMAP
  XChar xbuf;
#endif
  player *p=pl->contr;

  if(p->freeze_look)
    return;
  if(IS_REMOVED(pl) || pl->map == NULL || pl->map->in_memory != MAP_IN_MEMORY)
    return;
  if(out_of_map(pl->map,pl->x,pl->y))
    return;


  /* Eneq(@csd.uu.se): Display container or look. */

  if (pl->container==NULL)
    for(top=get_map_ob(pl->map,pl->x,pl->y);top!=NULL&&top->above!=NULL;
        top=top->above);
  else
    top = pl->container->inv;
  for(tmp=top,items=0;(tmp!=NULL)&&!(tmp->above&&IS_FLOOR(tmp->above));tmp=tmp->below)
    if(!tmp->invisible&&tmp!=pl)
      items++;
  if(p->scroll_look>items-p->look_size)
    p->scroll_look=items-p->look_size;
  if(p->scroll_look<0)
    p->scroll_look=0;
  XSetForeground(p->gdisp,p->gc_look_text,
                 p->gforeground);
  for(tmp=top,i=0;(tmp!=NULL)&&!(tmp->above&&IS_FLOOR(tmp->above));tmp=tmp->below,i++) {
    if(tmp==pl||tmp->invisible) {
      i--;
      continue;
    }
    if(i-p->scroll_look>=p->look_size||i<p->scroll_look)
      continue;
    if(p->look_face[i-p->scroll_look]==tmp->face)
      continue;
    p->look_face[i-p->scroll_look]=tmp->face;
#ifdef BEFORE_PIXMAP
    xbuf=FontindexToXChar(tmp->face);
    XDRAWIMAGESTRING(p->gdisp,p->win_look,
                     p->gc_look_icon,
                     4,40+24*(i-p->scroll_look),&xbuf,1);
#else
    draw_face(p,p->win_look,p->gc_look_icon,4,40+24*(i-p->scroll_look),tmp->face);
#endif
  }
  for(tmp=top,i=0;(tmp!=NULL)&&!(tmp->above&&IS_FLOOR(tmp->above));tmp=tmp->below,i++) {
    if(tmp==pl||tmp->invisible) {
      i--;
      continue;
    }
    if(i-p->scroll_look>=p->look_size||i<p->scroll_look)
      continue;
    if(NO_PICK(tmp)) {
      char *cp;
      strncpy(buf,query_name(tmp),p->look_chars); /* Don't worry about screenwidth */
      buf[p->look_chars]='\0';
      cp=strchr(buf,'\0');
      if(strlen(buf)<p->look_chars)
        memset(cp,' ',p->look_chars-strlen(buf));
      buf[p->look_chars]='\0';
    } else
      sprintf(buf,p->format_look,query_name(tmp),query_weight(tmp));
    if(!strncmp(buf,p->look_name[i-p->scroll_look],BIG_NAME))
      continue;
    strncpy(p->look_name[i-p->scroll_look],buf,BIG_NAME);
    XDrawImageString(p->gdisp,p->win_look,
                     p->gc_look_text,32,34+24*(i-p->scroll_look)
                     ,buf,strlen(buf));
  }
  if((i-p->scroll_look)<=p->nrofdrawn_look) {
    int start_height=24+24*(i-p->scroll_look);
    int height;
    for(j=i;(j-p->scroll_look)<=p->nrofdrawn_look&&
        (j-p->scroll_look)<p->look_size;j++)
      p->look_name[j-p->scroll_look][0]='\0';
    if(p->look_size<p->nrofdrawn_look)
      height=p->look_size*24;
    else
      height=p->nrofdrawn_look*24;
    XClearArea(p->gdisp,p->win_look,0,start_height,
               p->lookhint.width-23,height*24,0);
#if 0 /* Lots of XDrawImageString probably isn't the best way to clear an area*/
      XDrawImageString(p->gdisp,p->win_look,
                       p->gc_look_text,32,
                       34+24*(j-p->scroll_look),
                       "                              ",30);
    }
#endif
#ifdef BEFORE_PIXMAP
    xbuf=FontindexToXChar(G_BLANK);
#endif
    for(j=i;(j-p->scroll_look)<=p->nrofdrawn_look&&
        (j-p->scroll_look)<p->look_size;j++) {
      p->look_face[j-p->scroll_look]=G_BLANK;
#ifdef BEFORE_PIXMAP
      XDRAWIMAGESTRING(p->gdisp,p->win_look,
                       p->gc_look_icon,4,
                       40+24*(j-p->scroll_look),&xbuf,1);
#else
      draw_face(p,p->win_look,p->gc_look_icon,4,40+24*(j-p->scroll_look),G_BLANK);
#endif
    }
  }
  if(i<p->look_size)
    i=p->look_size;
  p->nrofdrawn_look=i;
  j=(p->look_size*p->barlength_look/i);
  if(p->scrollsize_look!=(unsigned short) j||
     p->last_scroll_look!=p->scroll_look) {
    int offset=p->scroll_look*p->barlength_look/i;
    XClearArea(p->gdisp,p->win_look,p->lookhint.width-20,
               17+p->scrollstart_look,16,p->scrollsize_look,0);
    p->scrollsize_look=j;
    p->scrollstart_look=offset;
    p->last_scroll_look=p->scroll_look;
    XFillRectangle(p->gdisp,p->win_look,p->gc_look_text,
                   p->lookhint.width-20,17+offset,16,p->scrollsize_look);
  }
}

void draw_look_faces(object *pl) {
  int i,j,items;
  object *tmp,*top;
#ifdef BEFORE_PIXMAP
  XChar xbuf;
#endif
  player *p=pl->contr;

  if(p->freeze_look)
    return;
  if(IS_REMOVED(pl)||out_of_map(pl->map,pl->x,pl->y))
    return;
  for(top=(pl->container?pl->container->inv:get_map_ob(pl->map,pl->x,pl->y));
      top!=NULL&&top->above!=NULL;
      top=top->above);
  for(tmp=top,items=0;(tmp!=NULL)&&!(tmp->above&&IS_FLOOR(tmp->above));
      tmp=tmp->below)
    if(!tmp->invisible&&tmp!=pl)
      items++;
  if(p->scroll_look>items-p->look_size)
    p->scroll_look=items-p->look_size;
  if(p->scroll_look<0)
    p->scroll_look=0;
  for(tmp=top,i=0;(tmp!=NULL)&&!(tmp->above&&IS_FLOOR(tmp->above));tmp=tmp->below,i++) {
    if(tmp->invisible||tmp==pl) {
      i--;
      continue;
    }
    if(i-p->scroll_look>=p->look_size||i<p->scroll_look)
      continue;
    if(p->look_face[i-p->scroll_look]==tmp->face)
      continue;
    p->look_face[i-p->scroll_look]=tmp->face;
#ifdef BEFORE_PIXMAP
    xbuf=FontindexToXChar(tmp->face);
    XDRAWIMAGESTRING(p->gdisp,p->win_look,
                     p->gc_look_icon,
                     4,40+24*(i-p->scroll_look),&xbuf,1);
#else
    draw_face(p,p->win_look,p->gc_look_icon,4,40+24*(i-p->scroll_look),tmp->face);
#endif
  }
  if((i-p->scroll_look)<=p->nrofdrawn_look) {
#ifdef BEFORE_PIXMAP
    xbuf=FontindexToXChar(G_BLANK);
#endif
    for(j=i;(j-p->scroll_look)<=p->nrofdrawn_look&&
        (j-p->scroll_look)<p->look_size;j++) {
      p->look_face[j-p->scroll_look]=G_BLANK;
#ifdef BEFORE_PIXMAP
      XDRAWIMAGESTRING(p->gdisp,p->win_look,
                       p->gc_look_icon,4,
                       40+24*(j-p->scroll_look),&xbuf,1);
#else
    draw_face(p,p->win_look,p->gc_look_icon,4,40+24*(j-p->scroll_look),G_BLANK);
#endif
    }
  }
}

void draw_all_look(object *pl) {
  int i;
  player *p=pl->contr;

  for(i=0;i<p->look_size;i++)
    p->look_face[i]=G_BLANK,p->look_name[i][0]='\0';
  XClearWindow(p->gdisp,p->win_look);
  XDrawRectangle(p->gdisp,p->win_look,p->gc_look_text,
                 p->lookhint.width-22,15,20,p->barlength_look+4);
  p->scrollsize_look=1,p->scrollstart_look=1,
  p->nrofdrawn_look=0;
  if (!pl->container)
  XDrawImageString(p->gdisp,p->win_look,
                     p->gc_look_text,8,13,"You see:     ",13);
  else
    XDrawImageString(p->gdisp,p->win_look,
                     p->gc_look_text,8,13,"In container:",13);
  p->freeze_look=0;
  draw_look(pl);
}

#define MAX_BARS_MESSAGE 80

void draw_message_window(object *pl) {
  int bar;
  bar=(pl->stats.hp*MAX_BARS_MESSAGE)/pl->stats.maxhp;
  if(bar<0) bar=0;
  if(bar!=pl->contr->scrollsize_hp) {
    if(pl->contr->scrollsize_hp)
      XClearArea(pl->contr->gdisp,pl->contr->win_message,40,
                 MAX_BARS_MESSAGE+4-pl->contr->scrollsize_hp,
                 10,pl->contr->scrollsize_hp,0);
    pl->contr->scrollsize_hp=bar;
    if(pl->contr->iscolor&&bar<MAX_BARS_MESSAGE/4)
      XSetForeground(pl->contr->gdisp,pl->contr->gc_look_text,
                     pl->contr->discolor[3].pixel);
    XFillRectangle(pl->contr->gdisp,pl->contr->win_message,
                   pl->contr->gc_look_text,40,
                   MAX_BARS_MESSAGE+4-pl->contr->scrollsize_hp,
                   10,pl->contr->scrollsize_hp);
    if(pl->contr->iscolor&&bar<MAX_BARS_MESSAGE/4)
      XSetForeground(pl->contr->gdisp,pl->contr->gc_look_text,
                     pl->contr->gforeground);
  }
  bar=(pl->stats.sp*MAX_BARS_MESSAGE)/pl->stats.maxsp;
  if(bar<0) bar=0;
  if(bar!=pl->contr->scrollsize_sp) {
    if(pl->contr->scrollsize_sp)
      XClearArea(pl->contr->gdisp,pl->contr->win_message,80,
                 MAX_BARS_MESSAGE+4-pl->contr->scrollsize_sp,
                 10,pl->contr->scrollsize_sp,0);
    pl->contr->scrollsize_sp=bar;
    if(pl->contr->iscolor&&bar<MAX_BARS_MESSAGE/4)
      XSetForeground(pl->contr->gdisp,pl->contr->gc_look_text,
                     pl->contr->discolor[3].pixel);
    XFillRectangle(pl->contr->gdisp,pl->contr->win_message,
                   pl->contr->gc_look_text,80,
                   MAX_BARS_MESSAGE+4-pl->contr->scrollsize_sp,
                   10,pl->contr->scrollsize_sp);
    if(pl->contr->iscolor&&bar<MAX_BARS_MESSAGE/4)
      XSetForeground(pl->contr->gdisp,pl->contr->gc_look_text,
                     pl->contr->gforeground);
  }
  bar=(pl->stats.food*MAX_BARS_MESSAGE)/999;
  if(bar<0) bar=0;
  if(bar!=pl->contr->scrollsize_food) {
    if(pl->contr->scrollsize_food)
      XClearArea(pl->contr->gdisp,pl->contr->win_message,120,
                 MAX_BARS_MESSAGE+4-pl->contr->scrollsize_food,
                 10,pl->contr->scrollsize_food,0);
    pl->contr->scrollsize_food=bar;
    if(pl->contr->iscolor&&bar<MAX_BARS_MESSAGE/4)
      XSetForeground(pl->contr->gdisp,pl->contr->gc_look_text,
                     pl->contr->discolor[3].pixel);
    XFillRectangle(pl->contr->gdisp,pl->contr->win_message,
                   pl->contr->gc_look_text,120,
                   MAX_BARS_MESSAGE+4-pl->contr->scrollsize_food,
                   10,pl->contr->scrollsize_food);
    if(pl->contr->iscolor&&bar<MAX_BARS_MESSAGE/4)
      XSetForeground(pl->contr->gdisp,pl->contr->gc_look_text,
                     pl->contr->gforeground);
  }
}

void xwritedown(object *pl,char *txt,int x) {
  int y=13;
  for(;*txt!='\0';txt++,y+=13)
    XDrawImageString(pl->contr->gdisp,pl->contr->win_message,
                     pl->contr->gc_look_text,x,y,txt,1);
}

void draw_all_message(object *pl) {
  int i;
  XClearWindow(pl->contr->gdisp,pl->contr->win_message);
  pl->contr->scrollsize_hp=pl->contr->scrollsize_sp=
  pl->contr->scrollsize_food=0;
  xwritedown(pl,"HP",26);
  XDrawRectangle(pl->contr->gdisp,pl->contr->win_message,
                 pl->contr->gc_look_text,38,2,14,MAX_BARS_MESSAGE+4);
  xwritedown(pl,"SP",66);
  XDrawRectangle(pl->contr->gdisp,pl->contr->win_message,
                 pl->contr->gc_look_text,78,2,14,MAX_BARS_MESSAGE+4);
  xwritedown(pl,"Food",106);
  XDrawRectangle(pl->contr->gdisp,pl->contr->win_message,
                 pl->contr->gc_look_text,118,2,14,MAX_BARS_MESSAGE+4);
  for (i=0;i<NUMDIRBUTTS;i++) {
    XDrawRectangle(pl->contr->gdisp,pl->contr->win_message,
                   pl->contr->gc_look_text,dirX[i],dirY[i],
                   dirW,18);
    XDrawImageString(pl->contr->gdisp,pl->contr->win_message,
                     pl->contr->gc_look_text,dirX[i]+2,dirY[i]+14,
                     dirnames[i],2);
  }
  for (i=0;i<NUMOPBUTTS;i++) {
    XDrawImageString(pl->contr->gdisp,pl->contr->win_message,
                     pl->contr->gc_look_text,opX[i]+2,opY[i]+14,
                     opnames[i],8);
    XDrawRectangle(pl->contr->gdisp,pl->contr->win_message,
                   pl->contr->gc_look_text,opX[i],opY[i],
                   64,18);
  } 
  draw_message_window(pl);
}

void info_all(char *str, int pri) {
  char buf[MAX_BUF];
  player *pl;

  if(ann_count) {
    if(!strncmp(str,ann_last,MAX_BUF)) {
      ann_count++;
      return;
    }
    if(ann_count>1)
      sprintf(buf,"%d times %s",ann_count,ann_last);
    else
      strncpy(buf,ann_last,MAX_BUF);
    for(pl=first_player;pl!=NULL;pl=pl->next)
      if(pl->ob!=NULL && !IS_REMOVED(pl->ob) && pl->listening>=ann_pri)
        draw_info(pl->ob,buf);
    if(8 > ann_pri)
      info_all_sockets(buf);
  }
  strncpy(ann_last,str,MAX_BUF);
  ann_count=1;
  ann_pri=pri;
}

void info_map(mapstruct *map, char *str) {
  player *pl;
  for(pl = first_player; pl != NULL; pl = pl->next)
    if(pl->ob != NULL && pl->ob->map == map)
      draw_info(pl->ob,str);
}

void mess_all(char *str, int pri, int colr) {
  char buf[MAX_BUF];
  player *pl;

  if(ann_count) {
    if(!strncmp(str,ann_last,MAX_BUF)) {
      ann_count++;
      return;
    }
    if(ann_count>1)
      sprintf(buf,"%d times %s",ann_count,ann_last);
    else
      strncpy(buf,ann_last,MAX_BUF);
    for(pl=first_player;pl!=NULL;pl=pl->next)
      if(pl->listening>=ann_pri)
        print_message(pl->ob,buf,colr);
  }
  strncpy(ann_last,str,MAX_BUF);
  ann_count=1;
  ann_pri=pri;
}

void info_flush() {
  info_all("blah",999);
  ann_count=0;
}

/****************************************************************************
print_message : This routine prints out the character string in tmp on the
                info window. If in color mode, then the text will show up in
                a color specified by the variable colr. If Black and white
                mode, then it prints a line of "="s before and after the
                messsage.
****************************************************************************/

void print_message(object *op,char *tmp, int colr){
  if(op->type!=PLAYER)
    return;
  if (op->contr->iscolor){
    XSetForeground(op->contr->gdisp,op->contr->gc_info,
      op->contr->discolor[colr].pixel);
    draw_info(op,tmp);
    XSetForeground(op->contr->gdisp,op->contr->gc_info,
      op->contr->discolor[0].pixel);
  }
  else{
    draw_info(op,"==========================================");
    draw_info(op,tmp);
    draw_info(op,"==========================================");
  }
}

void draw_info(object *pl,char *str) {
  static char blanks[]="                                                      ";
  char *cp;
  if(pl==NULL) {
#ifdef SERVER
    if(active_socket != (sockets *) NULL)
      draw_socket(active_socket->fd,str);
    else
#endif
      fprintf(stderr,"%s\n",str);
    return;
  }
  if(pl->type!=PLAYER) {
    LOG(llevMonster,"%s [%d]: %s\n",pl->name,pl->count,str);
    return;
  }
  if((cp=strchr(str,'\n'))!=NULL) {
    char buf[VERY_BIG_BUF];
    strncpy(buf,str,VERY_BIG_BUF);
    cp=strchr(buf,'\n');
    *(cp++)='\0';
    draw_info(pl,buf);
    draw_info(pl,cp);
    return;
  }
  /* Lets do the word wrap for messages - MSW (master@cats.ucsc.edu) */
  if (strlen(str)>=pl->contr->infochars) {
	int i=pl->contr->infochars-1;
	char buf[VERY_BIG_BUF];

	/* i=last space (or ')' for armor.  Wrap armor, because
	otherwise, the two sets of ()() can be about half the line */
	while ((str[--i]!=' ') && (str[i]!=')') && (i!=0)) ;
	/* if i==0, string has no space.  Just let it be truncated */
	if (i!=0) {
	    int j;

	    i++;	/* want to keep the ')'.  This also keeps
			the space, but that really doesn't matter */
	    strncpy(buf, str, i);
	    buf[i]='\0';
	    draw_info(pl, buf);

	    for (j=i; j<strlen(str); j++)	/* if the wrap portion is */
		if (str[j]!=' ') break;		/* only space, don't wrap it*/
	    if ((((strlen(str)-i)!=1) || (str[i]!='.')) && (j!=strlen(str)))
		draw_info(pl, (str+i));
	    return;
	}
  }
  strncpy(pl->contr->info[pl->contr->infoline],str,pl->contr->infochars);
  XDrawImageString(pl->contr->gdisp,pl->contr->win_info,
    pl->contr->gc_info,FONTWIDTH,(pl->contr->infoline+1)*FONTHEIGHT,
    pl->contr->info[pl->contr->infoline],
    strlen(pl->contr->info[pl->contr->infoline]));
  if(++(pl->contr->infoline)>=pl->contr->infolines){
    if(pl->contr->scroll) {
      XCopyArea(pl->contr->gdisp,pl->contr->win_info,pl->contr->win_info,
	        pl->contr->gc_info,0,FONTHEIGHT,pl->contr->infochars*FONTWIDTH,
                pl->contr->infolines*FONTHEIGHT,0,0);
      pl->contr->infoline--;
    }
    else
      pl->contr->infoline=0;
  }
  strncpy(pl->contr->info[pl->contr->infoline],blanks,pl->contr->infochars);
  XDrawImageString(pl->contr->gdisp,pl->contr->win_info,
    pl->contr->gc_info,FONTWIDTH,(pl->contr->infoline+1)*FONTHEIGHT,blanks,strlen(blanks));
  if(pl->contr->writing) {
    strncpy(pl->contr->info[pl->contr->infoline],
            pl->contr->write_buf,pl->contr->infochars);
    XDrawImageString(pl->contr->gdisp,pl->contr->win_info,pl->contr->gc_info,
                     FONTWIDTH,(pl->contr->infoline+1)*FONTHEIGHT,
                     pl->contr->write_buf,pl->contr->writing);
  }
  if(strlen(str)>pl->contr->infochars)
    draw_info(pl,str+pl->contr->infochars);
}

void delete_ch(object *pl) {
  if(pl->contr->no_echo&&pl->contr->writing==1) /* Can't delete that prompt */
    return;
  pl->contr->write_buf[pl->contr->writing]='\0';
  pl->contr->info[pl->contr->infoline][pl->contr->writing]='\0';
  pl->contr->writing--;
  XDrawImageString(pl->contr->gdisp,pl->contr->win_info,pl->contr->gc_info,
    (pl->contr->writing+1)*FONTWIDTH,(pl->contr->infoline+1)*FONTHEIGHT," ",1);
}

void write_ch(object *pl,unsigned char key) {
  char *c;

  if((key<32||key>127)&&key!=8) return;
  c= &pl->contr->write_buf[pl->contr->writing];
  if(key==8||key==127) {
    delete_ch(pl);
    return;
  } else {
    if(pl->contr->writing>=pl->contr->infochars)
      return;
    pl->contr->write_buf[pl->contr->writing]=key;
    pl->contr->info[pl->contr->infoline][pl->contr->writing]=pl->contr->no_echo?'?':key;
    pl->contr->writing++;
  }
  pl->contr->write_buf[pl->contr->writing]='\0';
  if(pl->contr->no_echo)
    XDrawImageString(pl->contr->gdisp,pl->contr->win_info,pl->contr->gc_info,
      pl->contr->writing*FONTWIDTH,(pl->contr->infoline+1)*FONTHEIGHT,"?",1);
  else
    XDrawImageString(pl->contr->gdisp,pl->contr->win_info,pl->contr->gc_info,
      pl->contr->writing*FONTWIDTH,(pl->contr->infoline+1)*FONTHEIGHT,c,1);
}

void refresh_win_info(object *op) {
  draw_all_info(op);
}

void clear_win_info(object *op) {
  int i;

	/* don't clear if in scroll mode MSW (master@cats.ucsc.edu */
  if((op->type==PLAYER)  &&  (op->contr->scroll)) return;

  for(i=0;i<op->contr->infolines;i++) {
    (void) memset(op->contr->info[i],' ',op->contr->infochars);
    op->contr->info[i][op->contr->infochars]='\0';
  }
  XClearWindow(op->contr->gdisp,op->contr->win_info);
/*refresh_win_info(op); */
  op->contr->infoline=0;
}

void refresh(object *pl) {
  XClearWindow(pl->contr->gdisp,pl->contr->win_game);
  if(pl->contr->viewmap) {
    draw_map(pl);
    return;
  }
  refresh_map(pl->map);
  (void)memset(pl->contr->drawn,'\0',
               sizeof(Fontindex)*(WINRIGHT-WINLEFT+1)*(WINLOWER-WINUPPER+1));
  draw(pl);
}

#define CD_DRAW_MAP

void draw_dots(object *op) {
  player *pl;
  op->contr->viewmap^=2;
  if(op->contr->viewmap&2) {
    int i=0;
    XSetForeground(op->contr->gdisp,op->contr->gc_game,op->contr->gforeground);
    XSetBackground(op->contr->gdisp,op->contr->gc_game,op->contr->gbackground);
    for(pl=first_player;pl!=NULL;pl=pl->next) {
      if(pl->ob->map!=op->map)
        break;
#ifdef CD_DRAW_MAP
      XFillRectangle(op->contr->gdisp,op->contr->win_game,op->contr->gc_game,
                     2+op->contr->mapres*(pl->ob->x - op->contr->mapxmin),
                     2+op->contr->mapres*(pl->ob->y - op->contr->mapymin),
                     op->contr->mapres,op->contr->mapres);
      op->contr->mapdelx[i]=pl->ob->x - op->contr->mapxmin;
      op->contr->mapdely[i]=pl->ob->y - op->contr->mapymin;
      i++;
#else /* CD_DRAW_MAP */
      XFillRectangle(op->contr->gdisp,op->contr->win_game,op->contr->gc_game,
                     2+op->contr->mapres*pl->ob->x,
                     2+op->contr->mapres*pl->ob->y,
                     op->contr->mapres,op->contr->mapres);
      op->contr->mapdelx[i]=pl->ob->x,
      op->contr->mapdely[i]=pl->ob->y,
      i++;
#endif /* CD_DRAW_MAP */
    }
    op->contr->mapdelx[i]= -1;
  } else {
    int i;
    XSetForeground(op->contr->gdisp,op->contr->gc_game,op->contr->gbackground);
    XSetBackground(op->contr->gdisp,op->contr->gc_game,op->contr->gforeground);
    for(i=0;op->contr->mapdelx[i]!=-1;i++)
      XFillRectangle(op->contr->gdisp,op->contr->win_game,op->contr->gc_game,
                     2+op->contr->mapres*op->contr->mapdelx[i],
                     2+op->contr->mapres*op->contr->mapdely[i],
                     op->contr->mapres,op->contr->mapres);
    XSetForeground(op->contr->gdisp,op->contr->gc_game,op->contr->gforeground);
    XSetBackground(op->contr->gdisp,op->contr->gc_game,op->contr->gbackground);
  }
}
    

    
#ifndef CD_DRAW_MAP

void draw_map(object *pl) {
  int res,x,y;
  pl->contr->viewmap=1;
  XSetForeground(pl->contr->gdisp,pl->contr->gc_game,pl->contr->gbackground);
  XSetBackground(pl->contr->gdisp,pl->contr->gc_game,pl->contr->gforeground);
  XClearWindow(pl->contr->gdisp,pl->contr->win_game);
  res=(WINRIGHT-WINLEFT+1)*24/
      (pl->map->mapx>pl->map->mapy?pl->map->mapx:pl->map->mapy);
  pl->contr->mapres=res;
  for(x=0;x<pl->map->mapx;x++)
    for(y=0;y<pl->map->mapy;y++)
      if(!blocks_view(pl->map,x,y)&&!wall(pl->map,x,y))
        XFillRectangle(pl->contr->gdisp,pl->contr->win_game,
                       pl->contr->gc_game,2+res*x,2+res*y,res,res);
  XSetForeground(pl->contr->gdisp,pl->contr->gc_game,pl->contr->gforeground);
  XSetBackground(pl->contr->gdisp,pl->contr->gc_game,pl->contr->gbackground);
}

#else /* CD_DRAW_MAP */
void magic_mapping_mark(object *pl, char *map_mark, int strength);
void magic_mapping_mark_recursive(object *pl, char *map_mark, int px, int py);

void draw_map(object *pl) 
{
  int res,x,y;
  char *map_mark = (char *) calloc(pl->map->mapx * pl->map->mapy, 1);
  int xmin = pl->map->mapx, xmax = 0, ymin = pl->map->mapy, ymax = 0;

  magic_mapping_mark(pl, map_mark, 3);
  for(x = 0; x < pl->map->mapx; x++) {
    for(y = 0; y < pl->map->mapy; y++) {
      if (map_mark[x + pl->map->mapx * y]) {
	xmin = x < xmin ? x : xmin;
	xmax = x > xmax ? x : xmax;
	ymin = y < ymin ? y : ymin;
	ymax = y > ymax ? y : ymax;
      }
    }
  }
  xmin--;
  xmin = xmin < 0 ? 0 : xmin;
  xmax++;
  xmax = xmax > pl->map->mapx - 1 ? pl->map->mapx - 1: xmax;
  ymin--;
  ymin = ymin < 0 ? 0 : ymin;
  ymax++;
  ymax = ymax > pl->map->mapy - 1? pl->map->mapy - 1: ymax;

  pl->contr->viewmap=1;
  XSetForeground(pl->contr->gdisp,pl->contr->gc_game,pl->contr->gbackground);
  XSetBackground(pl->contr->gdisp,pl->contr->gc_game,pl->contr->gforeground);
  XClearWindow(pl->contr->gdisp,pl->contr->win_game);
  res=(WINRIGHT-WINLEFT+1)*24/
    (xmax - xmin + 1 >  ymax - ymin + 1 ? xmax - xmin + 1 : ymax - ymin + 1); 
  pl->contr->mapres=res;
  pl->contr->mapxmin = xmin;
  pl->contr->mapymin = ymin;
  for (x = xmin; x <= xmax; x++) {
    for (y = ymin; y <= ymax; y++) {
      if (map_mark[x + pl->map->mapx * y]) {
        XFillRectangle(pl->contr->gdisp,pl->contr->win_game,
                       pl->contr->gc_game, 2 + res*(x - xmin), 
		       2 + res*(y - ymin), res, res);
      }
    }
  }
  XSetForeground(pl->contr->gdisp,pl->contr->gc_game,pl->contr->gforeground);
  XSetBackground(pl->contr->gdisp,pl->contr->gc_game,pl->contr->gbackground);
  free(map_mark);
}

void magic_mapping_mark(object *pl, char *map_mark, int strength)
{
  int x, y;
  int xmin = pl->x - strength + 1 < 0 ? 0 : pl->x - strength + 1;
  int xmax = pl->x + strength - 1 > pl->map->mapx - 1 ? 
    pl->map->mapx - 1 : pl->x + strength - 1;
  int ymin = pl->y - strength + 1 < 0 ? 0 : pl->y - strength + 1;
  int ymax = pl->y + strength - 1 > pl->map->mapy - 1 ? 
    pl->map->mapy - 1 : pl->y + strength - 1;

  for (x = xmin; x <= xmax; x++) {
    for (y = ymin; y <= ymax; y++) {
      if (!blocks_view(pl->map, x, y) && !wall(pl->map, x, y)) {
	map_mark[x + pl->map->mapx * y] = 1;
	magic_mapping_mark_recursive(pl, map_mark, x, y);
      }
    }
  }
}

void magic_mapping_mark_recursive(object *pl, char *map_mark, int px, int py)
{
  int x, y, dx, dy;

  for (dx = -1; dx <= 1; dx++) {
    for (dy = -1; dy <= 1; dy++) {
      x = px + dx;
      y = py + dy;
      if (x >= 0 && x < pl->map->mapx && y >= 0 && y < pl->map->mapy &&
	  !blocks_view(pl->map, x, y) && !wall(pl->map, x, y) &&
	  map_mark[x + pl->map->mapx * y] == 0) {
	map_mark[x + pl->map->mapx * y] = 1;
	magic_mapping_mark_recursive(pl, map_mark, x, y);
      }
    }
  }
}

#endif /* CD_DRAW_MAP */      

#ifdef CHRFONT
/* Eneq(@csd.uu.se): Merged the color-routine and bw to one and moved the
   actual drawing (XDRAWIMAGESTRING) to draw_X. It now draws the map cell
   by cell. However to minimize the "hackiness" I only draw special characters
   (chrfont defined) in the first scan and build up an array of Fontindex
   and draws them afterwards. This minimizes the actual WORK between cells */

void draw_X (object *pl, int x, int y, XChar *str, int len, int flag, 
             Fontindex f) {
    static int last_f=-1, last_b=-1;

    
    if (flag)
        XSetFont(pl->contr->gdisp, pl->contr->gc_game, pl->contr->game_font);

    if(pl->contr->iscolor) {
        if(pl->contr->discolor[itemcolor[f][0]].pixel!=last_f)
            XSetForeground(pl->contr->gdisp,
                           pl->contr->gc_game,
                          (last_f=pl->contr->discolor[itemcolor[f][0]].pixel));
        if(pl->contr->discolor[itemcolor[f][1]].pixel!=last_b)
            XSetBackground(pl->contr->gdisp,
                           pl->contr->gc_game,
                          (last_b=pl->contr->discolor[itemcolor[f][1]].pixel));
    }
    XDRAWIMAGESTRING(pl->contr->gdisp,pl->contr->win_game,
                     pl->contr->gc_game,x,y,str,len);
    if(pl->contr->iscolor) {
        XSetForeground(pl->contr->gdisp,pl->contr->gc_game,
                       pl->contr->gforeground);
        XSetBackground(pl->contr->gdisp,pl->contr->gc_game,
                       pl->contr->gbackground);
    }
}

void draw(object *pl) {
  int i,j,left_buf, left_scr,right,ax,ay; 
                           /* ax and ay goes from 0 to max-size of arrays */
  XChar buff[20];
  Fontindex tmp, face=0, below_face=0;
  object *play=NULL;
  int noofchars,noofcells;
  Fontindex disp[15][15];

  if(pl->contr->viewmap) {
    draw_dots(pl);
    return;
  }

  if(pl->contr->do_los)
    update_los(pl);

/* Eneq(@csd.uu.se): Patch to allow players to have their own bitmaps to 
   represent their characters. This is defined in the resources-manager
   (crossfire*chrfont) and it MUST be present in the font-path. One new fields
   has been added to player-structure (game_font). The gamewindows GC is 
   altered depending on which character-font is to be used. */

  for (i=0; i<16; i++)
      for (j=0; j<16;j++)
          disp[i][j]=0;
  
  for(j=pl->y+WINUPPER;j<=pl->y+WINLOWER;j++) {
      left_scr= -1, left_buf=0,right=0,ay=j-pl->y-WINUPPER;

      noofchars=0;
      for(i=pl->x+WINLEFT;i<=pl->x+WINRIGHT;i++) {
          ax=i-pl->x-WINLEFT;

          if((tmp=((out_of_map(pl->map,i,j)||
                    pl->contr->blocked_los[i+5-pl->x][j+5-pl->y])?
               G_BLOCKED:get_map(pl->map,i,j)))!= 
                   pl->contr->drawn[ax][ay])  {

/* Eneq(@csd.uu.se): Get the object at the bottom of the map-cell. If it's
   non-NULL then traverse it up to the topmost object. If it is a PLAYER
   and it has a special character_font then we'll write him in that font. */

              play=NULL;
              disp[ax][ay]=pl->contr->drawn[ax][ay]=tmp;

              if (tmp!=G_BLOCKED&&!out_of_map(pl->map,i,j)) {
                  for (play=get_map_ob(pl->map,i,j);
                       play&&play->above!=NULL;
                       play=play->above);
                  for (;play&&play->invisible;play=play->below);

                  if (play) {
                      if (play&&play->type==PLAYER&&
                          play->contr->font_str[0]!='\0') {
                          face=play->face-pface[play->contr->face][0];
                          buff[0]=FontindexToXChar(face);
                          play->contr->font=XLoadFont(pl->contr->gdisp, 
                                                      play->contr->font_str);
                          XSetFont(pl->contr->gdisp,pl->contr->gc_game,
                                   play->contr->font);
                          draw_X(pl, 2+ax*24,ay*24+26,buff,1,0,tmp );
                          XUnloadFont(pl->contr->gdisp, play->contr->font);
                          pl->contr->drawn[ax][ay]=disp[ax][ay]=G_USER_DEFINED; 
                      } else pl->contr->drawn[ax][ay]=disp[ax][ay]=play->face;
                  }
              }
          }
      }
  }
  
  for (i=0; i<16;i++) {
      noofcells=0;
      for (j=0; j<16;j++) {
          if (disp[j][i]!=0&&disp[j][i]!=G_USER_DEFINED) 
              buff[noofcells++]=FontindexToXChar(disp[j][i]);
	  if (noofcells>0 &&
	      (disp[j][i]==0||disp[j][i]==G_USER_DEFINED||
	       pl->contr->iscolor||j==15)) {
              draw_X(pl, 2+24*(j-noofcells),i*24+26,
		     buff, noofcells,1,disp[j][i]);
	      noofcells=0;
          }
      }
  }

  face=pl->face;
  if (pl->below != NULL)
      below_face=pl->below->face;       
  if(pl->invisible) {
      tmp=pl->invisible<5 ? face : 
          (pl->invisible&(pl->invisible<50 ? 4 : 1) ?
           face :
           (pl->below==NULL ? G_BLANK : below_face));
      
  /* Eneq(@csd.uu.se): We'll have to check here as well. */

      buff[0]=FontindexToXChar(tmp);
      if (pl->contr->font_str[0]!='\0') {
          pl->contr->font=XLoadFont(pl->contr->gdisp,pl->contr->font_str);
          XSetFont(pl->contr->gdisp,pl->contr->gc_game,pl->contr->font);
          draw_X(pl,122,146,buff,1,0,tmp);
          XUnloadFont(pl->contr->gdisp, pl->contr->font);
      } else {
          XSetFont(pl->contr->gdisp,pl->contr->gc_game,pl->contr->game_font);
          draw_X(pl,122,146,buff,1,1,tmp);
      }
  }
}

#else /* DRAW_X */

void draw(object *pl) {
  int i,j,last_f= -1,last_b= -1,ax,ay; /* ax and ay goes from 0 to max-size of arrays */
  XChar buf;
  Fontindex f;
#ifdef PLAYER_COLORS
  object *play=NULL;
#endif /* PLAYER_COLORS */

  if(pl->map == NULL || pl->map->in_memory != MAP_IN_MEMORY)
    return;
  if(pl->contr->viewmap) {
    draw_dots(pl);
    return;
  }
  if(pl->contr->do_los)
    update_los(pl);
  if(pl->contr->use_pixmaps) {
    draw_pix(pl);
    return;
  }
  if(!pl->contr->iscolor) {
    draw_bw(pl); /* This function optimizes better */
    return;
  }
  buf= FontindexToXChar((Fontindex) 0);
  for(j=pl->y+WINUPPER;j<=pl->y+WINLOWER;j++) {
    ay=j-pl->y-WINUPPER;
    for(i=pl->x+WINLEFT;i<=pl->x+WINRIGHT;i++) {
      ax=i-pl->x-WINLEFT;
      if(((f=(out_of_map(pl->map,i,j)||
             pl->contr->blocked_los[i+5-pl->x][j+5-pl->y])?
             G_BLOCKED:get_map(pl->map,i,j)))
          !=pl->contr->drawn[ax][ay])
      {
        buf=FontindexToXChar(f);
        pl->contr->drawn[ax][ay]=f;
#ifdef PLAYER_COLORS
        play=NULL;
        if (f!=G_BLOCKED&&!out_of_map(pl->map,i,j))
        {
          for (play=get_map_ob(pl->map,i,j);
               play&&play->above!=NULL;
               play=play->above);
          for (;play&&play->invisible;play=play->below);
          if(play && play->type==PLAYER)
          {
            if(pl->contr->discolor[play->contr->player_color].pixel!=last_f)
              XSetForeground(pl->contr->gdisp,pl->contr->gc_game,
                             (last_f=
                              pl->contr->discolor[play->contr->player_color].
                              pixel));
          }
          else
          {
            if(pl->contr->discolor[itemcolor[f][0]].pixel!=last_f)
              XSetForeground(pl->contr->gdisp,pl->contr->gc_game,
                             (last_f=
                              pl->contr->discolor[itemcolor[f][0]].pixel));
          }
        }
        else
        {
          if(pl->contr->discolor[itemcolor[f][0]].pixel!=last_f)
            XSetForeground(pl->contr->gdisp,pl->contr->gc_game,
                           (last_f=
                            pl->contr->discolor[itemcolor[f][0]].pixel));
        }
#else /* PLAYER_COLORS */
        if(pl->contr->discolor[itemcolor[f][0]].pixel!=last_f)
          XSetForeground(pl->contr->gdisp,pl->contr->gc_game,
                         (last_f=pl->contr->discolor[itemcolor[f][0]].pixel));
#endif /* PLAYER_COLORS */
        if(pl->contr->discolor[itemcolor[f][1]].pixel!=last_b)
          XSetBackground(pl->contr->gdisp,pl->contr->gc_game,
                           (last_b=pl->contr->discolor[itemcolor[f][1]].pixel));
        XDRAWIMAGESTRING(pl->contr->gdisp,pl->contr->win_game,
                         pl->contr->gc_game,2+24*ax,ay*24+26,&buf,1);
      }
    }
  }
  if(pl->invisible & (pl->invisible<50 ? 4 : 1)) {
    pl->contr->drawn[5][5]=pl->face;
    buf=FontindexToXChar(pl->face);
#ifdef PLAYER_COLORS
    XSetForeground(pl->contr->gdisp,pl->contr->gc_game,
		   pl->contr->discolor[pl->contr->player_color].pixel);
#else /* PLAYER_COLORS */
    XSetForeground(pl->contr->gdisp,pl->contr->gc_game,
      pl->contr->discolor[itemcolor[pl->face][0]].pixel);
#endif /* PLAYER_COLORS */
    XSetBackground(pl->contr->gdisp,pl->contr->gc_game,
      pl->contr->discolor[itemcolor[pl->face][1]].pixel);
    XDRAWIMAGESTRING(pl->contr->gdisp,pl->contr->win_game,
                     pl->contr->gc_game,122,146,&buf,1);
    }
}

void draw_pix(object *pl) {
  int i,j,last_f= -1,last_b= -1,ax,ay; /* ax and ay goes from 0 to max-size of arrays */
  Fontindex f;

  if(pl->contr->viewmap) {
    draw_dots(pl);
    return;
  }
  if(pl->contr->do_los)
    update_los(pl);
  for(j=pl->y+WINUPPER;j<=pl->y+WINLOWER;j++) {
    ay=j-pl->y-WINUPPER;
    for(i=pl->x+WINLEFT;i<=pl->x+WINRIGHT;i++) {
      ax=i-pl->x-WINLEFT;
      if(((f=(out_of_map(pl->map,i,j)||
             pl->contr->blocked_los[i+5-pl->x][j+5-pl->y])?
             G_BLOCKED:get_map(pl->map,i,j)))
          !=pl->contr->drawn[ax][ay]) {
        pl->contr->drawn[ax][ay]=f;
        if(pl->contr->iscolor) {
          if(pl->contr->discolor[itemcolor[f][0]].pixel!=last_f)
            XSetForeground(pl->contr->gdisp,pl->contr->gc_game,
                           (last_f=pl->contr->discolor[itemcolor[f][0]].pixel));
          if(pl->contr->discolor[itemcolor[f][1]].pixel!=last_b)
            XSetBackground(pl->contr->gdisp,pl->contr->gc_game,
                           (last_b=pl->contr->discolor[itemcolor[f][1]].pixel));
        }
        XCopyPlane(pl->contr->gdisp,*pl->contr->pixmaps[f],
                   pl->contr->win_game,pl->contr->gc_game,
                   0,0,24,24,2+24*ax,2+24*ay,1);
      }
    }
  }
  if(pl->invisible & (pl->invisible < 50 ? 4 : 1)) {
    pl->contr->drawn[5][5] = pl->face;
    if(pl->contr->iscolor) {
      XSetForeground(pl->contr->gdisp,pl->contr->gc_game,
        pl->contr->discolor[itemcolor[pl->face][0]].pixel);
      XSetBackground(pl->contr->gdisp,pl->contr->gc_game,
        pl->contr->discolor[itemcolor[pl->face][1]].pixel);
    }
    XCopyPlane(pl->contr->gdisp,*pl->contr->pixmaps[pl->face],
               pl->contr->win_game,pl->contr->gc_game,0,0,24,24,122,146,1);
  }
}

void draw_bw(object *pl) {
  int i,j,left,right,ax,ay; /* ax and ay goes from 0 to max-size of arrays */
  XChar buff[20];
  Fontindex tmp;

  for(j=pl->y+WINUPPER;j<=pl->y+WINLOWER;j++) {
    left= -1,right=0,ay=j-pl->y-WINUPPER;
    for(i=pl->x+WINLEFT;i<=pl->x+WINRIGHT;i++) {
      ax=i-pl->x-WINLEFT;
      if((tmp=(out_of_map(pl->map,i,j)||
              pl->contr->blocked_los[i+5-pl->x][j+5-pl->y])?
              G_BLOCKED:get_map(pl->map,i,j))!=
	 pl->contr->drawn[ax][ay]) {
	pl->contr->drawn[ax][ay]=tmp;
	if(left== (-1)) left=ax;
	right=ax;
      }
      buff[ax] = FontindexToXChar(tmp);
    }
    if(left!=-1)
      XDRAWIMAGESTRING(pl->contr->gdisp,pl->contr->win_game,
                       pl->contr->gc_game,2+24*left,ay*24+26,buff+left,
		       right-left+1);
  }
  if(pl->invisible & (pl->invisible < 50 ? 4 : 1)) {
    buff[0]=FontindexToXChar(pl->face);
    pl->contr->drawn[5][5] = pl->face;
    XDRAWIMAGESTRING(pl->contr->gdisp,pl->contr->win_game,
		     pl->contr->gc_game,122,146,buff,1);
  }
}

#endif
