/*
 * static char *rcsid_Cnv_c =
 *   "$Id: Cnv.c,v 1.1.1.1 1993/03/07 08:31:00 frankj Exp $";
 */

/*
	CrossEdit, A Editor for CrossFire, A Multiplayer game
	for X-windows	

	Copyright (C) 1993 Jarkko Sonninen & Petri Heinila

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	The authors can be reached via e-mail to Jarkko.Sonninen@lut.fi
	or Petri.Heinila@lut.fi .
*/

#include <stdlib.h>

#include <Cnv.h>
#include <debug.h>

/**********************************************************************/

String CnvOk[] = {
    "OK",
    NULL
};

String CnvOkCancel[] = {
    "OK",
    "Cancel",
    NULL
};

String CnvOkApplyCancel[] = {
    "OK",
    "Apply",
    "Cancel",
    NULL
};

String CnvOkCancelNo[] = {
    "OK",
    "Cancel",
    "No",
    NULL
};

String CnvOkNotOk[] = {
  "OK",
  "Not OK",
  NULL
  };

/**********************************************************************/

Widget CnvGetShell (Widget child)
{
    Widget use;

    for (use = child;
      use != 0 && XtIsShell (use) == False;
      use = XtParent (use)
      )
	/* Widget cannot live without top shell */
	if (use == 0) {
	    fprintf (stderr, "getShell: %s\n", "no shell found");
	}
    return use;
}

/**********************************************************************/

void CnvCenterWidget (Widget w)
{
    Position x, y;
    Dimension width, height, border;

    /* center dialog in root window */
    XtRealizeWidget (w);
    XtVaGetValues (w, XtNwidth, &width, XtNheight, &height,
      XtNborderWidth, &border, NULL);
    x = (WidthOfScreen (XtScreen (w)) - width - border) / 2;
    y = (HeightOfScreen (XtScreen (w)) - height - border) / 2;
    XtVaSetValues (w, XtNx, x, XtNy, y, NULL);
}

/**********************************************************************/
/* CnvNotify */

static int CnvNotifySelect = 0;

void CnvNotifyCb (Widget w, XtPointer client, XtPointer call)
{
    CnvNotifySelect = (int) client;	/* set cardinal */

    XtDestroyWidget (CnvGetShell (w));
}

int CnvNotify (Widget w, char *msg, char **list)
{
    Widget popup, dialog;
    int i;

    /* creating widgets */
    popup = XtVaCreatePopupShell ("Notify", transientShellWidgetClass,
      w, NULL);
    dialog = XtVaCreateManagedWidget ("dialog", dialogWidgetClass, popup,
      XtNlabel, msg, NULL);
    for (i = 0; list[i] != NULL; i++) {
	XawDialogAddButton (dialog, list[i], CnvNotifyCb, (XtPointer) (i + 1));
    }

    if (i == 0)
	return 0;		/* error */


    /* */
    CnvCenterWidget (popup);
    XtPopup (popup, XtGrabExclusive);
    while (!CnvNotifySelect) {
	XtAppProcessEvent (XtWidgetToApplicationContext (w), XtIMXEvent);
    }

    /* return user selection */
    i = CnvNotifySelect;
    CnvNotifySelect = 0;
    return i;
}

/**********************************************************************/
/* CnvPrompt */

static int CnvPromptSelect = 0;
static int CnvPromptActionFlag = 0;
static char *CnvPromptString;
static Widget CnvPromptWidgetText;
void CnvPromptCb (Widget w, XtPointer client, XtPointer call);

void CnvPromptAction (Widget w, XEvent * event, String * argv, Cardinal * argc)
{
    CnvPromptCb (w, (XtPointer) 1, NULL);	/* send Ok to callback */
}

XtActionsRec CnvPromptActionTable[] = {
    {"CnvPromptAction", CnvPromptAction}
};

void CnvPromptCb (Widget w, XtPointer client, XtPointer call)
{
    CnvPromptSelect = (int) client;	/* set cardinal */
    XtVaGetValues (CnvPromptWidgetText, XtNstring, &CnvPromptString, NULL);

    XtDestroyWidget (CnvGetShell (w));
}

/* ans is temporary stored for a while ( till next call ) */
int CnvPrompt (Widget w, char *msg, char *sgg, char **ans, char **list)
{
    Widget popup, form, label, ok, use;
    int i;

    /* adding actions, if needed */
    if (!CnvPromptActionFlag) {
	XtAppAddActions (XtWidgetToApplicationContext (w),
	  CnvPromptActionTable, XtNumber (CnvPromptActionTable));
    }
    /* create dialog */
    popup = XtVaCreatePopupShell ("popup", transientShellWidgetClass, w,
      NULL);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, popup, NULL);
    label = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
      XtNresizable, True,
      XtNlabel, msg,
      XtNborderWidth, 0,
      NULL);
    CnvPromptWidgetText = XtVaCreateManagedWidget ("text",
      asciiTextWidgetClass, form,
      XtNresizable, True,
      XtNfromVert, label,
      XtNstring, sgg,
      XtNeditType, XawtextEdit,
      XtNtranslations, XtParseTranslationTable
      ("#override <Key>Return: CnvPromptAction() \n"),
      NULL);
    if (list[0]) {
	ok = XtVaCreateManagedWidget ("ok", commandWidgetClass, form,
	  XtNfromVert, CnvPromptWidgetText,
	  NULL);
	XtAddCallback (ok, XtNcallback, CnvPromptCb, (XtPointer) 1);
	use = ok;
	for (i = 1; list[i]; i++) {
	    use = XtVaCreateManagedWidget (list[i], commandWidgetClass, form,
	      XtNfromHoriz, use,
	      XtNfromVert, CnvPromptWidgetText,
	      NULL);
	    XtAddCallback (use, XtNcallback, CnvPromptCb, (XtPointer) (i + 1));
	}
    }
    /* resize again XtRealizeWidget(popup); XtVaGetValues(form,
       XtNwidth,&width, XtNhorizDistance,&padding, NULL);
       XtVaSetValues(label, XtNwidth,( width - 2 * padding ), NULL);
       XtVaSetValues(CnvPromptWidgetText, XtNwidth,( width - 2 * padding ),
       NULL); XawFormDoLayout(form,True); */

    CnvCenterWidget (popup);
    XtPopup (popup, XtGrabExclusive);
    while (!CnvPromptSelect) {
	XtAppProcessEvent (XtWidgetToApplicationContext (w), XtIMXEvent);
    }

    /* return user selection */
    i = CnvPromptSelect;
    CnvPromptSelect = 0;
    *ans = CnvPromptString;
    return i;
}

/**********************************************************************/

Widget CnvAddButton (Widget parent, String label, XtCallbackProc proc)
{
    Widget button;

    button = XtVaCreateManagedWidget (label, commandWidgetClass, parent, NULL);
    XtAddCallback (button, XtNcallback, proc, NULL);
    return button;
}

/**********************************************************************/
/*** CnvTags ***/
/***************/

static _CnvTagsNrof;

void CnvTagsOpen (CnvTags tags[], Widget parent)
{
    Cardinal i;
    char buf[BUFSIZ];

    _CnvTagsNrof++;
    for (i = 0; tags[i].label; i++) {
	sprintf (buf, "tags_%i_%i", _CnvTagsNrof, i);
	tags[i].w = XtVaCreateWidget (buf, boxWidgetClass, parent,
	  XtNorientation, XtorientHorizontal,
	  NULL);
	XtVaCreateManagedWidget (tags[i].label, labelWidgetClass, tags[i].w,
	  NULL);
	sprintf (buf, "tagsText_%i_%i", _CnvTagsNrof, i);

	tags[i].t = XtVaCreateManagedWidget (buf, asciiTextWidgetClass, 
	  tags[i].w,
	  XtNtype, XawAsciiString,
	  XtNeditType, XawtextEdit,
	  XtNresize, XawtextResizeBoth,
	  XtNuseStringInPlace, True,
	  XtNlength, BUFSIZ,
	  XtNstring, tags[i].value,
	  NULL);
    }
}

void CnvTagsClose (CnvTags tags[])
{
    int i;

    for (i = 0; tags[i].w; i++)
	XtDestroyWidget (tags[i].w);	/* for sure */
    _CnvTagsNrof--;
}

void CnvTagsManage (CnvTags tags[], unsigned long select)
{
    int i;

    for (i = 0; tags[i].label; i++)
	XtUnmanageChild (tags[i].w);
    for (i = 0; tags[i].label; i++) {
	if (select & (1 << i)) {
	    XtVaSetValues (tags[i].t, XtNstring, tags[i].value, NULL);
	    XtManageChild (tags[i].w);
	}
    }
}

/**********************************************************************/

/* show usage and information of application */
void CnvSyntax (Widget top, XtActionsRec * actions, Cardinal num_actions)
{
    int i;
    String name, class, path;

    /* application name & class */
    XtGetApplicationNameAndClass (XtDisplay (top), &name, &class);
    fprintf (stderr, "Application Name:\t\"%s\"\n", name);
    fprintf (stderr, "Application Class:\t\"%s\"\n", class);

    /* User-specific class resource, page 29 on Xt  */
    fprintf (stderr, "User resource file:\t");
    if (getenv ("XUSERFILESEARCHPATH")) {
	path = XtResolvePathname (XtDisplay (top),
	  NULL,
	  NULL,
	  NULL,
	  getenv ("XUSERFILESEARCHPATH"),
	  (Substitution) NULL, 0,
	  NULL
	  );
	if (path)
	    fprintf (stderr, "%s\n", path);
	else
	    fprintf (stderr, "(none)\n");
    } else {
	fprintf (stderr, "(not used)\n");
    }

    /* Class-specific class resource, page 30 on Xt  */
    fprintf (stderr, "Class resource file:\t");
    path = XtResolvePathname (XtDisplay (top),
      "app-defaults",		/* type */
      NULL,
      NULL,			/* suffix */
      NULL,			/* path, default XFILESEARCHPATH */
      (Substitution) NULL, 0,	/* substitutions, numof */
      NULL
      );
    if (path)
	fprintf (stderr, "%s\n", path);
    else
	fprintf (stderr, "(none)\n");

    if (actions) {
	fprintf (stderr, "%i actions suplied:\n", num_actions);
	for (i = 0; i < num_actions; i++) {
	    fprintf (stderr, "\t%s()\n", actions[i].string);
	}
    }
    exit (2);
}

/**********************************************************************/

/* */
String CnvResourceFilePath (Widget w)
{
    String path = 0;

    if (getenv ("XUSERFILESEARCHPATH")) {
	path = XtResolvePathname (XtDisplay (w),
	  NULL,
	  NULL,
	  NULL,
	  getenv ("XUSERFILESEARCHPATH"),
	  NULL, 0,
	  NULL
	  );
    }
    if (path) {
	return path;
    } else {
	path = XtResolvePathname (XtDisplay (w),
	  "app-defaults",	/* type */
	  NULL,
	  NULL,			/* suffix */
	  NULL,			/* default; XFILESEARCHPATH */
	  NULL, 0,		/* substitutions, numof */
	  NULL
	  );
    }
    return path;
}

/*** End of Cnv.c ***/
