/*
 * static char *rcsid_crossedit_c =
 *   "$Id: crossedit.c,v 1.2 1993/03/09 06:32:26 frankj Exp $";
 */

/*
	Copyright (C) 1993 Jarkko Sonninen & Petri Heinila

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	The authors can be reached via e-mail to Jarkko.Sonninen@lut.fi
	or Petri.Heinila@lut.fi .
*/

#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#if !defined (NeXT)
#include <unistd.h>
#endif

#include <Cnv.h>
#include <debug.h>
#include <editor.h>
#include <crossedit.h>
#include <X11/bitmaps/xlogo11>

/**********************************************************************/
/* ACTIONS */

XtActionsRec actions[] = {
    /* Look */
    {"lookDraw", AcLookDraw},
    {"lookInsert", AcLookInsert},
    {"lookSelect", AcLookSelect},
    {"lookDelete", AcLookDelete},
    /* Arch */
    {"archDraw", AcArchDraw},
    {"archSelect", AcArchSelect},
    /* Item */
    {"itemDraw", AcItemDraw},
    /* Edit */
    {"editDraw", AcEditDraw},
    {"editInsert", AcEditInsert},
    {"editSelect", AcEditSelect},
    {"editDelete", AcEditDelete},
    {"editMove", AcEditMove},
    {"editResize", AcEditResize},
    {"editScroll", AcEditScroll},
};

/**********************************************************************/
/* GLOBALS */

App app;
EditType editType;
Pixmap mark;
CnvTags app_attr_tags[] = {
    {NULL, NULL, "Name", ""},
    {NULL, NULL, "Message", ""},
    {NULL, NULL, "X", ""},
    {NULL, NULL, "Y", ""},
    {NULL, NULL, "Path", ""},
    {NULL, NULL, "Weight", ""},
    {NULL, NULL, "Connect", ""},
    {NULL, NULL, "Hp", ""},
    {NULL, NULL, "Trigger", ""},
    {NULL, NULL, "Sacrifice", ""},
    {NULL, NULL, "Count", ""},
    {NULL, NULL, NULL, ""},
};

/**********************************************************************/
/* RESOURCES */

struct _app_resources {
    String selectMaps;
    String wallMaps;
    Dimension mapWidth;
    Dimension mapHeight;
    Font mapFont;
    Pixel mapForeground;
}   app_resources;

#define Offset(field) XtOffsetOf(struct _app_resources, field)
static XtResource resources[] = {
    {"selectMaps", "SelectMaps", XtRString, sizeof (char *),
    Offset (selectMaps), XtRString, NULL},
    {"wallMaps", "WallMaps", XtRString, sizeof (char *),
    Offset (wallMaps), XtRString, NULL},
    {"mapWidth", "MapWidth", XtRDimension, sizeof (Dimension),
    Offset (mapWidth), XtRDimension, NULL},
    {"mapHeight", "MapHeight", XtRDimension, sizeof (Dimension),
    Offset (mapHeight), XtRDimension, NULL},

    {"mapFont", "MapFont", XtRFont, sizeof (Font),
    Offset (mapFont), XtRFont, NULL},
    {"mapForeground", "MapForeGround", XtRPixel, sizeof (Pixel),
    Offset (mapForeground), XtRPixel, NULL},
};

#undef Offset

String fallback_resources[] = {
#ifndef __CEXTRACT__
#include <Crossedit.ad.h>
#endif
    NULL
};

/**********************************************************************/
void main (int argc, char **argv, char **env)
{
    XtAppContext app_con;
    Widget pane, box;
    Widget file, fileButton, edit, editButton, info, infoButton, wc, ws;
    XSetWindowAttributes wattr;
    Arg args[2];
    char *str;
    int i;
    char buf[BUFSIZ];

    app.w = XtAppInitialize (
      &app_con,
      "Crossedit",
      NULL, 0,
      &argc, argv,
      fallback_resources,
      NULL, 0
      );
    XtAppAddActions (app_con, actions, XtNumber (actions));
    XtGetApplicationResources (app.w, (XtPointer) & app_resources,
      resources, XtNumber (resources), NULL, 0);

    if (argc > 1)
	CnvSyntax (app.w, actions, XtNumber (actions));

    mark = XCreateBitmapFromData (XtDisplay (app.w),
      RootWindowOfScreen (XtScreen (app.w)),
      (char *) xlogo11_bits,
      xlogo11_width, xlogo11_height);
    /*** vertical Pane of widgets ***/
    pane = XtCreateManagedWidget ("pane", panedWidgetClass, app.w, NULL, 0);

    /*** MenuButton - box ***/
    box = XtVaCreateManagedWidget ("box", boxWidgetClass, pane,
      XtNorientation, XtorientHorizontal,
      NULL);

    /*** File -menu ***/
    file = XtVaCreatePopupShell ("file", simpleMenuWidgetClass, app.w, NULL);
    fileButton = XtVaCreateManagedWidget ("fileButton",
      menuButtonWidgetClass, box,
      XtNmenuName, "file",
      NULL);
    ws = XtVaCreateManagedWidget ("new", smeBSBObjectClass, file, NULL);
    XtAddCallback (ws, XtNcallback, CbEditOpen, NULL);
    ws = XtVaCreateManagedWidget ("open", smeBSBObjectClass, file, NULL);
    XtAddCallback (ws, XtNcallback, CbMainLoad, NULL);
    ws = XtVaCreateManagedWidget ("line", smeLineObjectClass, file, NULL);
    ws = XtVaCreateManagedWidget ("crossfire", smeBSBObjectClass, file, NULL);
    XtAddCallback (ws, XtNcallback, CbMainCrossfire, NULL);
    ws = XtVaCreateManagedWidget ("line", smeLineObjectClass, file, NULL);
    ws = XtVaCreateManagedWidget ("quit", smeBSBObjectClass, file, NULL);
    XtAddCallback (ws, XtNcallback, CbMainQuit, NULL);

    /*** Edit -menu ***/
    edit = XtVaCreatePopupShell ("edit", simpleMenuWidgetClass, app.w, NULL);
    editButton = XtVaCreateManagedWidget ("editButton",
      menuButtonWidgetClass, box,
      XtNmenuName, "edit",
      NULL);
    ws = XtVaCreateManagedWidget ("editable", smeBSBObjectClass, edit, NULL);
    XtAddCallback (ws, XtNcallback, CbMainToggleEditable, NULL);
    XtVaSetValues (ws, XtNleftBitmap, mark, NULL);
    ws = XtVaCreateManagedWidget ("monsters", smeBSBObjectClass, edit, NULL);
    XtAddCallback (ws, XtNcallback, CbMainToggleMonsters, NULL);
    XtVaCreateManagedWidget ("line", smeLineObjectClass, edit, NULL);
    XtVaSetValues (ws, XtNleftBitmap, mark, NULL);

    /* selectMaps-menu */
    str = strtok (app_resources.selectMaps, ",\n\t ");
    while (str) {
	ws = XtVaCreateManagedWidget (str, smeBSBObjectClass, edit, NULL);
	XtAddCallback (ws, XtNcallback, CbMainSelectMaps, (XtPointer) str);
	str = strtok (NULL, ",\n\t ");
    }
    XtVaCreateManagedWidget ("line", smeLineObjectClass, edit, NULL);
    /* wallMaps-menu */
    str = strtok (app_resources.wallMaps, ",\n\t ");
    while (str) {
	ws = XtVaCreateManagedWidget (str, smeBSBObjectClass, edit, NULL);
	XtAddCallback (ws, XtNcallback, CbMainWallMaps, (XtPointer) str);
	str = strtok (NULL, ",\n\t ");
    }

    /*** Info - menu ***/
    sprintf (buf, "%s/doc/help", LibDir);
    if (!access (buf, R_OK)) {
	info = XtVaCreatePopupShell ("infoMenu", simpleMenuWidgetClass,
	  app.w, NULL);
	infoButton = XtVaCreateManagedWidget ("infoButton",
	  menuButtonWidgetClass, box,
	  XtNmenuName, "infoMenu",
	  NULL);
	{
	    String newPath;
	    newPath = XtNewString (buf);
	    ws = XtVaCreateManagedWidget ("help", smeBSBObjectClass, info, NULL);
	    XtAddCallback (ws, XtNcallback, CbMainInfo, (XtPointer) newPath);
	}
    }
    /*** select ***/
    wc = XtVaCreateManagedWidget ("look", viewportWidgetClass, pane, NULL);
    app.look.w = XtVaCreateManagedWidget ("cross", widgetClass, wc, NULL);

    /*** arch ***/
    wc = XtVaCreateManagedWidget ("arch", viewportWidgetClass, pane, NULL);
    app.arch.w = XtVaCreateManagedWidget ("cross", widgetClass, wc, NULL);

    /*** item ***/
    app.item.w = XtVaCreateManagedWidget ("item", widgetClass, pane, NULL);

    EdInitialize ();
    {
	XGCValues gcval;
	gcval.foreground = app_resources.mapForeground;
	gcval.font = app_resources.mapFont;
	app.text_gc = XtGetGC (app.w, GCForeground | GCFont, &gcval);
    }

    XtSetArg (args[0], XtNwidth, app.arch.width * FONTSIZE);
    XtSetArg (args[1], XtNheight, app.arch.height * FONTSIZE);
    XtSetValues (app.arch.w, args, 2);

    /*** initialize editorlist ***/
    for (i = 0; i < MAPLIST_SIZE; i++)
	app.list[i].id = -1;

    /*** get it run ***/
    XtRealizeWidget (app.w);
    wattr.backing_store = WhenMapped;
    XChangeWindowAttributes (XtDisplay (app.w), XtWindow (app.arch.w),
      CWBackingStore, &wattr);
    XtAppMainLoop (app_con);
}


/**********************************************************************/

/* return negative, if no editor available */
int UiEditGetIndex ()
{
    int i;

    for (i = 0; app.list[i].id != -1 && i < MAPLIST_SIZE; i++)
	debug1 ("UiEditGetIndex(): map %i reserved\n", i);
    if (i >= MAPLIST_SIZE) {
	CnvNotify (app.w, "No more maps", CnvOk);
	return -1;
    } else
	return i;
}

int UiEditIndexFromWidget (Widget w)
{
    Widget w2;
    int map;

    /* find free mapslot */
    w2 = CnvGetShell (w);
    for (map = 0; app.list[map].top != w2 && map < MAPLIST_SIZE; map++);
    if (map >= MAPLIST_SIZE)
	efatal ("map without widget");

    return map;
}

/**********************************************************************/
/* select window */

/* for exposes */
void AcLookDraw (Widget w, XEvent * event, String * argv, Cardinal * argc)
{
    debug0 ("selectPosDraw\n");
    EdDrawLook ();
}

void AcLookInsert (Widget w, XEvent * event, String * argv, Cardinal * argc)
{
    int index;

    index = event->xbutton.y / FONTSIZE;
    debug1 ("selectInsert %i\n", index);

    EdInsertObject (app.look.map, app.look.x, app.look.y, index);
}

void AcLookSelect (Widget w, XEvent * e, String * argv, Cardinal * argc)
{
    int i;

    debug1 ("argc: %i\n", *argc);
    if (argv)
	for (i = 0; i < *argc; i++)
	    debug1 ("Arg: %s\n", argv[i]);

    CbAttrOpen (w, (XtPointer) e, NULL);
}

void AcLookDelete (Widget w, XEvent * event, String * argv, Cardinal * argc)
{
    int index;

    index = event->xbutton.y / FONTSIZE;
    debug1 ("selectDelete %i\n", index);

    EdDeleteObject (app.look.map, app.look.x, app.look.y, index);
}

/**********************************************************************/
/* arch - window, select object to insert from arch square  */

/* for exposes */
void AcArchDraw (Widget w, XEvent * event, String * argv, Cardinal * argc)
{
    Arg args[2];

    debug0 ("AcArchDraw()\n");
    /* force window size , dirty */
    XtSetArg (args[0], XtNwidth, app.arch.width * FONTSIZE);
    XtSetArg (args[1], XtNheight, app.arch.height * FONTSIZE);
    XtSetValues (w, args, 2);
    EdDrawArch ();
}

void AcArchSelect (Widget w, XEvent * event, String * argv, Cardinal * argc)
{
    int x, y;

    x = event->xbutton.x / FONTSIZE;	/* grid in future 17/11/92 */
    y = event->xbutton.y / FONTSIZE;	/* grid in future 17/11/92 */

    debug0 ("selectChose()\n");
    ArchChose (x, y);		/* update to item structure */

    /* DrawChose(); not yet needed */
}

/**********************************************************************/
/* item - window to show inserted archetype */

/* for exposes */
void AcItemDraw (Widget w, XEvent * event, String * argv, Cardinal * argc)
{
    debug0 ("drawChosen()\n");
    EdDrawItem ();
}

/**********************************************************************/

/* for exposes */
void AcEditDraw (Widget w, XEvent * e, String * argv, Cardinal * argc)
{
    int index;
    debug4 ("Expose %i,%i %i,%i\n",
      e->xexpose.x, e->xexpose.y, e->xexpose.width, e->xexpose.height);

    index = UiEditIndexFromWidget (w);
    DrawMap (index,
	     e->xexpose.x / app.list[index].fontsize, 
	     e->xexpose.y / app.list[index].fontsize, 
	     e->xexpose.width / app.list[index].fontsize + 2, 
	     e->xexpose.height / app.list[index].fontsize + 2);
}

/* insert object to map */
void AcEditInsert (Widget w, XEvent * event, String * argv, Cardinal * argc)
{
    int map, x, y;
    map = UiEditIndexFromWidget (w);
    x = event->xbutton.x / app.list[map].fontsize;
    y = event->xbutton.y / app.list[map].fontsize;
    debug3 ("insertPos() in map %i at %ix%i\n", map, x, y);

    EdInsertObject (map, x, y, 0);
}

/* select objectlist from map to select-window */
void AcEditSelect (Widget w, XEvent * event, String * argv, Cardinal * argc)
{
    int x, y, map;
    char buf[BUFSIZ];

    debug2 ("AcEditSelect() %ix%i\n", app.look.x, app.look.y);
    if (argv)
	debug2 ("arg: %s, argc: %i\n", argv[0], *argc);

    map = UiEditIndexFromWidget (w);
    x = event->xbutton.x / app.list[map].fontsize;
    y = event->xbutton.y / app.list[map].fontsize;

    if (argv) {
	if (!strcmp (argv[0], "archetype")) {
	    if (app.list[map].auto_choose) {
		EdSelectItem (map, x, y);
	    } else {
		app.look.x = x;
		app.look.y = y;
		app.look.map = map;
		EdDrawLook ();
		if (app.attr.w)
		    CbAttrOpen (app.list[map].w, event, NULL);
	    }
	} else if (!strcmp (argv[0], "coordinate") && app.attr.w) {
	
  	    sprintf (buf, "%s", app.list[map].mappath);
	    strcpy (app_attr_tags[I_Path].value, buf);
	    XtVaSetValues (app_attr_tags[I_Path].t, XtNstring, buf, NULL);

	    sprintf (buf, "%i", x);
	    strcpy (app_attr_tags[I_X].value, buf);
	    XtVaSetValues (app_attr_tags[I_X].t, XtNstring, buf, NULL);

	    sprintf (buf, "%i", y);
	    strcpy (app_attr_tags[I_Y].value, buf);
	    XtVaSetValues (app_attr_tags[I_Y].t, XtNstring, buf, NULL);

	} else {
	    debug1 ("AcEditSelect(): illegal param: %s\n", argv[0]);
	}
    } else {
	debug0 ("AcEditSelect(): no param\n");
    }
}

/* delete object from map */
void AcEditDelete (Widget w, XEvent * event, String * argv, Cardinal * argc)
{
    int map, x, y;
    map = UiEditIndexFromWidget (w);
    x = event->xbutton.x / app.list[map].fontsize;
    y = event->xbutton.y / app.list[map].fontsize;

    debug3 ("deletePos() in map %i at %ix%i\n", map, x, y);
    EdDeleteObject (map, x, y, 0);
}

void UiEditRefresh (int map)
{
    char buf[BUFSIZ];

    /*** info ***/
    sprintf (buf, "Path: %s", app.list[map].mappath);
    XtVaSetValues (app.list[map].iw.path,
      XtNlabel, buf, NULL);
    sprintf (buf, "Size: %ix%i", app.list[map].emap->mapx,
      app.list[map].emap->mapy);
    XtVaSetValues (app.list[map].iw.size,
      XtNlabel, buf, NULL);
    sprintf (buf, "Start: %ix%i", app.list[map].emap->startx,
      app.list[map].emap->starty);
    XtVaSetValues (app.list[map].iw.start,
      XtNlabel, buf, NULL);

    /*** toggles ***/
    XtVaSetValues (app.list[map].iw.read_only, XtNleftBitmap, app.list[map].read_only ? mark : None, NULL);
    XtVaSetValues (app.list[map].iw.overwrite, XtNleftBitmap, app.list[map].overwrite ? mark : None, NULL);
    XtVaSetValues (app.list[map].iw.auto_choose, XtNleftBitmap, app.list[map].auto_choose ? mark : None, NULL);
    XtVaSetValues (app.list[map].iw.sparse, XtNleftBitmap, app.list[map].sparse ? mark : None, NULL);
}

void UiEditSizeRefresh (int map)
{
    XtVaSetValues (app.list[map].w,
      XtNwidth, app.list[map].emap->mapx * app.list[map].fontsize,
      XtNheight, app.list[map].emap->mapy * app.list[map].fontsize,
      NULL);
}

/* set widget size to mapsize and draws it */
void UiEditShow (int map)
{
    XClearWindow (XtDisplay (app.w), windowOfMap (map));
    UiEditRefresh (map);
    Draw (map);
}

void AcEditMove (Widget w, XEvent * e, String * argv, Cardinal * argc)
{
    Widget horizontal, vertical;

    if (argv) {
	horizontal = XtNameToWidget (XtParent (w), "horizontal");
	vertical = XtNameToWidget (XtParent (w), "vertical");

	if (!strcmp (argv[0], "right")) {
	    String argv[1];
	    argv[0] = XtNewString ("Forward");
	    XtCallActionProc (horizontal, "StartScroll", e, argv, 1);
	    XtFree (argv[0]);
	    argv[0] = XtNewString ("Proportional");
	    XtCallActionProc (horizontal, "NotifyScroll", e, argv, 1);
	    XtFree (argv[0]);
	    XtCallActionProc (horizontal, "EndScroll", e, NULL, 0);
	} else if (!strcmp (argv[0], "left")) {
	    String argv[1];
	    argv[0] = XtNewString ("Backward");
	    XtCallActionProc (horizontal, "StartScroll", e, argv, 1);
	    XtFree (argv[0]);
	    argv[0] = XtNewString ("Proportional");
	    XtCallActionProc (horizontal, "NotifyScroll", e, argv, 1);
	    XtFree (argv[0]);
	    XtCallActionProc (horizontal, "EndScroll", e, NULL, 0);
	} else if (!strcmp (argv[0], "down")) {
	    String argv[1];
	    argv[0] = XtNewString ("Forward");
	    XtCallActionProc (vertical, "StartScroll", e, argv, 1);
	    XtFree (argv[0]);
	    argv[0] = XtNewString ("Proportional");
	    XtCallActionProc (vertical, "NotifyScroll", e, argv, 1);
	    XtFree (argv[0]);
	    XtCallActionProc (vertical, "EndScroll", e, NULL, 0);
	} else if (!strcmp (argv[0], "up")) {
	    String argv[1];
	    argv[0] = XtNewString ("Backward");
	    XtCallActionProc (vertical, "StartScroll", e, argv, 1);
	    XtFree (argv[0]);
	    argv[0] = XtNewString ("Proportional");
	    XtCallActionProc (vertical, "NotifyScroll", e, argv, 1);
	    XtFree (argv[0]);
	    XtCallActionProc (vertical, "EndScroll", e, NULL, 0);
	}
    }
}

void AcEditResize (Widget w, XEvent * e, String * argv, Cardinal * argc)
{
    int map;

    if (argv) {
	debug1 ("AcEditScroll() %s\n", argv[0]);
	map = UiEditIndexFromWidget (w);
	if (app.list[map].read_only)
	    return;
	if (!strcmp (argv[0], "right")) {
	    ResizeMap (map, app.list[map].emap->mapx + 1,
	      app.list[map].emap->mapy);
	} else if (!strcmp (argv[0], "left")) {
	    ResizeMap (map, app.list[map].emap->mapx - 1,
	      app.list[map].emap->mapy);
	} else if (!strcmp (argv[0], "down")) {
	    ResizeMap (map, app.list[map].emap->mapx,
	      app.list[map].emap->mapy + 1);
	} else if (!strcmp (argv[0], "up")) {
	    ResizeMap (map, app.list[map].emap->mapx,
	      app.list[map].emap->mapy - 1);
	}
	UiEditSizeRefresh (map);
	UiEditShow (map);
    }
}

void AcEditScroll (Widget w, XEvent * e, String * argv, Cardinal * argc)
{
    int map;

    if (argv) {
	debug1 ("AcEditResize(): %s\n", argv[0]);
	map = UiEditIndexFromWidget (w);
	if (app.list[map].read_only)
	    return;
	if (!strcmp (argv[0], "right")) {
	    ScrollMap (map, 1, 0);
	} else if (!strcmp (argv[0], "left")) {
	    ScrollMap (map, -1, 0);
	} else if (!strcmp (argv[0], "down")) {
	    ScrollMap (map, 0, 1);
	} else if (!strcmp (argv[0], "up")) {
	    ScrollMap (map, 0, -1);
	}
	UiEditShow (map);
    }
}


/**********************************************************************/

void CbEditOpen (Widget w, XtPointer client, XtPointer call)
{
    Widget vbox, hbox;
    XSetWindowAttributes wattr;
    int map;
    int sizex = 16, sizey = 16;
    char *path = (char *) client;
    Widget wc, ws;		/* widget-container, widget-small */

    debug1 ("CbEditOpen() for path %s\n", path);
    if ((map = UiEditGetIndex ()) < 0)
	return;
    app.list[map].id = map;

    /*** ***/
    if (path) {
	strcpy (app.list[map].mappath, path);	/*** before load !!! ***/
	switch (EdLoadMap (map, app.list[map].mappath)) {
	case 0:
	    break;
	case -1:
	    CnvNotify (app.w, "Map already in memory", CnvOk);
	    return;
	default:
	    CnvNotify (app.w, "Error in loading map", CnvOk);
	    return;
	}
	sizex = app.list[map].emap->mapx;
	sizey = app.list[map].emap->mapy;
	debug1 ("CbEditOpen(): Path: %s\n", app.list[map].mappath);
    } else {
	editType = Regular;
	EdEmptyMap (map, sizex, sizey);
    }
    app.list[map].type = editType;	/*** implicit ! ***/
    app.list[map].overwrite = 0;
    app.list[map].modified = 0;
    app.list[map].sparse = 0;
    app.list[map].fontsize = 24;
    switch (app.list[map].type) {
    case Regular:
	app.list[map].read_only = 0;
	app.list[map].auto_choose = 0;
	break;
    case Pick:
	app.list[map].read_only = 1;
	app.list[map].auto_choose = 1;
	break;
    case Wall:
	app.list[map].read_only = 1;
	app.list[map].auto_choose = 1;
	break;
    default:
	efatal ("unknown MapType");
    }

    /*** shell ***/
    app.list[map].top = XtVaCreatePopupShell ("edit",
      topLevelShellWidgetClass, app.w,
      XtNwidth,
      (app.list[map].emap->mapx > app_resources.mapWidth ?
	app_resources.mapWidth * app.list[map].fontsize :
	app.list[map].emap->mapx * app.list[map].fontsize) + 16,	/* kludge */
      XtNheight,
      (app.list[map].emap->mapy > app_resources.mapHeight ?
	app_resources.mapHeight * app.list[map].fontsize :
	app.list[map].emap->mapy * app.list[map].fontsize) + 46,	/* kludge */
      NULL);
    vbox = XtVaCreateManagedWidget ("vbox", panedWidgetClass,
      app.list[map].top, NULL);
    hbox = XtVaCreateManagedWidget ("hbox", boxWidgetClass,
      vbox, NULL);

    /*** file-menu ***/
    ws = XtVaCreateManagedWidget ("fileButton", menuButtonWidgetClass, hbox,
      XtNmenuName, "mapFile",
      NULL);
    wc = XtVaCreatePopupShell ("mapFile", simpleMenuWidgetClass, ws, NULL);
    ws = XtVaCreateManagedWidget ("save", smeBSBObjectClass, wc, NULL);
    XtAddCallback (ws, XtNcallback, CbEditSave,
      (XtPointer) & app.list[map]);
    ws = XtVaCreateManagedWidget ("saveAs", smeBSBObjectClass, wc, NULL);
    XtAddCallback (ws, XtNcallback, CbEditSaveAs,
      (XtPointer) & app.list[map]);
    ws = XtVaCreateManagedWidget ("load", smeBSBObjectClass, wc, NULL);
    XtAddCallback (ws, XtNcallback, CbEditLoad,
      (XtPointer) & app.list[map]);
    ws = XtVaCreateManagedWidget ("enter", smeBSBObjectClass, wc, NULL);
    XtAddCallback (ws, XtNcallback, CbEditEnter,
      (XtPointer) & app.list[map]);

    XtVaCreateManagedWidget ("line", smeLineObjectClass, wc, NULL);
    ws = XtVaCreateManagedWidget ("quit", smeBSBObjectClass, wc, NULL);
    XtAddCallback (ws, XtNcallback, CbEditQuit,
      (XtPointer) & app.list[map]);

    /*** edit-menu ***/
    ws = XtVaCreateManagedWidget ("editButton", menuButtonWidgetClass, hbox,
      XtNmenuName, "mapEdit",
      NULL);
    wc = XtVaCreatePopupShell ("mapEdit", simpleMenuWidgetClass, ws, NULL);
    app.list[map].iw.path = XtVaCreateManagedWidget ("path",
      smeBSBObjectClass, wc, NULL);
    XtAddCallback (app.list[map].iw.path, XtNcallback,
      CbEditSetPath, (XtPointer) & app.list[map]);
    app.list[map].iw.size = XtVaCreateManagedWidget ("size",
      smeBSBObjectClass, wc, NULL);
    XtAddCallback (app.list[map].iw.size, XtNcallback,
      CbEditResize, (XtPointer) & app.list[map]);
    app.list[map].iw.start = XtVaCreateManagedWidget ("start",
      smeBSBObjectClass, wc, NULL);
    XtAddCallback (app.list[map].iw.start, XtNcallback,
      CbEditStart, (XtPointer) & app.list[map]);

    /*** toggles ***/
    XtVaCreateManagedWidget ("line", smeLineObjectClass, wc, NULL);
    app.list[map].iw.read_only
      = XtVaCreateManagedWidget ("readOnly", smeBSBObjectClass, wc, NULL);
    XtAddCallback (app.list[map].iw.read_only,
      XtNcallback, CbEditToggleRead, (XtPointer) & app.list[map]);
    app.list[map].iw.overwrite
      = XtVaCreateManagedWidget ("overWrite", smeBSBObjectClass, wc, NULL);
    XtAddCallback (app.list[map].iw.overwrite,
      XtNcallback, CbEditToggleOver, (XtPointer) & app.list[map]);
    app.list[map].iw.auto_choose
      = XtVaCreateManagedWidget ("autoChoose", smeBSBObjectClass, wc, NULL);
    XtAddCallback (app.list[map].iw.auto_choose,
      XtNcallback, CbEditToggleAuto, (XtPointer) & app.list[map]);
    app.list[map].iw.sparse
      = XtVaCreateManagedWidget ("sparse", smeBSBObjectClass, wc, NULL);
    XtAddCallback (app.list[map].iw.sparse,
      XtNcallback, CbEditToggleSparse, (XtPointer) & app.list[map]);

    /*** editor-widget ***/
    app.list[map].view = XtVaCreateManagedWidget ("view",
      viewportWidgetClass, vbox, NULL);
    app.list[map].w = XtVaCreateManagedWidget ("cross",
      widgetClass, app.list[map].view, NULL);

    /*** show-map-editor ***/
    UiEditRefresh (map);
    UiEditSizeRefresh (map);
    XtRealizeWidget (app.list[map].top);
    wattr.backing_store = WhenMapped;
    XChangeWindowAttributes (XtDisplay (app.w), windowOfMap (map),
      CWBackingStore, &wattr);
    XtPopup (app.list[map].top, XtGrabNone);
    UiEditShow (map);
}

void CbMainLoad (Widget w, XtPointer client, XtPointer call)
{
    char *path = (char *) client;

    switch (CnvPrompt (app.w, "Open map", path, &path, CnvOkCancel)) {
    case 1:
	if (!*path) {
	    CnvNotify (app.w, "Cannot open map without name !", CnvOk);
	    return;
	}
	editType = Regular;
	CbEditOpen (w, (XtPointer) path, call);
	break;
    case 2:
	return;
    default:
	efatal ("CbMainLoad(),CnvPrompt()");
    }
}

void CbMainQuit (Widget w, XtPointer client, XtPointer call)
{
    int map;

    /*** for all maps ***/
    for (map = 0; map < MAPLIST_SIZE; map++) {
	if (app.list[map].emap) {
	    CbEditQuit (app.w, (XtPointer) & app.list[map], call);
	    if (app.list[map].emap)
		return;
	}
    }

    EdQuit ();
    XtDestroyApplicationContext (XtWidgetToApplicationContext (w));
    debug0 ("Leaving editor, good bye ... see you soon\n");
    exit (0);
}

void CbMainInfoClose (Widget w, XtPointer client, XtPointer call)
{
    Widget popup = (Widget) client;

    XtDestroyWidget (popup);
}

void CbMainInfo (Widget w, XtPointer client, XtPointer call)
{
    /* absolute filepath are given here */
    char *path = (char *) client;

    if (path) {
	Widget popup, vpane, hbox, dismiss, text;

	debug1 ("CbMainHelp() by %s\n", path);
	popup = XtVaCreatePopupShell ("info", topLevelShellWidgetClass,
	  app.w, NULL);
	vpane = XtVaCreateManagedWidget ("vpane", panedWidgetClass, popup,
	  NULL);
	hbox = XtVaCreateManagedWidget ("hbox", boxWidgetClass, vpane, NULL);
	dismiss = XtVaCreateManagedWidget ("dismiss", commandWidgetClass,
	  hbox, NULL);
	/* give widget to destroy */
	XtAddCallback (dismiss, XtNcallback, CbMainInfoClose, (XtPointer) popup);
	text = XtVaCreateManagedWidget ("text", asciiTextWidgetClass, vpane,
	  XtNeditType, XawtextRead,
	  XtNtype, XawAsciiFile,
	  XtNstring, path,
	  NULL);
	XtPopup (popup, XtGrabNone);
    }
}

void CbMainCrossfire  (Widget w, XtPointer client, XtPointer call)
{
  int pid;
  switch(pid = fork()) {
  case -1:
    CnvNotify(app.w,"Cannot open new process",CnvOk);
    return;
  case 0: /* child */
    if(execlp("crossfire","crossfire",NULL) == -1) {
      die("Cannot execute crossfire");
    }
  default: /* parent */
    debug1("CbMainCrossfire() %i started\n",pid);
    return;
  }
}

/**********************************************************************/

void CbEditLoad (Widget w, XtPointer client, XtPointer call)
{
    int i;
    Edit *map = (Edit *) client;
    char buf[BUFSIZ];
    char *path;

    debug1 ("CbEditLoad() %s\n", map->mappath);

    /*** save, if modified ***/
    if (map->modified) {
	switch (CnvNotify (map->w, "Map modified, save", CnvOkCancelNo)) {
	case 1:
	    CbEditSave (map->w, client, call);
	    break;
	case 2:
	    return;
	default:
	    break;
	}
    }
    switch (i = CnvPrompt (map->w, "Enter level (path)", map->mappath,
	&path, CnvOkCancel)) {
    case 1:
	strcpy (map->mappath, path);
	switch (EdLoadMap (map->id, map->mappath)) {
	case 0:
	    UiEditShow (map->id);
	    break;
	case 1:
	    CnvNotify (map->w, "Map already in memory", CnvOk);
	    return;
	case 2:
	default:
	    sprintf (buf, "Error in loading map %s", map->mappath);
	    CnvNotify (map->w, buf, CnvOk);
	    return;
	}
	break;
    case 2:
	return;
    default:
	debug1 ("CnvPrompt error %i\n", i);
	break;
    }
    UiEditSizeRefresh (map->id);
}

void CbEditQuit (Widget w, XtPointer client, XtPointer call)
{
    char buf[BUFSIZ];
    Edit *map = (Edit *) client;

    debug1 ("CbEditQuit() %i\n", map->id);

    if (map->modified) {
	sprintf (buf, "Quitting, save %s ?", map->mappath);
	switch (CnvNotify (map->w, buf, CnvOkCancelNo)) {
	case 1:		/* ok */
	    CbEditSave (w, client, call);
	    break;
	case 2:		/* cancel */
	    return;
	case 3:		/* no */
	    break;
	default:
	    efatal ("CnvNotify()");
	}
    }
    XtPopdown (map->top);
    XtDestroyWidget (map->top);
    EdFreeMap (map->id);

    if (map->id == app.look.map) {
	CbAttrCancel (NULL, NULL, NULL);
	EdDrawLook ();
    }
    map->id = -1;
}


void CbEditSave (Widget w, XtPointer client, XtPointer call)
{
    char *path;
    Edit *map = (Edit *) client;

    if (map->read_only) {
	CnvNotify (map->w, "Map is read-only!", CnvOk);
	return;
    }
    if (!map->modified) {
	CnvNotify (map->w, "No changes to save", CnvOk);
	return;
    }
    if (!*(map->mappath)) {
	switch (CnvPrompt (map->w, "Save map as", "", &path, CnvOkCancel)) {
	case 1:
	    strcpy (map->mappath, path);
	    UiEditRefresh (map->id);
	    break;
	case 2:
	    return;
	default:
	    efatal ("CnvPrompt()");
	}
    }
    if (EdSaveMap (map->id, map->mappath)) {
	CnvNotify (map->w, "Error in saving map !", CnvOk);
    }
}

void CbEditSaveAs (Widget w, XtPointer client, XtPointer call)
{
    Edit *map = (Edit *) client;
    *(map->mappath) = 0;
    map->modified = 1;
    CbEditSave (w, client, call);
}


void CbEditResize (Widget w, XtPointer client, XtPointer call)
{
    Edit *map = (Edit *) client;
    char buf[1025], *path;
    unsigned x, y;
    int sx = 0, sy = 0, res;

    if (map->read_only)
	return;

    sprintf (buf, "%ix%i+0+0", map->emap->mapx, map->emap->mapy);

    switch (CnvPrompt (app.w, "ResizeScroll", buf, &path, CnvOkCancel)) {
    case 1:
	res = XParseGeometry (path, &sx, &sy, &x, &y);

	if (!(res & WidthValue))
	    x = map->emap->mapx;
	if (!(res & HeightValue))
	    y = map->emap->mapy;

	/* if (x < 0) x += map->emap->mapx; if (y < 0) y += map->emap->mapy; */

	if (!(res & XValue))
	    sx = 0;
	if (!(res & YValue))
	    sy = 0;

	if (x < 3 || y < 3 || x > 100 || y > 100) {
	    CnvNotify (app.w, "Illegal size!", CnvOk);
	    return;
	}
	ResizeMap (map->id, x, y);
	ScrollMap (map->id, sx, sy);
	UiEditShow (map->id);
	UiEditSizeRefresh (map->id);
	break;
    case 2:
	return;
    default:
	efatal ("CbEditResize(),CnvPrompt()");
    }
}

void CbEditStart (Widget w, XtPointer client, XtPointer call)
{
    Edit *edit = (Edit *) client;
    char buf[BUFSIZ], *reply;
    int x, y, flags;
    unsigned width, height;

    debug1 ("CbEditStart(), editor %i\n", edit->id);
    sprintf (buf, "%ix%i", edit->emap->startx,
      edit->emap->starty);
    switch (CnvPrompt (app.w, "Start of map", buf, (char **) &reply,
	CnvOkCancel)) {
    case 1:			/* ok */
	flags = XParseGeometry (reply, &x, &y, &width, &height);
	if ((flags != (WidthValue | HeightValue))) {
	    CnvNotify (app.w, "Illegal size", CnvOk);
	    return;
	}
	if (EdSetStart (edit->id, width, height)) {
	    sprintf (buf, "Error at size %ix%i", width, height);
	    CnvNotify (app.w, buf, CnvOk);
	    return;
	}
	UiEditRefresh (edit->id);
	return;
    case 2:			/* cancel */
	return;
    default:
	return;
    }
}

void CbEditSetPath (Widget w, XtPointer client, XtPointer call)
{
    char *path;
    Edit *map = (Edit *) client;

    switch (CnvPrompt (map->w, "Set path", map->mappath, &path, CnvOkCancel)) {
    case 1:
	strcpy (map->mappath, path);
	UiEditRefresh (map->id);
	break;
    case 2:
	return;
    default:
	efatal ("CnvPrompt()");
    }
}

/* toggle write accces */
void CbEditToggleRead (Widget w, XtPointer client, XtPointer call)
{
    ((Edit *) client)->read_only++;
    UiEditRefresh (((Edit *) client)->id);
    debug0 ("CbEditToggleRead()\n");
}

/* Toggle overwrite mode */
void CbEditToggleOver (Widget w, XtPointer client, XtPointer call)
{
    ((Edit *) client)->overwrite++;
    UiEditRefresh (((Edit *) client)->id);
    debug0 ("CbEditToggleOver()\n");
}

/* Toggle auto choose */
void CbEditToggleAuto (Widget w, XtPointer client, XtPointer call)
{
    ((Edit *) client)->auto_choose++;
    UiEditRefresh (((Edit *) client)->id);
    debug0 ("CbEditToggleAuto()\n");
}

/* Toggle sparse mode */
void CbEditToggleSparse (Widget w, XtPointer client, XtPointer call)
{
    Edit *m = (Edit *) client;
    m->sparse++;
    m->fontsize = m->sparse ? STACK_SIZE : FONTSIZE; 
    UiEditSizeRefresh (m->id);
    UiEditShow (m->id);
    debug0 ("CbEditToggleSparse()\n");
}

/* enter to map in Look window */
void CbEditEnter (Widget w, XtPointer client, XtPointer call)
{
    Edit *edit = (Edit *) client;
    object *obj;
    char buf[BUFSIZ];

    debug0("CbEditEnter()\n");

    obj = get_nth_map_ob (app.list[app.look.map].emap,
			  app.look.x,app.look.y,0);
    if (!obj) {
	CnvNotify(app.w,"Select exit to enter",CnvOk);
	return;
      }
    if (obj->head)
	obj = obj->head;

    do {
      switch (obj->type) {
      case STAIR:
      case TELEPORTER:
      case EXIT:

      if (edit->modified) {
	switch (CnvNotify (edit->w, "Map modified, save", CnvOkCancelNo)) {
	case 1:
	    CbEditSave (edit->w, client, call);
	    break;
	case 2:
	    return;
	default:
	    break;
	}
       }
       if(obj->stats.food < 1) {
	 CnvNotify(app.w,"Exit leads nowhere",CnvOk);
	 return;
       }
       sprintf(buf,"%i",obj->stats.food);
       strcpy(edit->mappath,buf);
       EdLoadMap(edit->id,buf);
       UiEditSizeRefresh (edit->id);
       UiEditRefresh (edit->id);
       UiEditShow (edit->id);
       return;
     }
     obj = obj->below;
    } while(obj);
   CnvNotify(app.w,"That's not exit",CnvOk);
}



/**********************************************************************/

void CbMainSelectMaps (Widget w, XtPointer client, XtPointer call)
{
    editType = Pick;
    debug0 ("CbMainSelectMaps()\n");
    CbEditOpen (app.w, client, call);
}

void CbMainWallMaps (Widget w, XtPointer client, XtPointer call)
{
    editType = Wall;
    debug0 ("CbWallMaps()\n");
    CbEditOpen (app.w, client, call);
}

void CbMainToggleEditable (Widget w, XtPointer client, XtPointer call)
{
    Arg arglist[1];
    debug0 ("CbToggle()\n");
    app.arch.editable = !app.arch.editable;
    EdInitArch ();
    EdDrawArch ();
    XtVaSetValues (app.arch.w,
      XtNheight, app.arch.height * FONTSIZE,
      NULL);
    XtSetArg (arglist[0], XtNleftBitmap, app.arch.editable ? mark : None);
    XtSetValues (w, arglist, 1);
}

void CbMainToggleMonsters (Widget w, XtPointer client, XtPointer call)
{
    Arg arglist[1];
    debug0 ("CbToggle()\n");
    app.arch.monsters = !app.arch.monsters;
    EdInitArch ();
    EdDrawArch ();
    XtVaSetValues (app.arch.w,
      XtNheight, app.arch.height * FONTSIZE,
      NULL);
    XtSetArg (arglist[0], XtNleftBitmap, app.arch.monsters ? mark : None);
    XtSetValues (w, arglist, 1);
}

/**********************************************************************/


void CbAttrOpen (Widget w, XtPointer client, XtPointer call)
{
    Widget view, box;
    int map, x, y, index;
    XEvent *e = (XEvent *) client;

    if (w == app.look.w) {
	map = app.look.map;
	x = app.look.x;
	y = app.look.y;
	index = e->xbutton.y / FONTSIZE;
    } else {			/* its map */
	map = UiEditIndexFromWidget (w);
	x = e->xbutton.x / app.list[map].fontsize;
	y = e->xbutton.y / app.list[map].fontsize;
	index = 0;
    }
    debug4 ("CbAttrOpen(): %i, %ix%i, %i\n", map, x, y, index);

    if (GetObjectValues (map, x, y, index)) {
	/* CnvNotify (app.w, "GetObjectValues()", CnvOk); */
	return;
    }
    if (!app.attr.w) {
	Widget vpane, hbox;
	app.attr.w = XtVaCreatePopupShell ("attr", topLevelShellWidgetClass,
	  app.w,
	  NULL);

	vpane = XtVaCreateManagedWidget ("vpane", panedWidgetClass,
	  app.attr.w, NULL);
	hbox = XtVaCreateManagedWidget ("hbox", boxWidgetClass, vpane,
	  XtNorientation, XtorientHorizontal,
	  NULL);
	CnvAddButton (hbox, "ok", CbAttrQuit);
	CnvAddButton (hbox, "apply", CbAttrApply);
	CnvAddButton (hbox, "cancel", CbAttrCancel);
	view = XtVaCreateManagedWidget ("view", viewportWidgetClass, vpane,
	  XtNallowHoriz, True,
	  XtNallowVert, True,
	  NULL);
	box = XtVaCreateManagedWidget ("box", boxWidgetClass, view,
	  XtNorientation, XtorientVertical,
	  NULL);
	CnvTagsOpen (app_attr_tags, box);

	CnvTagsManage (app_attr_tags, app.attr.flags);
	XtPopup (app.attr.w, XtGrabNone);
    } else {
	CnvTagsManage (app_attr_tags, app.attr.flags);
    }
}

void CbAttrQuit (Widget w, XtPointer client, XtPointer call)
{
    CbAttrApply (w, client, call);
    CbAttrCancel (w, client, call);
}

void CbAttrApply (Widget w, XtPointer client, XtPointer call)
{
    debug1( "CbAttrApply() %s\n" ,app_attr_tags[I_Path].value);
    SetObjectValues ();
}

void CbAttrCancel (Widget w, XtPointer client, XtPointer call)
{
    debug0 ("CbAttrCancel()\n");

    if (app.attr.w)
	XtDestroyWidget (app.attr.w);
    app.attr.w = 0;
}

/*** End of crossfire.c ***/
