/*
 * static char *rcsid_editor_c =
 *   "$Id: editor.c,v 1.1.1.1 1993/03/07 08:31:02 frankj Exp $";
 */

/*
	CrossEdit, A Editor for CrossFire, A Multiplayer game
	for X-windows	

	Copyright (C) 1993 Jarkko Sonninen & Petri Heinila

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	The authors can be reached via e-mail to Jarkko.Sonninen@lut.fi
	or Petri.Heinila@lut.fi .
*/

#include "editor.h"
#include "debug.h"
#include "Cnv.h"

extern object *get_nth_map_ob (mapstruct * emap, int x, int y, int i);

/**********************************************************************/

void EdInitialize ()
{
    XGCValues gcval;

    /* crossfire init */
    init_library ();
    init_archetypes ();

    if (!check_font_path(XtDisplay (app.w))) {
      fprintf(stderr, "Trying to fix fontpath...\n");
      fflush(stderr);
      set_font_path (XtDisplay (app.w), FONTDIR);
      if (!check_font_path(XtDisplay (app.w))) {
        fprintf(stderr, "Failed to set fontpath.\n");
        exit(1);
      }
    }

    app.font = gcval.font = XLoadFont (XtDisplay (app.w), FONTNAME);
    app.gc = XtAllocateGC (app.w, 0, GCFont, &gcval,
      GCBackground | GCForeground, 0);


#ifdef DEBUG
    debug = 1;
#else
    debug = 0;
#endif


    app.arch.monsters = 1;
    app.arch.editable = 1;
    EdInitArch ();
    find_colors ();
}

void EdQuit ()
{
    XUnloadFont (XtDisplay (app.w), app.font);
    XtReleaseGC (app.w, app.gc);
    XtReleaseGC (app.w, app.text_gc);
}


/**********************************************************************/

int EdFreeMap (int map)
{
    if (app.item.wall_map == app.list[map].emap) {
	app.item.wall_map = NULL;
	EdDrawItem ();
    }
    app.list[map].emap->level = -2;
    *app.list[map].emap->name = '\0';
    free_map (app.list[map].emap, 1);
    app.list[map].emap = NULL;
    app.list[map].modified = 0;
    *app.list[map].mappath = '\0';
    return (0);
}

int EdEmptyMap (int map, int x, int y)
{
    app.list[map].emap = get_empty_map (x, y);
    if (!app.list[map].emap)
	efatal ("Empty map allocation failed");
    app.list[map].emap->level = 0;
    *app.list[map].emap->name = '\0';
    *app.list[map].mappath = '\0';
    return (0);
}

int EdLoadMap (int map, char *name)
{
    int level;
    mapstruct *tmp;

    level = atoi (name);	/******/
    if (level < 1) {
	debug1 ("Loading Map by name %s\n", name);
	if (check_level (name))
	    return (-2);

	if (app.list[map].emap) {
	    /* I don't want more to do with it for now */
	    app.list[map].emap->level = -2;
	    free_map (app.list[map].emap, 1);
	    if (app.item.wall_map == app.list[map].emap) {
		app.item.wall_map = NULL;
		EdDrawItem ();
	    }
	}
	app.list[map].type = Regular;
	tmp = ready_map_name (name, 1);
	tmp->level = 0;
	tmp->path = add_string (name);

    } else {
	if (map_in_memory_p (level))
	    return (-1);
	debug1 ("Loading Map by level %d\n", level);

	if (app.list[map].emap) {
	    if (app.item.wall_map == app.list[map].emap) {
		app.item.wall_map = NULL;
		EdDrawItem ();
	    }
	    /* I don't want more to do with it for now */
	    app.list[map].emap->level = -2;
	    free_map (app.list[map].emap, 1);
	}
	app.list[map].type = Regular;
	tmp = ready_map (level, 1);
    }
    app.list[map].modified = 0;
    app.list[map].emap = tmp;
    if (tmp)
	return (0);
    return (-2);
}

int EdSaveMap (int map, char *name)
{
    mapstruct *m = app.list[map].emap;
    char *path;
    int i, j;
    int level;

    for (i = 0; i < m->mapx; i++)
	for (j = 0; j < m->mapy; j++)
	    set_omap (m, i, j, 35);

    if (m->path)
	free_string (m->path);
    m->path = NULL;

    level = atoi (name);	/******/
    if (level < 1) {
	debug0 ("Saving Map by name\n");
	m->level = 0;
	m->path = add_string (name);
    } else {
	debug0 ("Saving Map by level\n");
	m->level = level;
	switch (CnvPrompt (app.list[map].w, "Save map as name",
	    m->name, &path, CnvOkCancel)) {
	case 1:
	    strcpy (m->name, path);
	    break;

	case 2:
	    return (-1);

	default:
	    efatal ("CnvPrompt()");
	}
	update_map_name (m->level, m->name);
    }

    save_omap (m, 1);
    save_map (m, 1);

    app.list[map].modified = 0;
    /*EdLoadMap (map, name);*/	/* grrr */
    return (0);
}

void Draw (int i)
{
    XClearWindow (XtDisplay (app.w), windowOfMap (i));
    DrawMap (i, 0, 0, app.list[i].emap->mapx, app.list[i].emap->mapy);
}

void DrawMap (int map, int x, int y, int width, int height)
{
    int i, j;

    height += y;
    if (width > app.list[map].emap->mapx)
	width = app.list[map].emap->mapx;
    if (height > app.list[map].emap->mapy)
	height = app.list[map].emap->mapy;

    for (j = x; j < width; j++)
	for (i = y; i < height; i++) {
	    update_position (map, j, i);
	}
    if (app.look.map == map)
	EdDrawLook ();
}

/**********************************************************************/

int direction_x[] = {0, -1, 0, 1};
int direction_y[] = {1, 0, -1, 0};
int pow2[] = {1, 2, 4, 8};

int draw_add (int map, int x, int y)
{
    int i, mask = 0;

    if (!app.item.wall_map)
	efatal ("No Wall Map\n");
    if (app.item.wall_map->mapx < 16)
	efatal ("Wall Map has wrong width\n");

    for (i = 0; i < 4; i++)
	if (update_wall (map, x + direction_x[i], y + direction_y[i], 
			 pow2[i], 0))
	    mask |= pow2[i ^ 2];
    /* mask  = 4 * update_wall(windowOfMap(map), app.list[map].emap, x, y +
       1, 1, 0); mask += 8 * update_wall(windowOfMap(map),
       app.list[map].emap, x - 1, y, 2, 0);    mask += 1 *
       update_wall(windowOfMap(map), app.list[map].emap, x, y - 1, 4, 0);
       mask += 2 * update_wall(windowOfMap(map), app.list[map].emap, x + 1,
       y, 8, 0); */
    update_wall (map, x, y, mask, 0);
    return (0);
}

int draw_remove (int map, int x, int y)
{
    if (!app.item.wall_map)
	efatal ("No Wall Map\n");
    if (app.item.wall_map->mapx < 16)
	efatal ("Wall Map has wrong width\n");

    update_wall (map, x, y + 1, 0, 1);
    update_wall (map, x - 1, y, 0, 2);
    update_wall (map, x, y - 1, 0, 4);
    update_wall (map, x + 1, y, 0, 8);
    return (0);
}


int EdDeleteObject (int map, int x, int y, int i)
{
    if (!app.list[map].read_only &&
      app.list[map].emap && x >= 0 && y >= 0 && i >= 0 &&
      x < app.list[map].emap->mapx && y < app.list[map].emap->mapy) {
	e_remove_object (map, x, y, i);
	app.list[map].modified = 1;
	if (app.item.wall_map)
	    draw_remove (map, x, y);
    }
    return (0);
}

int EdInsertObject (int map, int x, int y, int i)
{
    object *tmp;

    /* debug4("InsertObject(%i %i %i %i)\n", map, x, y, i); */
    if (app.list[map].read_only || !app.list[map].emap ||
	out_of_map (app.list[map].emap, x, y))
	return (-1);

    if (!app.item.at) {
	CnvNotify (app.w, "Select item to insert.", CnvOk);
	return (-1);
    }
    app.list[map].modified = 1;
    if (app.list[map].overwrite)
	while (get_map_ob (app.list[map].emap, x, y))
	    e_remove_object (map, x, y, 0);
    else {
	tmp = get_nth_map_ob (app.list[map].emap, x, y, i);
	if (tmp && (tmp->arch == app.item.at ||
	    (app.item.wall_map &&
	      find_draw_arch (app.item.wall_map,
		app.item.wall_set,
		tmp->arch) >= 0)))
	    return (0);
    }

    e_insert_object (map, x, y, i, app.item.at);
    if (app.item.wall_map)
	draw_add (map, x, y);
    return (0);
}

/**********************************************************************/

int EdSelectItem (int map, int x, int y)
{
    object *tmp;
    tmp = get_nth_map_ob (app.list[map].emap, x, y, 0);
    if (tmp && tmp->head)
	tmp = tmp->head;

    if (tmp && tmp->arch) {
	/* magical joining of objects */
	if (app.list[map].type == Wall && tmp->x == 0) {
	    app.item.wall_set = y;
	    app.item.wall_map = tmp->map;
	    debug0 ("Auto Joining Turned ON\n");
	} else
	    app.item.wall_map = NULL;
	app.item.at = tmp->arch;
	EdDrawItem ();
    }
    return (0);
}

int EdDrawLook ()
{
    XClearWindow (XtDisplay (app.w), XtWindow (app.look.w));
    if (app.list[app.look.map].emap &&
      !(out_of_map (app.list[app.look.map].emap,
	  app.look.x, app.look.y))) {
	e_draw_selection (XtWindow (app.look.w),
	  app.list[app.look.map].emap,
	  app.look.x, app.look.y);

    }
    return (0);
}

int EdDrawArch ()
{
    XClearWindow (XtDisplay (app.w), XtWindow (app.arch.w));

    draw_old_list (XtWindow (app.arch.w), app.arch.height = app.arch.total,
      app.arch.ch_table);
    return (0);
}

int EdDrawItem ()
{
    int minx = 0, miny = 0, maxx = 0;
    archetype *at;

    XClearWindow (XtDisplay (app.w), XtWindow (app.item.w));
    /* search for size */
    for (at = app.item.at; at; at = at->more) {
	if (maxx < at->clone.x)
	    maxx = at->clone.x;
	if (minx > at->clone.x)
	    minx = at->clone.x;
	if (miny > at->clone.y)
	    miny = at->clone.y;
    }
    for (at = app.item.at; at; at = at->more) {
	draw_object (XtWindow (app.item.w),
	  at->clone.x - minx, at->clone.y - miny, &at->clone);
    }
    if (app.item.at) {
	XDrawString (XtDisplay (app.w), XtWindow (app.item.w),
	  app.text_gc, (maxx - minx) * FONTSIZE + 32, 16,
	  app.item.at->name, strlen (app.item.at->name));
	if (app.item.wall_map)
	    XDrawString (XtDisplay (app.w), XtWindow (app.item.w),
	      app.text_gc,
	      (maxx - minx) * FONTSIZE + 32, 16 + FONTSIZE,
	      "(auto-joining)", 14);
	debug1 ("Draw Chosen %s\n", app.item.at->clone.name);
    }
    return (0);
}

int ArchChose (int x, int y)
{
    debug2 ("Arch Chose %i %i\n", x, y);
    if (y < app.arch.total && y >= 0) {
	app.item.at = app.arch.ch_table[y];
	app.item.wall_map = NULL;
	EdDrawItem ();
    }
    return (0);
}

int EdSetStart (int map, int x, int y)
{
    if (out_of_map (app.list[map].emap, x, y))
	return (-1);
    app.list[map].modified = 1;
    app.list[map].emap->startx = x;
    app.list[map].emap->starty = y;
    return (0);
}


int ResizeMap (int map, int x, int y)
{
    mapstruct *tmp;

    debug3 ("ResizeMap %i %i %i\n", map, x, y);
    if (x == app.list[map].emap->mapx && y == app.list[map].emap->mapy)
	return (0);

    tmp = get_empty_map (x, y);
    /* copy_map (app.list[map].emap, tmp); */
    app.list[map].modified = 1;
    app.list[map].emap->level = -2;
    move_all_objects (app.list[map].emap, tmp);
    free_map (app.list[map].emap, 0);
    /* The 0 means we don't free any objects */
    app.list[map].emap = tmp;
    return (0);
}

int ScrollMap (int map, int x, int y)
{
    mapstruct *tmp;

    debug3 ("ScrollMap %i %i %i\n", map, x, y);
    if (!x && !y)
	return (0);

    tmp = get_empty_map (app.list[map].emap->mapx,
      app.list[map].emap->mapy);
    app.list[map].modified = 1;
    app.list[map].emap->level = -2;
    scroll_move_objects (app.list[map].emap, tmp, x, y);
    free_map (app.list[map].emap, 0);
    /* The 0 means we don't free any objects */
    app.list[map].emap = tmp;
    return (0);
}

/**********************************************************************/

int GetObjectValues (int map, int x, int y, int i)
{
    object *tmp;

    tmp = get_nth_map_ob (app.list[app.look.map].emap, x, y, i);
    if (!tmp)
	return (-1);

    if (tmp->head)
	tmp = tmp->head;

    app.attr.op = tmp;
    if (tmp->name)
	strcpy (app_attr_tags[I_Name].value, tmp->name);
    else
	*app_attr_tags[I_Name].value = '\0';

    if (tmp->slaying) {
	strcpy (app_attr_tags[I_Sacrifice].value, tmp->slaying);
	strcpy (app_attr_tags[I_Trigger].value, tmp->slaying);
    } else {
	*app_attr_tags[I_Sacrifice].value = '\0';
	*app_attr_tags[I_Trigger].value = '\0';
    }

    if (tmp->msg)
	strcpy (app_attr_tags[I_Message].value, tmp->msg);
    else
	*app_attr_tags[I_Message].value = '\0';

    /* if (tmp->path) strcpy(app_attr_tags[I_Path].value, tmp->path); else */
    sprintf (app_attr_tags[I_Path].value, "%d", EXIT_LEVEL (tmp));

    sprintf (app_attr_tags[I_X].value, "%d", EXIT_X (tmp));
    sprintf (app_attr_tags[I_Y].value, "%d", EXIT_Y (tmp));

    sprintf (app_attr_tags[I_Weight].value, "%d", tmp->weight);
    sprintf (app_attr_tags[I_Connect].value, "%d", tmp->stats.sp);
    sprintf (app_attr_tags[I_Hp].value, "%d", tmp->stats.hp);
    sprintf (app_attr_tags[I_Count].value, "%d", tmp->stats.food);

    switch (tmp->type) {
    case STAIR:
    case TELEPORTER:
    case EXIT:
	app.attr.flags = T_Exit;
	break;

    case TRAPDOOR:
	app.attr.flags = T_Trapdoor;
	break;

    case HOLE:
	app.attr.flags = T_Pit;
	break;

    case GATE:
    case BUTTON:
    case HANDLE:
	app.attr.flags = T_Connected;
	break;

    case ALTAR:
	app.attr.flags = T_Altar;
	break;

    case PEDESTAL:
	app.attr.flags = T_Pedestal;
	break;

    case BOOK:
    case SIGN:
	app.attr.flags = T_Sign;
	break;

    default:
	if (IS_MONSTER (app.attr.op))
	    app.attr.flags = T_Monster;
	else
	    app.attr.flags = T_Default;
	break;
    }
    debug2 ("Getvalues %s type %02X\n", tmp->name, app.attr.flags);
    return (0);
}

int SetObjectValues ()
{
    object *tmp = app.attr.op;
    int level = -1;
    int i;

    if (!tmp)
	efatal ("No Object\n");

    while (tmp) {
	if (app.attr.flags & T_Name) {
	    if (tmp->name)
		free_string (tmp->name);
	    tmp->name = add_string (app_attr_tags[I_Name].value);
	}

	if (app.attr.flags & T_Message) {	
	    if (tmp->msg)
		free_string (tmp->msg);
	    tmp->msg = NULL;
	    debug1("Store message:<%s>\n", app_attr_tags[I_Message].value);
	    if ((i = strlen(app_attr_tags[I_Message].value))) {
		if (app_attr_tags[I_Message].value[i - 1] != '\n')
		    strcat (app_attr_tags[I_Message].value, "\n");
		if (i > 1 ||  *app_attr_tags[I_Message].value != '\n')
		    tmp->msg = add_string (app_attr_tags[I_Message].value);
	    }
	}

	if (app.attr.flags & T_Path) {
	    String val;
	    level = atoi (app_attr_tags[I_Path].value);
	    /*
	       XtVaGetValues(app_attr_tags[I_Path].t, XtNstring,&val, NULL);
	       */
	    val = app_attr_tags[I_Path].value;
	    debug2("SetObjectValues() %s vs %s\n",app_attr_tags[I_Path].value,val);
	    level = atoi (val);
	    if (level || *val == '0') {
		EXIT_LEVEL (tmp) = level;
	    } else {
		char buf[BUFSIZ];
		sprintf(buf,"CrossFire doesn't support pathnames as '%s' :-(",
			app_attr_tags[I_Path].value);
		CnvNotify(app.w,buf,CnvOk);
	    }
	    /* else tmp->path = app_attr_tags[I_Path].value; */
	}
	
	if (app.attr.flags & T_Sacrifice) {
	    if (tmp->slaying)
		free_string (tmp->slaying);
	    tmp->slaying = add_string (app_attr_tags[I_Sacrifice].value);
	}

	if (app.attr.flags & T_Trigger) {
	    if (tmp->slaying)
		free_string (tmp->slaying);
	    tmp->slaying = add_string (app_attr_tags[I_Trigger].value);
	}
	
	if (app.attr.flags & T_X)
	    EXIT_X (tmp) = atoi (app_attr_tags[I_X].value);
	
	if (app.attr.flags & T_Y)
	    EXIT_Y (tmp) = atoi (app_attr_tags[I_Y].value);
	
	if (app.attr.flags & T_Weight)
	    tmp->weight = atoi (app_attr_tags[I_Weight].value);
	
	if (app.attr.flags & T_Connect)
	    tmp->stats.sp = atoi (app_attr_tags[I_Connect].value);
	
	if (app.attr.flags & T_Hp)
	    tmp->stats.hp = atoi (app_attr_tags[I_Hp].value);
	
	if (app.attr.flags & T_Count)
	    tmp->stats.food = atoi (app_attr_tags[I_Count].value);
	
	EdDrawLook ();
	update_position (app.look.map, app.look.x, app.look.y);
	tmp = tmp->more;
    }
    app.list[app.look.map].modified = 1;
    return (0);
}


/**********************************************************************/

int arch_show (archetype * at)
{
    return ((!app.arch.editable || EDITABLE ((&at->clone))) &&
      (!app.arch.monsters || IS_MONSTER ((&at->clone))
	|| IS_GENERATOR ((&at->clone))));
}

int EdInitArch ()
{
    int j, total = 0;
    archetype *at = first_archetype;

    if (app.arch.ch_table)
	free (app.arch.ch_table);

    while (at) {
	if (arch_show (at))
	    total++;
	at = at->next;
    }

    app.arch.width = 8;
    app.arch.total = app.arch.height = total;
    debug1 ("%i objects\n", total);
    app.arch.ch_table = (archetype **) calloc (total, sizeof (archetype *));

    for (j = 0, at = first_archetype; at; at = at->next)
	if (arch_show (at))
	    app.arch.ch_table[j++] = at;
    return (total);
}
