/*
 * static char *rcsid_editor_h =
 *   "$Id: editor.h,v 1.1.1.1 1993/03/07 08:31:03 frankj Exp $";
 */

/*
	CrossEdit, A Editor for CrossFire, A Multiplayer game
	for X-windows	

	Copyright (C) 1993 Jarkko Sonninen & Petri Heinila

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	The authors can be reached via e-mail to Jarkko.Sonninen@lut.fi
	or Petri.Heinila@lut.fi .
*/

#include <global.h>
#include <Cnv.h>
#ifndef __CEXTRAXT__
#ifndef __STDC__
#define __STDC__ 1
#endif
#include "proto.h"
#endif

#ifdef FONTSIZE
#undef FONTSIZE
#endif

#define FONTSIZE 24
#define STACK_SIZE 48
#define STACK_DEPTH 6
#define MAPLIST_SIZE 10
#define STYPELIST_SIZE 10
#define FILENAME_SIZE 256

#define windowOfMap(map) XtWindow(app.list[map].w)

/**********************************************************************/


extern CnvTags app_attr_tags[];

#define I_Name      0
#define I_Message   1
#define I_X         2
#define I_Y         3
#define I_Path      4
#define I_Weight    5
#define I_Connect   6
#define I_Hp        7
#define I_Trigger   8
#define I_Sacrifice 9
#define I_Count     10

#define T_Name      (1<<I_Name)
#define T_Message   (1<<I_Message)
#define T_X         (1<<I_X)
#define T_Y         (1<<I_Y)
#define T_Path      (1<<I_Path)
#define T_Weight    (1<<I_Weight)
#define T_Connect   (1<<I_Connect)
#define T_Hp        (1<<I_Hp)
#define T_Trigger   (1<<I_Trigger)
#define T_Sacrifice (1<<I_Sacrifice)
#define T_Count     (1<<I_Count)

#define T_Exit      (T_Name | T_Path | T_X | T_Y )
#define T_Trapdoor  (T_Name | T_Weight | T_X | T_Y )
#define T_Connected (T_Name | T_Connect )
#define T_Pit       (T_Name | T_Connect | T_X | T_Y )
#define T_Monster   (T_Name | T_Hp )
#define T_Pedestal  (T_Name | T_Connected | T_Trigger )
#define T_Altar     (T_Name | T_Message | T_Connected | T_Sacrifice | T_Count )
#define T_Sign      (T_Name | T_Message )
#define T_Default   (T_Name)

/**********************************************************************/

typedef enum {
    Regular = 1, Pick, Wall
}   EditType;

typedef struct {
    Widget w;			/* mapEditor widget */
    Widget view;		/* viewport of mapEditor */
    Widget top;			/* shell widget of map */
    mapstruct *emap;
    char mappath[FILENAME_SIZE + 1];

    int modified:1;
    /* toggle-flags */
    int read_only:1;
    int overwrite:1;
    int auto_choose:1;
    int sparse:1;
    int fontsize;
    /* info widgets */
    struct {
	/* Widget level; */
	Widget path;
	Widget size;
	Widget start;
	/* Widget name; */
	/* toggles */
	Widget read_only;
	Widget overwrite;
	Widget auto_choose;
	Widget sparse;
    }   iw;
    int id;			/* which map, negative, if not exist, ui only */
    EditType type;
}   Edit;

typedef struct {
    Widget w;
    unsigned long flags;	/* type */
    object *op;
}   Attr;


typedef struct {
    Widget w;
    int map;
    int x;
    int y;
    int nrof;			/* stack size */
}   Look;

typedef struct {
    Widget w;
    int width;
    int height;
    int editable:1;
    int monsters:1;
    int total;
    archetype **ch_table;
}   Arch;

typedef struct {
    Widget w;
    archetype *at;
    /* object *clone; */
    /* int drawing; */
    int wall_set;
    mapstruct *wall_map;
}   Item;

typedef struct {
    Widget w;
    GC  gc;
    GC  text_gc;
    Font font;
    Edit list[MAPLIST_SIZE];
    Attr attr;
    Look look;
    Arch arch;
    Item item;
}   App;

extern App app;

extern Colormap colormap;
extern XColor exactcolor, discolor[];
extern char colorname[13][16];
extern char coloralias[13][16];
extern int itemcolor[][2];
